package example;

import java.io.*;
import java.util.*;
import javax.swing.*;

import evolve.*;
import evolve.data.*;
import evolve.visualization.*;

import step.*;
import step.io.*;

import example.jvmpi.*; // events

public class EVolveDemo implements DataSource
{
  private File            _inputFile;
  private StepRecordInput _stepIn;
  private Iterator        _iEntities;
  private Iterator        _iElementDefs;

  private JFileChooser _fileChooser = new JFileChooser();
  
  private EntityBuilder
    _envIdBuilder,
    _arenaIdBuilder,
    _classIdBuilder;

  private EventBuilder
    _allocBuilder;

  private FieldDefinition
    _envIdDef,
    _arenaIdDef,
    _classIdDef,
    _sizeDef,
    _objIdDef;
  
  private Map
    _envIdToEntity   = new java.util.HashMap(),
    _arenaIdToEntity = new java.util.HashMap(),
    _classIdToEntity = new java.util.HashMap();

  public void init() throws DataProcessingException
  {
    if ( _fileChooser.showOpenDialog(EVolve.getFrame()) ==
           JFileChooser.APPROVE_OPTION )
    {
      _inputFile = _fileChooser.getSelectedFile();

      if ( ! (_inputFile.isFile() && _inputFile.canRead()) )
        throw new DataProcessingException("File loading failed.");
    }
    else
    {
      throw new DataProcessingException("File loading canceled.");
    }
  }

  public void startBuildDefinition() throws DataProcessingException
  {
    List elementDefs = new java.util.LinkedList();

    _envIdBuilder = new EntityBuilder("envId", "Environment Identifier");
    elementDefs.add(_envIdBuilder.buildDefinition());

    _arenaIdBuilder = new EntityBuilder("arenaId", "Memory Arena Identifier");
    elementDefs.add(_arenaIdBuilder.buildDefinition());

    _classIdBuilder = new EntityBuilder("classId", "Class Identifier");
    elementDefs.add(_classIdBuilder.buildDefinition());


    _allocBuilder =
      new EventBuilder("Object Allocation", "an object was allocated on the heap");

    // envId
    _envIdDef = _allocBuilder.buildReferenceDefinition(
      "Environment Identifier", _envIdBuilder, null, "the (thread) environment");

    // arenaId
    _arenaIdDef = _allocBuilder.buildReferenceDefinition(
      "Allocation Arena", _arenaIdBuilder, null, "the memory arena");

    // classId
    _classIdDef = _allocBuilder.buildReferenceDefinition(
      "Class Identifier", _classIdBuilder, null, "the class identifier");

    // size
    {
      String[] sizeProperties = { "amount" };

      _sizeDef = _allocBuilder.buildValueDefinition(
        "Object Size", sizeProperties, "bytes allocated");
    }

    // objId
    {
      String[] objIdProperties = { "coordinate" };

      _objIdDef = _allocBuilder.buildValueDefinition(
        "Object Address", objIdProperties, "object's initial address");
    }

    elementDefs.add(_allocBuilder.buildDefinition());


    _iElementDefs = elementDefs.iterator();
  }

  public ElementDefinition getNextDefinition() throws DataProcessingException
  {
    if ( _iElementDefs.hasNext() )
    {
      return (ElementDefinition) _iElementDefs.next();
    }
    else
    {
      _iElementDefs = null; // free
      return null;
    }
  }

  public void startBuildEntity() throws DataProcessingException
  {
    List entities = new java.util.LinkedList();

    try
    {
      _stepIn =
        new StepEncodedInput(
        new java.io.BufferedInputStream(
        new java.io.FileInputStream(_inputFile), 65536));

      // skip header record
      AdaptHeader header = (AdaptHeader) _stepIn.readRecord();

      while (true)
      {
        StepRecord record = _stepIn.readRecord();

        if ( record instanceof OBJECT_ALLOC )
        {
          OBJECT_ALLOC r = (OBJECT_ALLOC) record;
          Entity entity;

          if ( ! _envIdToEntity.containsKey(r.envId) )
          {
            _envIdBuilder.newEntity(r.envId.toString());
            _envIdToEntity.put(r.envId, entity = _envIdBuilder.buildEntity());
            entities.add(entity);
          }

          if ( ! _arenaIdToEntity.containsKey(r.arenaId) )
          {
            _arenaIdBuilder.newEntity(r.arenaId.toString());
            _arenaIdToEntity.put(r.arenaId, entity = _arenaIdBuilder.buildEntity());
            entities.add(entity);
          }

          if ( ! _classIdToEntity.containsKey(r.classId) )
          {
            _classIdBuilder.newEntity(r.classId.toString());
            _classIdToEntity.put(r.classId, entity = _classIdBuilder.buildEntity());
            entities.add(entity);
          }
        }
        else if ( record == null ) // end of input
        {
          break;
        }
        else
          throw new java.io.IOException("unknown event: "+record);
      }

      _iEntities = entities.iterator();
    }
    catch (Exception e)
    {
      throw new DataProcessingException(e.toString());
    }
  }

  public Entity getNextEntity() throws DataProcessingException
  {
    if ( _iEntities.hasNext() )
    {
      return (Entity) _iEntities.next();
    }
    else
    {
      _iEntities = null; // free
      return null;
    }
  }

  public void startBuildEvent() throws DataProcessingException
  {
    System.out.println("startBuildEvent(): enter");
    try
    {
      _stepIn =
        new StepEncodedInput(
        new java.io.BufferedInputStream(
        new java.io.FileInputStream(_inputFile), 65536));

      // skip header record
      AdaptHeader header = (AdaptHeader) _stepIn.readRecord();
    }
    catch (Exception e)
    {
      System.out.println("startBuildEvent(): error");
      e.printStackTrace();
      throw new DataProcessingException(e.toString());
    }
    System.out.println("startBuildEvent(): exit");
  }

  public Event getNextEvent() throws DataProcessingException
  {
    System.out.println("getNextEvent(): enter");
    try
    {
      StepRecord record = _stepIn.readRecord();

/*
      if ( record instanceof ARENA_DELETE )
      {
        ARENA_DELETE r = (ARENA_DELETE) record;

        output.writeByte(0x00); // ADAPT_ARENA_DELETE

        output.writeInt(r.envId.intValue());
        output.writeInt(r.arenaId.intValue());
      }
      else if ( record instanceof ARENA_NEW )
      {
        ARENA_NEW r = (ARENA_NEW) record;

        output.writeByte(0x01); // ADAPT_ARENA_NEW

        output.writeInt(r.envId.intValue());
        output.writeInt(r.arenaId.intValue());
        output.writeShort(r.arenaName.toString().length());
        output.write(r.arenaName.toString().getBytes());
      }
      else if ( record instanceof GC_FINISH )
      {
        GC_FINISH r = (GC_FINISH) record;

        output.writeByte(0x09); // ADAPT_GC_FINISH

        output.writeInt(r.envId.intValue());
        output.writeLong(r.usedObjects.longValue());
        output.writeLong(r.usedObjectSpace.longValue());
        output.writeLong(r.usedTotalSpace.longValue());
      }
      else if ( record instanceof GC_START )
      {
        GC_START r = (GC_START) record;

        output.writeByte(0x0A); // ADAPT_GC_START

        output.writeInt(r.envId.intValue());
      }
      else if ( record instanceof JVM_INIT_DONE )
      {
        JVM_INIT_DONE r = (JVM_INIT_DONE) record;

        output.writeByte(0x10); // ADAPT_JVM_INIT_DONE

        output.writeInt(r.envId.intValue());
      }
      else if ( record instanceof JVM_SHUT_DOWN )
      {
        JVM_SHUT_DOWN r = (JVM_SHUT_DOWN) record;

        output.writeByte(0x11); // ADAPT_JVM_SHUT_DOWN

        output.writeInt(r.envId.intValue());
      }
      else*/ if ( record instanceof OBJECT_ALLOC )
      {
        OBJECT_ALLOC r = (OBJECT_ALLOC) record;

        _allocBuilder.newEvent();

        assert _envIdDef != null;
        assert _envIdToEntity.get(r.envId) != null;
        _allocBuilder.addReferenceField(
          _envIdDef, (Entity) _envIdToEntity.get(r.envId));

        _allocBuilder.addReferenceField(
          _arenaIdDef, (Entity) _arenaIdToEntity.get(r.arenaId));

        _allocBuilder.addReferenceField(
          _classIdDef, (Entity) _classIdToEntity.get(r.classId));

        // isArray ?

        _allocBuilder.addValueField(_sizeDef, r.size.intValue());

        _allocBuilder.addValueField(_objIdDef, r.objId.intValue());

        return _allocBuilder.buildEvent();
      }
/*
      else if ( record instanceof OBJECT_FREE )
      {
        OBJECT_FREE r = (OBJECT_FREE) record;

        output.writeByte(0x1D); // ADAPT_OBJECT_FREE

        output.writeInt(r.envId.intValue());
        output.writeInt(r.objId.intValue());
      }
      else if ( record instanceof METHOD_ENTRY2 )
      {
        METHOD_ENTRY2 r = (METHOD_ENTRY2) record;

        output.writeByte(0x13); // ADAPT_METHOD_ENTRY2

        output.writeInt(r.envId.intValue());
        output.writeInt(r.methodId.intValue());
        output.writeInt(r.objId.intValue());
      }
*/
/*
      else if ( record instanceof METHOD_EXIT )
      {
        METHOD_EXIT r = (METHOD_EXIT) record;

        output.writeByte(0x14); // ADAPT_METHOD_EXIT

        output.writeInt(r.envId.intValue());
        output.writeInt(r.methodId.intValue());
      }
      else if ( record instanceof INSTRUCTION_START )
      {
        INSTRUCTION_START r = (INSTRUCTION_START) record;

        output.writeByte(0x24); // ADAPT_INSTRUCTION_START

        output.writeInt(r.envId.intValue());
        output.writeInt(r.methodId.intValue());
        output.writeInt(r.fieldOffset.intValue());
        output.writeInt(r.fieldIsTrue.intValue());
        output.writeInt(r.fieldKey.intValue());
        output.writeInt(r.fieldLow.intValue());
        output.writeInt(r.fieldHigh.intValue());
        output.writeInt(r.pairIndex.intValue());
        output.writeInt(r.pairTotal.intValue());
      }
*/
      else if ( record == null ) // end of input
      {
        return null;
      }
      else
        throw new java.io.IOException("unknown event: "+record);
    }
    catch (Exception e)
    {
      System.out.println("getNextEvent(): error");
      e.printStackTrace();
      throw new DataProcessingException(e.toString());
    }
  }

  public static void main(String[] args)
  {
    DataSource dataSource = new EVolveDemo();

    VisualizationFactory[] factory = new VisualizationFactory[7];
    factory[0] = new TableVizFactory();
    factory[1] = new HotSpotAmountFactory();
    factory[2] = new HotSpotCoordinateFactory();
    factory[3] = new BarChartVerticalFactory();
    factory[4] = new BarChartHorizontalFactory();
    factory[5] = new CorrelationVizFactory();
    factory[6] = new PredictionVizFactory();
    ((PredictionVizFactory)(factory[6])).addPredictorFactory(new DefaultPredictorFactory());

    EVolve.start(dataSource, factory);
  }
}
