package example;

// Java imports
import java.io.BufferedOutputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Hashtable;

// STEP imports
import step.*;
import step.io.*;

// Events
import example.soot.*;

public class ProfileCollector
{
  public static final int BUF_SIZE = 65536;
  public static final String DEFAULT_FILENAME = "profile.step";

  public static final int MAX_EVENTS;

  private static ProfileCollector _theInstance = new ProfileCollector();

  static 
  {
    MAX_EVENTS = Integer.parseInt(System.getProperty("profile.maxevents", "100000000"));
  }

  private StepRecordOutput _stepOut;

  private Hashtable _threads = new Hashtable();
  private Hashtable _types   = new Hashtable();
  private Hashtable _methods = new Hashtable();
  private Hashtable _fields  = new Hashtable();
  private Hashtable _sites   = new Hashtable();

  private int _eventCount = 0;

  private ProfileCollector()
  {
    Runtime.getRuntime().addShutdownHook(new Thread() {
      public void run()
      {
        try
        {
          _stepOut.close();
        }
        catch (Exception e) {}

        System.err.println("STEP: recorded "+_eventCount+" events");
      }
    });

    try
    {
      _stepOut = new StepEncodedOutput(
        new BufferedOutputStream(new FileOutputStream(fileName()), BUF_SIZE));
    }
    catch (Exception e) { throw new RuntimeException(e); }
  }

  private void count()
  {
    _eventCount++;

    if ( _eventCount >= MAX_EVENTS )
    {
      System.err.println("STEP: reached max event count, terminating");
      System.exit(0);
    }
  }

  public static ProfileCollector v() { return _theInstance; }

  protected example.java.Thread getThread()
  {
    java.lang.Thread javaThread = Thread.currentThread();
    example.java.Thread thread  = (example.java.Thread) _threads.get(javaThread);

    if ( thread == null )
    {
      thread = new example.java.Thread(
        new StepString(javaThread.getName()),
        new StepString(javaThread.getThreadGroup().getName()));

      _threads.put(javaThread, thread);
    }

    return thread;
  }

  protected example.java.Type getType(String typeName)
  {
    example.java.Type type = (example.java.Type) _types.get(typeName);

    if ( type == null )
    {
      type = new example.java.Type(new StepString(typeName));
      _types.put(typeName, type);
    }

    return type;
  }

  protected example.java.Method getMethod(String signature)
  {
    example.java.Method method = (example.java.Method) _methods.get(signature);

    if ( method == null )
    {
      // "<class: method(args)>"
      String[] parts = signature.substring(1, signature.length()-1).split(":");

      String typeName   = parts[0].trim();
      String methodName = parts[1].trim();

      method = new example.java.Method(new StepString(methodName), getType(typeName));
      _methods.put(signature, method);
    }

    return method;
  }

  protected example.java.Field getField(
    String fieldName, String type, String declaringClass, boolean isStatic)
  {
    String key = declaringClass+fieldName;
    example.java.Field field = (example.java.Field) _fields.get(key);

    if ( field == null )
    {
      field = new example.java.Field(
        new StepString(fieldName),
        getType(type),
        getType(declaringClass),
        new StepInt( (isStatic) ? 1 : 0 ));

      _fields.put(key, field);
    }

    return field;
  }

  protected Site getSite(String method, int siteNum)
  {
    String key = method+siteNum;
    Site site = (Site) _sites.get(key);

    if ( site == null )
    {
      site = new Site(getMethod(method), new StepInt(siteNum));
      _sites.put(key, site);
    }

    return site;
  }

  public synchronized void allocation(String type, String method, int siteNum)
  {
    try
    {
      _stepOut.writeRecord(
        new Allocation(getThread(), getType(type), getSite(method, siteNum)));
    }
    catch (Exception e) { throw new RuntimeException(e); }

    count();
  }

  public synchronized void methodEnter(String method)
  {
    try
    {
      _stepOut.writeRecord(new MethodEnter(getThread(), getMethod(method)));
    }
    catch (Exception e) { throw new RuntimeException(e); }

    count();
  }

  public synchronized void methodExit(String method, int siteNum)
  {
    try
    {
      _stepOut.writeRecord(
        new MethodExit(getThread(), getMethod(method), getSite(method, siteNum)));
    }
    catch (Exception e) { throw new RuntimeException(e); }

    count();
  }

  public synchronized void fieldAccess(
    String field, String type, String declaringType, boolean isStatic,
    boolean read, String method, int siteNum)
  {
    try
    {
      StepRecord accessRecord = (read) ?
        (StepRecord) new FieldRead(getThread(),
          getField(field, type, declaringType, isStatic), getSite(method, siteNum)) :
        (StepRecord) new FieldWrite(getThread(),
          getField(field, type, declaringType, isStatic), getSite(method, siteNum)) ;

      _stepOut.writeRecord(accessRecord);
    }
    catch (Exception e) { throw new RuntimeException(e); }

    count();
  }

  static String fileName()
  {
    return System.getProperty("profile.file", DEFAULT_FILENAME);
  }
}
