/* STEP - An Framework for the Efficient Encoding of Trace Data
 * Copyright (C) 2002 Rhodes Brown
 *
 * STEP is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * STEP is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with STEP; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA
 */

/* STEP is distributed through the Sable Research Group at:
 * http://www.sable.mcgill.ca/step/
 */

/* Reference Version: $ProjectVersion: 0.9.2.dev.36 $ */

package step;

import java.util.Iterator;
import java.util.Collection;

import step.typedef.*;

/**
 * Implementation for arrays of STEP objects.
 */
public class StepArray implements StepObject
{
  /** The prototype definition for the length field of array objects. */
  public static final FieldDef LENGTH_FIELD;

  /** The prototype definition for the element fields of array objects. */
  public static final FieldDef ELEMENT_FIELD;

  static
  {
    LENGTH_FIELD = FieldDef.builder().newFieldDef("length", StepInt.TYPE)
      .addAttribute(new Attribute("property", "unsigned"))
      .addAttribute(new Attribute("encoding", "size=creep"))
      .makeFieldDef();

    ELEMENT_FIELD = FieldDef.builder().newFieldDef("element", StepObject.TYPE)
      .makeFieldDef();
  }

  private StepObject[] _elements;

  /**
   * Create a new STEP array.
   * <p>
   * NOTE: The element values are assumed to be non-null.
   */
  public StepArray(StepObject[] elements)
  {
    _elements = elements; assert _elements != null;
  }

  /**
   * Create a new STEP array from a collection of objects.
   * <p>
   * NOTE: The given collection is assumed to contain objects of the same type.
   *
   * @throws ClassCastException if an element is not a {@link StepObject}.
   */
  public StepArray(Collection elements)
  {
    _elements = new StepObject[elements.size()];

    Iterator iElements = elements.iterator();
    for (int i=0; iElements.hasNext(); i++)
    {
      _elements[i] = (StepObject) iElements.next(); assert _elements[i] != null;
    }
  }

  /** Get the underlying array of elements. */
  public StepObject[] getElements()
  {
    return _elements;
  }

  /** Get the length of this array. */
  public int length()
  {
    return _elements.length;
  }

  /** Get an iterator over the elements of this array. */
  public Iterator iterator()
  {
    return new FieldIterator(_elements);
  }

  /**
   * Compare two array objects for equality.
   * <p>
   * The objects are equivalent if they are the same size and contain the same
   * data values.
   * <p>
   * The algorithm is fail-fast.
   */
  public boolean equals(Object o) //: StepObject
  {
    try
    {
      StepObject[] rhs = ((StepArray) o)._elements;

      if ( _elements.length != rhs.length )
        return false;

      // NOTE: We do an explicit length check (vs. catching an exception) since
      //       it is a common source of difference.

      for (int i=0; i<_elements.length; i++)
      {
        if ( ! _elements[i].equals(rhs[i]) )
          return false;
      }

      return true;
    }
    catch (ClassCastException e)
    {
      return false;
    }
    catch (NullPointerException e)
    {
      return false;
    }
  }

  /**
   * Get the hash-code for this object.
   *
   * <h3>Algorithm:</h3>
   * <pre>
   * for (i=0; i&lt;value.length; i++)
   *   code = (code &lt;&lt; 1) ^ (hashCode[i]);
   * </pre>
   */
  public int hashCode() //: StepObject
  {
    int code = 0;

    for (int i=0; i<_elements.length; i++)
      code = (code << 1) ^ _elements[i].hashCode();

    return code;
  }

  /**
   * Get a string representation of this array.
   * The elements are assumed to implement a valid version of
   * <code>toString()</code>.
   */
  public String toString()
  {
    StringBuffer buf = new StringBuffer();

    buf.append('[');
    for (int i=0; i<_elements.length; i++)
    {
      if ( i > 0 ) buf.append(',');
      buf.append(_elements[i]);
    }
    buf.append(']');

    return buf.toString();
  }
}
