/* STEP - An Framework for the Efficient Encoding of Trace Data
 * Copyright (C) 2002 Rhodes Brown
 *
 * STEP is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * STEP is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with STEP; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA
 */

/* STEP is distributed through the Sable Research Group at:
 * http://www.sable.mcgill.ca/step/
 */

/* Reference Version: $ProjectVersion: 0.9.2.dev.36 $ */

package step;

import java.util.List;

import step.typedef.*;

/**
 * Implementation of the built-in STEP <code>data</code> type.
 */
public class StepData implements StepObject
{
  /** The STEP type for <code>data</code> objects. */
  public static final Type TYPE;

  public static final FieldDef LENGTH_FIELD;

  static
  {
    final String name = "StepData";
    final step.typedef.Package pkg = step.typedef.Package.v("step");

    TYPE = Type.v(pkg.qualify(name));
    TYPE.setParent(StepObject.TYPE);
    TYPE.setStepDL("data");

    LENGTH_FIELD = FieldDef.builder().newFieldDef("length", StepInt.TYPE)
      .addAttribute(new Attribute("property", "unsigned"))
      .addAttribute(new Attribute("encoding", "size=creep"))
      .makeFieldDef();

    List fieldDefs = new java.util.ArrayList(1);
    fieldDefs.add(LENGTH_FIELD);

    new TypeDef(name, pkg, fieldDefs, null)
      .setAsDefault();
  }

  private byte[] _data;

  /** Create a new <code>data</code> object from an array of bytes. */
  public StepData(byte[] data)
  {
    _data = data; assert _data != null;
  }

  /** Get the underlying byte data for this object. */
  public byte[] getData()
  {
    return _data;
  }

  /**
   * Compare two <code>data</code> objects for equality.
   * <p>
   * The objects are equivalent if their data are the same size and contain the
   * same values.
   * <p>
   * The algorithm is fail-fast.
   */
  public boolean equals(Object o) //: StepObject
  {
    try
    {
      byte[] rhs = ((StepData) o)._data;

      if ( _data.length != rhs.length )
        return false;

      // NOTE: We do an explicit length check (vs. catching an exception) since
      //       it is a common source of difference.

      for (int i=0; i<_data.length; i++)
      {
        if ( _data[i] != rhs[i] )
          return false;
      }

      return true;
    }
    catch (ClassCastException e)
    {
      return false;
    }
    catch (NullPointerException e)
    {
      return false;
    }
  }

  /**
   * Get the hash-code for this object.
   *
   * <h3>Algorithm:</h3>
   * <pre>
   * for (i=0; i&lt;value.length; i++)
   *   code = (code &lt;&lt; 1) ^ (hashCode[i]);
   * </pre>
   *
   * @see #equals(Object)
   */
  public int hashCode() //: StepObject
  {
    int code = 0;

    for (int i=0; i<_data.length; i++)
      code = (code << 1) ^ (int) _data[i];

    return code;
  }

  /**
   * Get a text version of this data object.
   * <p>
   * The output is a list of byte values in hexadecimal.
   *
   * @see step.util.Hex#toString(byte[])
   */
  public String toString()
  {
    return step.util.Hex.toString(_data);
  }
}
