/* STEP - An Framework for the Efficient Encoding of Trace Data
 * Copyright (C) 2002 Rhodes Brown
 *
 * STEP is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * STEP is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with STEP; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA
 */

/* STEP is distributed through the Sable Research Group at:
 * http://www.sable.mcgill.ca/step/
 */

/* Reference Version: $ProjectVersion: 0.9.2.dev.36 $ */

package step;

import step.typedef.*;

/**
 * Implementation of the built-in STEP <code>int</code> type.
 */
public class StepInt extends Number implements StepObject
{
  /** The STEP type for <code>int</code> objects. */
  public static final Type TYPE;

  static
  {
    final String name = "StepInt";
    final step.typedef.Package pkg = step.typedef.Package.v("step");

    TYPE = Type.v(pkg.qualify(name));
    TYPE.setParent(StepObject.TYPE);
    TYPE.setStepDL("int");

    new TypeDef(name, pkg, null, null)
      .setAsDefault();
  }
 
  private long _value;

  /** Create a STEP int from a <code>byte</code> value. */
  public StepInt(byte byteValue)
  {
    _value = (long) byteValue;
  }

  /** Create a STEP int from a <code>short</code> value. */
  public StepInt(short shortValue)
  {
    _value = (long) shortValue;
  }

  /** Create a STEP int from a (Java) <code>int</code> value. */
  public StepInt(int intValue)
  {
    _value = (long) intValue;
  }

  /** Create a STEP int from a <code>long</code> value. */
  public StepInt(long longValue)
  {
    _value = longValue;
  }

  /** Set the value for this STEP int from a <code>byte</code> value. */
  public void setValue(byte byteValue)
  {
    _value = (long) byteValue;
  }

  /** Set the value for this STEP int from a <code>short</code> value. */
  public void setValue(short shortValue)
  {
    _value = (long) shortValue;
  }

  /** Set the value for this STEP int from a (Java) <code>int</code> value. */
  public void setValue(int intValue)
  {
    _value = (long) intValue;
  }

  /** Set the value for this STEP int from a <code>long</code> value. */
  public void setValue(long longValue)
  {
    _value = longValue;
  }

  /**
   * Convert this object to a <code>byte</code> value.
   * <p>
   * NOTE: Values outside the [-128..127] range are truncated.
   */
  public byte byteValue() //: Number
  {
    return (byte) _value;
  }

  /** Convert this object to a <code>double</code> value. */
  public double doubleValue() //: Number
  {
    return new Long(_value).doubleValue();
  }

  /** Convert this object to a <code>float</code> value. */
  public float floatValue() //: Number
  {
    return new Long(_value).floatValue();
  }

  /**
   * Convert this object to a (Java) <code>int</code> value.
   * <p>
   * NOTE: Values outside the [-2147483648..2147483647] range are truncated.
   */
  public int intValue() //: Number
  {
    return (int) _value;
  }

  /** Convert this object to a <code>long</code> value. */
  public long longValue() //: Number
  {
    return _value;
  }

  /**
   * Convert this object to a <code>short</code> value.
   * <p>
   * NOTE: Values outside the [-32768..32767] range are truncated.
   */
  public short shortValue() //: Number
  {
    return (short) _value;
  }

  /**
   * Add two STEP int values.
   * @throws NullPointerException if the argument is <code>null</code>.
   */
  public StepInt add(StepInt rhs)
  {
    return new StepInt(_value + rhs._value);
  }

  /**
   * Subtract a STEP int value from this value.
   * @throws NullPointerException if the argument is <code>null</code>.
   */
  public StepInt sub(StepInt rhs)
  {
    return new StepInt(_value - rhs._value);
  }

  /**
   * Multiply two STEP int values.
   * @throws NullPointerException if the argument is <code>null</code>.
   */
  public StepInt mul(StepInt rhs)
  {
    return new StepInt(_value * rhs._value);
  }

  /**
   * Divide this value by another STEP int value.
   * @throws NullPointerException if the argument is <code>null</code>.
   */
  public StepInt div(StepInt rhs)
  {
    return new StepInt(_value / rhs._value);
  }

  /**
   * Get the division remainder of this value divided by another STEP int value.
   * @throws NullPointerException if the argument is <code>null</code>.
   */
  public StepInt mod(StepInt rhs)
  {
    return new StepInt(_value % rhs._value);
  }

  /** Compare two STEP int values for equality. */
  public boolean equals(Object o) //: StepObject
  {
    try
    {
      return _value == ((StepInt) o)._value;
    }
    catch (ClassCastException e)
    {
      return false;
    }
    catch (NullPointerException e)
    {
      return false;
    }
  }

  /** Get a hash code for this value. */
  public int hashCode() //: StepObject
  {
    return ((int) _value) ^ ((int) (_value << 32));
  }

  /** Get a text version of this value. */
  public String toString()
  {
    return Long.toString(_value);
  }
}
