/* STEP - An Framework for the Efficient Encoding of Trace Data
 * Copyright (C) 2002 Rhodes Brown
 *
 * STEP is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * STEP is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with STEP; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA
 */

/* STEP is distributed through the Sable Research Group at:
 * http://www.sable.mcgill.ca/step/
 */

/* Reference Version: $ProjectVersion: 0.9.2.dev.36 $ */

package step.encoding;

import java.util.Iterator;
import java.util.List;

import step.*;
import step.typedef.*;

/**
 * An encoder for STEP array objects.
 *
 * @see StepArray
 */
public class ArrayEncoder extends Encoder
{
  private Encoder _lengthEncoder;
  private Encoder _elementEncoder;

  private final StepInt _length = new StepInt(0);
    // use a tmp field to avoid unnecessary allocation

  /**
   * Create a new encoder for STEP array objects.
   *
   * @param name        the name of the encoder.
   * @param lengthDef   a definition for the length field
   * @param elementDef  a definition for the element field
   * @param factory     the {@link EncoderFactory} used to create the encoder
   *
   * @throws IllegalArgumentException if the length definition does not handle
   *   {@link StepInt} values.
   */
  public ArrayEncoder(String name, FieldDef lengthDef, FieldDef elementDef,
                      EncoderFactory factory)
  {
    super(name, factory);

    if ( !lengthDef.getType().equals(StepInt.TYPE) )
      throw new IllegalArgumentException("length field must encode StepInt values");

    // .length field
    {
      String lengthName = makeFieldName(lengthDef.getName());

      TypeDef lengthTypeDef = lengthDef.hasAltTypeDef() ?
        lengthDef.getAltTypeDef() :
        TypeDef.getDefault(lengthDef.getType());

      List lengthAttrs = lengthDef.getAttributes();

      _lengthEncoder = getFactory().newEncoder(lengthName, lengthTypeDef, lengthAttrs);
    }

    // .element field
    {
      String elementName = makeFieldName(elementDef.getName());

      TypeDef elementTypeDef = elementDef.hasAltTypeDef() ?
        elementDef.getAltTypeDef() :
        TypeDef.getDefault(elementDef.getType());

      List elementAttrs = elementDef.getAttributes();

      _elementEncoder = getFactory().newEncoder(elementName, elementTypeDef, elementAttrs);
    }
  }

  /**
   * Encode a STEP array object.
   *
   * @throws EncoderException if the object is not a {@link StepArray}.
   */
  public void encode(StepObject o, EncodeContext context)
    throws EncoderException, java.io.IOException
  {
    StepArray a;
    try
    {
      a = (StepArray) o;
    }
    catch (ClassCastException e)
    {
      throw new EncoderException(this, "incorrect value type");
    }

    _length.setValue(a.length());
    _lengthEncoder.encode(_length, context);

    Iterator iElements = a.iterator();
    while (iElements.hasNext())
      _elementEncoder.encode((StepObject) iElements.next(), context);
  }

  /**
   * Decode a STEP array object.
   */
  public StepObject decode(DecodeContext context)
    throws EncoderException, java.io.IOException
  {
    StepInt length = (StepInt) _lengthEncoder.decode(context);

    StepObject[] elements = new StepObject[length.intValue()];
    for (int i=0; i<elements.length; i++)
      elements[i] = _elementEncoder.decode(context);

    return new StepArray(elements);
  }

  public void dump(int level)
  {
    indent(level); System.out.println(getName());

    _lengthEncoder.dump(level+1);
    _elementEncoder.dump(level+1);
  }
}
