/* STEP - An Framework for the Efficient Encoding of Trace Data
 * Copyright (C) 2002 Rhodes Brown
 *
 * STEP is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * STEP is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with STEP; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA
 */

/* STEP is distributed through the Sable Research Group at:
 * http://www.sable.mcgill.ca/step/
 */

/* Reference Version: $ProjectVersion: 0.9.2.dev.36 $ */

package step.encoding;

import java.util.Map;
import java.util.List;

import step.*;
import step.typedef.*;

/**
 * Use a simple caching strategy to replace values with a reference to a cache
 * slot. When a miss occurs, the strategy just replaces the next slot (on a
 * rotational basis) with the new value.
 * <p>
 * It is possible to chain cache strategies to get a multi-level exclusive cache
 * hierarchy.
 * <p>
 * NOTE: If the cache has a lot of misses, it will likely have a negative effect
 * on overall reduction; introducing lots of meta-data and artificial randomness
 * (as values get mapped to different slots).
 */
public class CacheStrategy extends RegularValueStrategy
{
  public static final String STRATEGY_NAME = "cache";

  protected StepObject[] _cache;
  protected StepInt[]    _slots;
  protected Map          _valueToSlot;

  private int _nextSlot = 0;

  private Encoder _slotEncoder;

  /**
   * Create a new "cache" strategy.
   *
   * @param baseEncoder the encoder to use for cache misses.
   * @param cacheSize   the number of elements to keep in the cache.
   */
  public CacheStrategy(Encoder baseEncoder, int cacheSize)
  {
    super(baseEncoder, STRATEGY_NAME);

    // step the cache slot identifiers (ints: 0..cacheSize-1)
    {
      _slots = new StepInt[cacheSize];
      for (int i=0; i<cacheSize; i++)
        _slots[i] = new StepInt(i);
    }

    // create an empty cache
    _cache = new StepObject[cacheSize];

    // how to find slot values:
    _valueToSlot = new java.util.HashMap(2*cacheSize);

    // create the encoder for slot ID values
    {
      List slotEncAttrs = new java.util.ArrayList(2);
      slotEncAttrs.add(new Attribute("encoding", "unsigned"));
      if ( cacheSize <= 256 ) // 2^8 => 1 byte
        slotEncAttrs.add(new Attribute("encoding", "size=1"));
      else if ( cacheSize <= 65536 ) // 2^16 => 2 bytes
        slotEncAttrs.add(new Attribute("encoding", "size=2"));
      else if ( cacheSize <= 16777216 ) // 2^24 => 3 bytes
        slotEncAttrs.add(new Attribute("encoding", "size=3.."));
      else // ...a very big cache!
        slotEncAttrs.add(new Attribute("encoding", "size=4.."));

      _slotEncoder =
        getFactory().newEncoder(makeFieldName("slot"), StepInt.TYPE, slotEncAttrs);
    }

    setNextIsRegular(false);
  }

  /**
   * Encode a value using its cache slot ID.
   */
  public void encode(StepObject o, EncodeContext context)
    throws EncoderException, java.io.IOException
  {
    StepInt slot = (StepInt) _valueToSlot.get(o);

    if ( slot == null )
    {
      slot = nextSlot();

      _valueToSlot.remove(_cache[slot.intValue()]);
      _valueToSlot.put(o, slot);
      _cache[slot.intValue()] = o;

      context.metaWrite(new IrregularValueEvent(new StepString(getName())));
      baseEncode(o, context);
    }
    else
    {
      _slotEncoder.encode(slot, context);
    }
  }

  /**
   * Decode a value using its cache slot ID.
   */
  public StepObject decode(DecodeContext context)
    throws EncoderException, java.io.IOException
  {
    if ( _nextIsRegular )
    {
      StepInt slot = (StepInt) _slotEncoder.decode(context);

      assert _cache[slot.intValue()] != null;

      return _cache[slot.intValue()];
    }
    else // => new mapping
    {
      setNextIsRegular(true); // IMPORTANT: set this before decoding

      return ( _cache[nextSlot().intValue()] = baseDecode(context) );
    }
  }

  /**
   * Rotate to the next cache slot.
   */
  protected StepInt nextSlot()
  {
    StepInt slot = _slots[_nextSlot];
    if ( ++_nextSlot >= _slots.length ) _nextSlot = 0;
    return slot;
  }

  public void dump(int level)
  {
    indent(level);   System.out.println(getName());
    indent(level+1); System.out.println("size="+_cache.length);
    indent(level+1); System.out.println("cache="+_valueToSlot);

    _slotEncoder.dump(level+1);
    _baseEncoder.dump(level+1);
  }
}
