/* STEP - An Framework for the Efficient Encoding of Trace Data
 * Copyright (C) 2002 Rhodes Brown
 *
 * STEP is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * STEP is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with STEP; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA
 */

/* STEP is distributed through the Sable Research Group at:
 * http://www.sable.mcgill.ca/step/
 */

/* Reference Version: $ProjectVersion: 0.9.2.dev.36 $ */

package step.encoding;

import step.*;

/**
 * Assume values are mostly the same.
 * If the value remains constant, there is no need to write anything to the
 * output. The decoder simply returns the default value instead of reading from
 * the input.
 * <p>
 * The "constant" variant of the algorithm throws an exception if an irregular
 * value is encountered.
 *
 * The "default" variant allows deviations from the initial value, signalling
 * the irregularity with an {@link IrregularValueEvent}.
 *
 * The "repeat" variant shifts to a new default every time the value differs
 * from the previous, signalling the change with an {@link IrregularValueEvent}.
 *
 * <h3>Algorithm:</h3>
 * <pre>
 * if current == default
 *     then -&gt; <i>nothing</i>
 *     else if constant
 *         then ERROR!
 *         else -&gt; meta:{@link IrregularValueEvent}, current
 * </pre>
 */
public class ConstantStrategy extends RegularValueStrategy
{
  public static final String CONSTANT_STRATEGY_NAME = "constant";
  public static final String DEFAULT_STRATEGY_NAME  = "default";
  public static final String REPEAT_STRATEGY_NAME   = "repeat";

  public static final int CONSTANT_STRATEGY = 1;
  public static final int DEFAULT_STRATEGY  = 2;
  public static final int REPEAT_STRATEGY   = 3;

  private StepObject _default;
  private int        _strategy;

  /**
   * Create a new constant value strategy.
   *
   * @param baseEncoder the encoder for the values passed to this strategy.
   * @param strategy one of
   * <ul>
   *   <li> {@link #CONSTANT_STRATEGY} - <em>all</em> values are the same;
   *     deviations should generate an error.
   *   <li> {@link #DEFAULT_STRATEGY} - values are rarely different from the
   *     initial value; deviants are indicated with a meta event.
   *   <li> {@link #REPEAT_STRATEGY} - values consist of long sequences of the
   *     same value; new default values are indicated with a meta event.
   * </ul>
   *
   * @throws IllegalArgumentException if <code>strategy</code> is invalid.
   */
  public ConstantStrategy(Encoder baseEncoder, int strategy)
  {
    super(baseEncoder, strategyToName(strategy));

    _strategy = strategy;

    setNextIsRegular(false); // => initial read is default value
  }

  private static String strategyToName(int strategy)
  {
    switch (strategy)
    {
      case CONSTANT_STRATEGY: return CONSTANT_STRATEGY_NAME;
      case DEFAULT_STRATEGY:  return DEFAULT_STRATEGY_NAME;
      case REPEAT_STRATEGY:   return REPEAT_STRATEGY_NAME;
      default:
        throw new IllegalArgumentException("unknown strategy type: "+strategy);
    }
  }

  protected int getStrategyVariant()
  {
    return _strategy;
  }

  /**
   * Encode a constant value.
   *
   * @throws EncoderException if the value is different from the default
   *   (initial) value and the "constant" option is set.
   */
  public void encode(StepObject o, EncodeContext context)
    throws EncoderException, java.io.IOException
  {
    try
    {
      if ( _default.equals(o) ) // => encode nothing
      {
        return;
      }
    }
    catch (NullPointerException e)
    {
      assert _default == null; // => initial write

      baseEncode(_default = o, context);
      return;
    }
    // => different from default

    if ( _strategy == CONSTANT_STRATEGY )
    {
      throw new EncoderException(this, "value differs from constant");
    }
    else
    {
      if ( _strategy == REPEAT_STRATEGY ) // => change default
      {
        _default = o;
      }
      else
        assert _strategy == DEFAULT_STRATEGY;

      context.metaWrite(new IrregularValueEvent(new StepString(getName())));
      baseEncode(o, context);
    }
  }

  /**
   * Decode a constant value.
   */
  public StepObject decode(DecodeContext context)
    throws EncoderException, java.io.IOException
  {
    if ( _nextIsRegular )
    {
      assert _default != null;

      return _default;
    }
    else // => deviant value
    {
      setNextIsRegular(true); // IMPORTANT: set this before decoding

      StepObject current = baseDecode(context);

      if ( (_strategy == REPEAT_STRATEGY) || (_default == null) )
      {
        _default = current;
      }

      return current;
    }
  }

  public void dump(int level)
  {
    indent(level);   System.out.println(getName());
    indent(level+1); System.out.println("value="+_default);

    _baseEncoder.dump(level+1);
  }
}
