/* STEP - An Framework for the Efficient Encoding of Trace Data
 * Copyright (C) 2002 Rhodes Brown
 *
 * STEP is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * STEP is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with STEP; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA
 */

/* STEP is distributed through the Sable Research Group at:
 * http://www.sable.mcgill.ca/step/
 */

/* Reference Version: $ProjectVersion: 0.9.2.dev.36 $ */

package step.encoding;

import step.StepObject;
import step.StepInt;
import step.StepString;

import step.util.*;

/**
 * Encoder for STEP <code>int</code> types that uses a variable size encoding.
 * <p>
 * Values are encoded one byte at a time, using 7 bits for data and marking the
 * high bit if more bytes should be read. The encoding is thus pseudo-LITTLE
 * endian (least significant -&gt; most significant), using 7 bit bytes.
 * <p>
 * The maximum number of bytes used by this encoding is 10, in signed mode:
 * 9 (=63/7) for the 63 low bits plus one more to sign extend.
 *
 * @see StepInt
 * @see IntEncoder
 */
public class CreepIntEncoder extends Encoder
{
  private byte[]  _buf = new byte[10];
  private boolean _unsigned;

  /**
   * Create a new encoder for STEP <code>int</code> types.
   *
   * @param name      the name for this encoder.
   * @param unsigned  <code>true</code> if encoded/decoded values are always &gt;= 0
   * @param factory   the {@link EncoderFactory} used to create this encoder
   */
  public CreepIntEncoder(String name, boolean unsigned, EncoderFactory factory)
  {
    super(name, factory);

    _unsigned = unsigned;
  }

  /**
   * Encode the given STEP <code>int</code> value.
   *
   * @throws EncoderException if
   * <ul>
   *   <li> the value was not a {@link StepInt}.
   *   <li> a negative value was passed while the "unsigned" option was set.
   * </ul>
   */
  public void encode(StepObject o, EncodeContext context)
    throws EncoderException, java.io.IOException
  {
    long n;
    try
    {
      n = ((StepInt) o).longValue();
    }
    catch (ClassCastException e)
    {
      throw new EncoderException(this, "incorrect value type");
    }

    int used = 0;

    if ( n < 0L )
    {
      if ( _unsigned )
        throw new EncoderException(this, "attempt to encode negative value as unsigned");

      long n_comp = n ^ 0xFFFFFFFFFFFFFFFFL; // invert the bits
      do
      {
        _buf[used] = (byte) (n & 0x7F); // grab the low 7 bits
        n >>= 7; n_comp >>= 7;          // shift right 7 (not 8!)

        if ( n_comp != 0L ) // => more bits to come
        {
          _buf[used] |= (byte) (0x80); // mark the high bit
        }

        used++;
      }
      while (n_comp != 0L);

      // sign extend
      {
        _buf[used-1] |= (byte) (0x80); // mark the high bit
        _buf[used++] = 0x7F;           // add a dummy byte
      }
    }
    else // n >= 0
    {
      do
      {
        _buf[used] = (byte) (n & 0x7F); // grab the low 7 bits
        n >>= 7;                        // shift right 7 (not 8!)

        if ( n != 0L ) // => more bits to come
        {
          _buf[used] |= (byte) (0x80); // mark the high bit
        }

        used++;
      }
      while (n != 0L);

      if ( (!_unsigned) && ((_buf[used-1] & 0x40) != 0) ) // => sign extend
      {
        _buf[used-1] |= (byte) (0x80); // mark the high bit
        _buf[used++] = 0x00;           // add a dummy byte
      }
    }

    byte[] bytes = ByteArray.allocator().alloc(used);
    {
      for (int i=0; i<used; i++) bytes[i] = _buf[i];
      context.write(bytes);
    }
    ByteArray.allocator().free(bytes);
  }

  /**
   * Decode a STEP <code>int</code> value.
   */
  public StepObject decode(DecodeContext context)
    throws EncoderException, java.io.IOException
  {
    long n = 0L;
    byte[] aByte = ByteArray.allocator().alloc(1);
    {
      int i = 0;
      do
      {
        context.read(aByte);

        byte bits = (byte) (aByte[0] & 0x7F); // grab the low 7 bits

        n |= ( (((long) bits) << i) & (0x7FL << i) );

        i += 7;
      }
      while ( (aByte[0] & 0x80) != 0 );

      if ( (!_unsigned) && ((aByte[0] & 0x40) != 0) ) // negative => sign extend
      {
        for ( ; i < 63; i += 7 ) n |= (0x7FL << i);
        n |= (1L << 63);
      }
    }
    ByteArray.allocator().free(aByte);

    return new StepInt(n);
  }

  public void dump(int level)
  {
    indent(level);   System.out.println(getName()+" (int:creep)");
    indent(level+1); System.out.println("unsigned="+_unsigned);
  }
}
