/* STEP - An Framework for the Efficient Encoding of Trace Data
 * Copyright (C) 2002 Rhodes Brown
 *
 * STEP is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * STEP is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with STEP; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA
 */

/* STEP is distributed through the Sable Research Group at:
 * http://www.sable.mcgill.ca/step/
 */

/* Reference Version: $ProjectVersion: 0.9.2.dev.36 $ */

package step.encoding;

import java.util.List;

import step.*;
import step.typedef.*;
import step.util.ByteArray;

/**
 * An encoder for STEP <code>data</code> objects.
 *
 * @see StepData
 */
public class DataEncoder extends Encoder
{
  private Encoder _lengthEncoder;

  private final StepInt _length = new StepInt(0);
    // avoid unnecessary allocation

  /**
   * Create a new encoder for STEP <code>data</code> objects.
   *
   * @param name       the name for this encoder.
   * @param lengthDef  a definition for the length field of data objects.
   * @param factory    the {@link EncoderFactory} used to create this encoder
   *
   * @throws IllegalArgumentException if the length field does not handle
   *   {@link StepInt} values.
   */
  public DataEncoder(String name, FieldDef lengthDef, EncoderFactory factory)
  {
    super(name, factory);

    if ( ! lengthDef.getType().equals(StepInt.TYPE) )
      throw new IllegalArgumentException("length field must encode StepInt values");

    String lengthName = makeFieldName(lengthDef.getName());

    TypeDef lengthTypeDef = lengthDef.hasAltTypeDef() ?
      lengthDef.getAltTypeDef() :
      TypeDef.getDefault(lengthDef.getType());

    List lengthAttrs = lengthDef.getAttributes();

    _lengthEncoder = getFactory().newEncoder(lengthName, lengthTypeDef, lengthAttrs);
  }

  /**
   * Encode a STEP <code>data</code> object.
   *
   * @throws EncoderException if the object was not a {@link StepData}.
   */
  public void encode(StepObject o, EncodeContext context)
    throws EncoderException, java.io.IOException
  {
    byte[] bytes;
    try
    {
      bytes = ((StepData) o).getData();
    }
    catch (ClassCastException e)
    {
      throw new EncoderException(this, "incorrect value type");
    }

    _length.setValue(bytes.length);
    _lengthEncoder.encode(_length, context);

    context.write(bytes);
  }

  /**
   * Decode a STEP <code>data</code> object.
   */
  public StepObject decode(DecodeContext context)
    throws EncoderException, java.io.IOException
  {
    StepInt length = (StepInt) _lengthEncoder.decode(context);

    byte[] bytes = ByteArray.allocator().alloc(length.intValue());

    context.read(bytes);

    return new StepData(bytes);
  }

  public void dump(int level)
  {
    indent(level); System.out.println(getName()+" (data)");

    _lengthEncoder.dump(level+1);
  }
}
