/* STEP - An Framework for the Efficient Encoding of Trace Data
 * Copyright (C) 2002 Rhodes Brown
 *
 * STEP is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * STEP is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with STEP; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA
 */

/* STEP is distributed through the Sable Research Group at:
 * http://www.sable.mcgill.ca/step/
 */

/* Reference Version: $ProjectVersion: 0.9.2.dev.36 $ */

package step.encoding;

import step.*;

/**
 * Assume that the next (integer) value is relatively close to the last, and
 * only transmit on the difference from the last value.
 * <p>
 * NOTE: Although values may be more than 4 bytes (i.e. <code>long</code>),
 * their difference must be in the <code>int</code> range. Otherwise, the
 * behavior of this encoder is undefined.
 *
 * <h3>Algorithm:</h3>
 * <pre>
 * delta := current - last;
 * if |delta| &lt;= threshold
 *     then -&gt; delta
 *     else -&gt; meta:{@link IrregularValueEvent}, current
 * </pre>
 *
 * <b>Credit:</b> A. Samples (Mache)
 *
 * @see StepInt
 */
public class DeltaIntStrategy extends RegularValueStrategy
{
  public static final String STRATEGY_NAME = "delta";

  private StepInt _last;
  private int     _threshold;
  private Encoder _deltaEncoder;

  /**
   * Create a new "delta" strategy.
   *
   * @param baseEncoder the encoder to defer to for irregular values
   * @param threshold   the maximum (absolute) delta value to encode
   *
   * @throws IllegalArgumentException if the threshold is &lt; 0.
   */
  public DeltaIntStrategy(Encoder baseEncoder, int threshold)
  {
    super(baseEncoder, STRATEGY_NAME);

    if ( threshold < 0 )
      throw new IllegalArgumentException("threshold must be >= 0");

    _threshold    = threshold;
    _deltaEncoder = getFactory().newEncoder(makeFieldName("delta"), StepInt.TYPE);

    setNextIsRegular(false); // => initial read is starting point
  }

  /**
   * Encode an integer value.
   *
   * @throws EncoderException if the value is not a {@link StepInt}.
   */
  public void encode(StepObject o, EncodeContext context)
    throws EncoderException, java.io.IOException
  {
    StepInt current;
    try
    {
      current = (StepInt) o;
    }
    catch (ClassCastException e)
    {
      throw new EncoderException(this, "incorrect value type");
    }

    StepInt delta;
    try
    {
      delta = current.sub(_last);
    }
    catch (NullPointerException e)
    {
      assert _last == null; // => initial value

      baseEncode(_last = current, context);
      return;
    }

    if ( Math.abs(delta.intValue()) <= _threshold )
    {
      _deltaEncoder.encode(delta, context);
    }
    else
    {
      context.metaWrite(new IrregularValueEvent(new StepString(getName())));
      baseEncode(current, context);
    }

    _last = current;
  }

  public StepObject decode(DecodeContext context)
    throws EncoderException, java.io.IOException
  {
    StepInt current;

    if ( _nextIsRegular )
    {
      assert _last != null;

      StepInt delta = (StepInt) _deltaEncoder.decode(context);

      current = _last.add(delta);
    }
    else // => deviant value
    {
      setNextIsRegular(true); // IMPORTANT: set this before decoding

      current = (StepInt) baseDecode(context);
    }

    return ( _last = current );
  }

  public void dump(int level)
  {
    indent(level);   System.out.println(getName());
    indent(level+1); System.out.println("last="+_last);
    indent(level+1); System.out.println("threshold="+_threshold);

    _deltaEncoder.dump(level+1);
    _baseEncoder.dump(level+1);
  }
}
