/* STEP - An Framework for the Efficient Encoding of Trace Data
 * Copyright (C) 2002 Rhodes Brown
 *
 * STEP is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * STEP is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with STEP; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA
 */

/* STEP is distributed through the Sable Research Group at:
 * http://www.sable.mcgill.ca/step/
 */

/* Reference Version: $ProjectVersion: 0.9.2.dev.36 $ */

package step.encoding;

import java.io.IOException;

import java.util.Map;
import java.util.List;
import java.util.Iterator;

import step.StepObject;

/**
 * An agent for encoding/decoding objects to/from the STEP binary format.
 * <p>
 * Encoders encapsulate the information necessary to convert STEP objects to and
 * from a set of bytes. Some encoders apply a standard rule, while others use an
 * adaptive strategy that attempts to minimize the average number of bytes per
 * value.
 * <p>
 * Encoders are normally instantiated using a
 * {@linkplain EncoderFactory factory} object.
 *
 * @see EncodeContext
 * @see DecodeContext
 */
public abstract class Encoder
{
  private String         _name;
  private EncoderFactory _factory;

  public Encoder(String name, EncoderFactory factory)
  {
    _name    = name;    assert _name    != null;
    _factory = factory; assert _factory != null;
  }

  /**
   * Get the name of this encoder.
   */
  public String getName()
  {
    return _name;
  }

  /**
   * Get the factory used to create this encoder.
   */
  public EncoderFactory getFactory()
  {
    return _factory;
  }

  /**
   * Create an appropriate name for a sub-field encoder.
   */
  protected String makeFieldName(String fieldName)
  {
    return getName()+"."+fieldName;
  }

  /**
   * Encode a value.
   *
   * @throws EncoderException if the encode operation encountered an error.
   * @throws IOException if the encode operation encountered an I/O failure.
   */
  public abstract void encode(StepObject o, EncodeContext context)
    throws EncoderException, IOException;

  /**
   * Decode a value.
   *
   * @throws EncoderException if the decode operation encountered an error.
   * @throws IOException if the decode operation encountered an I/O failure.
   */
  public abstract StepObject decode(DecodeContext context)
    throws EncoderException, IOException;

  /**
   * Send a text message to this encoder.
   * <p>
   * In general, messages are intended for the decoding process.
   * <p>
   * The default behavior is to ignore the message.
   *
   * @see EncoderMessageEvent
   */
  protected void eventMessage(String msg)
  {
    // do nothing
  }

  public abstract void dump(int level);

  protected static void indent(int level)
  {
    for (int i=0; i<level; i++) System.out.print("  ");
  }
}
