/* STEP - An Framework for the Efficient Encoding of Trace Data
 * Copyright (C) 2002 Rhodes Brown
 *
 * STEP is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * STEP is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with STEP; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA
 */

/* STEP is distributed through the Sable Research Group at:
 * http://www.sable.mcgill.ca/step/
 */

/* Reference Version: $ProjectVersion: 0.9.2.dev.36 $ */

package step.encoding;

import java.io.PrintWriter;

import java.util.Iterator;
import java.util.List;
import java.util.Map;

import step.*;
import step.typedef.*;
import step.encoding.profile.*;

/**
 */
public class EncoderFactory
{
  private Map  _nameToEncoder    = new java.util.HashMap();
  private Map  _nameToEventList  = new java.util.HashMap();
  private List _profiledEncoders = new java.util.LinkedList();

  public EncoderFactory()
  {
    Runtime.getRuntime().addShutdownHook(new Thread()
    {
      public void run()
      {
        if ( _profiledEncoders.size() > 0 )
        {
          System.out.println("----------------");
          System.out.println("Encoder Profiles");
          System.out.println("----------------");

          PrintWriter out = new PrintWriter(System.out, true);

          Iterator iProfiledEncoders = _profiledEncoders.iterator();

          while (iProfiledEncoders.hasNext())
          {
            out.println();
            ((Profiled) iProfiledEncoders.next()).printProfile(out);
          }
        }
      }
    });
  }

  /**
   * Create a "universal" record encoder, capable of encoding all STEP records.
   */
  public Encoder newRecordEncoder(String name)
  {
    return register(new RecordEncoder(name, this));
  }

  /**
   * Create a new encoder using the default definition for the given type.
   * @see TypeDef#getDefault(Type)
   */
  public Encoder newEncoder(String name, Type valueType)
  {
    return newEncoder(name, TypeDef.getDefault(valueType));
  }

  /**
   * Create a new encoder for the given type definition.
   */
  public Encoder newEncoder(String name, TypeDef valueTypeDef)
  {
    return newEncoder(name, valueTypeDef, new java.util.LinkedList());
  }

  /**
   * Create a new encoder using the default definition for the given type.
   * @see TypeDef#getDefault(Type)
   */
  public Encoder newEncoder(String name, Type valueType, List attributes)
  {
    return newEncoder(name, TypeDef.getDefault(valueType), attributes);
  }

  /**
   * Create a new encoder for the given type definition with the given
   * attributes.
   *
   * @throws IllegalArgumentException if the given type definition
   *   cannot be handled by this factory.
   */
  public Encoder newEncoder(String name, TypeDef typeDef, List attributes)
  {
    // ensure the definition is complete...
    typeDef.resolve();

    Type type = typeDef.getType();
    Encoder encoder;

    /*
     * int
     *   size=<fixed width>|<start width>..|<min width>+|creep
     *   unsigned
     *   delta=<threshold>
     *   offset|offset=<base>
     *   stride=<increment>
     *   window=<threshold>
     *
     *   <property:"unsigned"> => unsigned
     *   <property:"address">  => unsigned
     *
     * defaults:
     *   size=creep
     */
    if ( type.equals(StepInt.TYPE) )
    {
      // property attributes
      boolean unsigned = false;
      {
        String values[] = splitValues(attributes, "property");
        for (int i=0; i<values.length; i++)
        {
          String value = values[i];

          if ( value.equalsIgnoreCase("unsigned") || value.equalsIgnoreCase("address") )
          {
            unsigned = true;
          }
        }
      }

      final int CREEP = -1;

      // encoding attributes
      int sizeOpt = CREEP;
      int size    = 1;
      {
        String values[] = splitValues(attributes, "encoding");

        // basic encoding
        for (int i=0; i<values.length; i++)
        {
          String value = values[i];

          if ( value.toLowerCase().startsWith("size") )
          {
            value = value.substring(value.indexOf("=")+1).trim();

            if ( value.equals("creep") )
            {
              sizeOpt = CREEP;
            }
            else
            {
              int grow    = value.indexOf("..");
              int elastic = value.indexOf("+");
              if ( grow >= 0 )
              {
                size    = Integer.parseInt(value.substring(0, grow));
                sizeOpt = IntEncoder.GROW_SIZE;
              }
              else if ( elastic >= 0 )
              {
                size    = Integer.parseInt(value.substring(0, elastic));
                sizeOpt = IntEncoder.ELASTIC_SIZE;
              }
              else // => fixed size
              {
                size    = Integer.parseInt(value);
                sizeOpt = IntEncoder.FIXED_SIZE;
              }
              
              if (( size < 0 ) || ( size > 8 ))
                throw new RuntimeException("invalid size for int: "+size); // FIXME => verifier
            }
          }
          else if ( value.equalsIgnoreCase("unsigned") )
          {
            unsigned = true;
          }
        }

        if ( sizeOpt == CREEP )
          encoder = new CreepIntEncoder(name, unsigned, this);
        else
          encoder = new IntEncoder(name, size, sizeOpt, unsigned, this);

        // strategy
        Strategy strategy = null;
        for (int i=0; i<values.length; i++)
        {
          String value = values[i];

          if ( value.toLowerCase().startsWith("delta") )
          {
            int threshold = Integer.parseInt(value.substring(value.indexOf("=")+1).trim());

            if ( threshold < 0 )
              throw new RuntimeException("invalid threshold for delta strategy: "+threshold);
              // FIXME => verifier

            if ( profile() )
              _profiledEncoders.add(strategy = new PDeltaIntStrategy(encoder, threshold));
            else
              strategy = new DeltaIntStrategy(encoder, threshold);
          }
          else if ( value.toLowerCase().startsWith("offset") )
          {
            if ( value.indexOf("=") >= 0 )
            {
              int base = Integer.parseInt(value.substring(value.indexOf("=")+1).trim());
              strategy = new OffsetIntStrategy(encoder, new StepInt(base));
            }
            else
              strategy = new OffsetIntStrategy(encoder);
          }
          else if ( value.toLowerCase().startsWith("stride") )
          {
            int stride = Integer.parseInt(value.substring(value.indexOf("=")+1).trim());

            if ( profile() )
              _profiledEncoders.add(strategy = new PStrideIntStrategy(encoder, stride));
            else
              strategy = new StrideIntStrategy(encoder, stride);
          }
          else if ( value.toLowerCase().startsWith("window") )
          {
            int threshold = Integer.parseInt(value.substring(value.indexOf("=")+1).trim());

            if ( threshold < 0 )
              throw new RuntimeException("invalid threshold for window strategy: "+threshold);
              // FIXME => verifier

            if ( profile() )
              _profiledEncoders.add(strategy = new PWindowIntStrategy(encoder, threshold));
            else
              strategy = new WindowIntStrategy(encoder, threshold);
          }
        }

        register(encoder);

        if ( strategy != null )
          encoder = register(strategy);
      }
    }

    /*
     * string
     *   charset=US-ASCII|ISO-8859-1|UTF-8|UTF-16|...
     *   .length : int
     *
     * defaults:
     *   charset=UTF-8
     *   .length : unsigned,size=creep
     */
    else if ( type.equals(StepString.TYPE) )
    {
      String charset = "UTF-8";

      // encoding
      String values[] = splitValues(attributes, "encoding");
      for (int i=0; i<values.length; i++)
      {
        String value = values[i];

        if ( value.toLowerCase().startsWith("charset") )
          charset = value.substring(value.indexOf("=")+1).trim();
      }

      FieldDef lengthDef = typeDef.getField(StepString.LENGTH_FIELD.getName());

      encoder = register(new StringEncoder(name, charset, lengthDef, this));
    }

    /*
     * data
     *   .length : int
     *
     * defaults:
     *   .length : unsigned,size=creep
     */
    else if ( type.equals(StepData.TYPE) )
    {
      FieldDef lengthDef = typeDef.getField(StepData.LENGTH_FIELD.getName());

      encoder = register(new DataEncoder(name, lengthDef, this));
    }

    /*
     * <type>[]
     *   .length  : int
     *   .element : <type>
     *
     * defaults:
     *   .length : unsigned,size=creep
     */
    else if ( type.isArrayType() )
    {
      FieldDef lengthDef  = typeDef.getField(StepArray.LENGTH_FIELD.getName());
      FieldDef elementDef = typeDef.getField(StepArray.ELEMENT_FIELD.getName());

      encoder = register(new ArrayEncoder(name, lengthDef, elementDef, this));
    }

    /*
     * <record type>
     *   type=variable|default|constant
     *
     * defaults:
     *   type=default
     */
    else if ( typeDef instanceof RecordDef )
    {
      Attribute typeEncAttr = new Attribute("encoding", "default"); // default: "default"
      String values[] = splitValues(attributes, "encoding");
      for (int i=0; i<values.length; i++)
      {
        String value = values[i];

        if ( value.toLowerCase().startsWith("type") )
        {
          String typeStrategy = value.substring(value.indexOf("=")+1).trim();

          if ( typeStrategy.equals("variable") )
            typeEncAttr = new Attribute("encoding", "identifier");
          else if ( typeStrategy.equals("default") )
            typeEncAttr = new Attribute("encoding", "default");
          else if ( typeStrategy.equals("constant") )
            typeEncAttr = new Attribute("encoding", "constant");
        }
      }

      List typeEncAttrs = new java.util.LinkedList();
      typeEncAttrs.add(typeEncAttr);

      encoder =
        register(new RecordEncoder(name, (RecordDef) typeDef, typeEncAttrs, this));
    }
    else
    {
      throw new IllegalArgumentException("can't handle type "+type);
    }

    // general strategy:
    //   
    //   identifier
    //   cache=<size>
    //   constant
    //   default
    //   repeat
    {
      Strategy strategy = null;

      String values[] = splitValues(attributes, "encoding");
      for (int i=0; i<values.length; i++)
      {
        String value = values[i];

        if ( value.equalsIgnoreCase("identifier") )
        {
          if ( profile() )
            _profiledEncoders.add(strategy = new PIdentifierStrategy(encoder));
          else
            strategy = new IdentifierStrategy(encoder);
        }
        else if ( value.toLowerCase().startsWith("cache") )
        {
          int size = Integer.parseInt(value.substring(value.indexOf("=")+1).trim());

          if ( profile() )
            _profiledEncoders.add(strategy = new PCacheStrategy(encoder, size));
          else
            strategy = new CacheStrategy(encoder, size);
        }
        else if ( value.equalsIgnoreCase("constant") )
        {

          if ( profile() )
            _profiledEncoders.add(
              strategy = new PConstantStrategy(encoder, ConstantStrategy.CONSTANT_STRATEGY));
          else
            strategy = new ConstantStrategy(encoder, ConstantStrategy.CONSTANT_STRATEGY);
        }
        else if ( value.equalsIgnoreCase("default") )
        {
          if ( profile() )
            _profiledEncoders.add(
              strategy = new PConstantStrategy(encoder, ConstantStrategy.DEFAULT_STRATEGY));
          else
            strategy = new ConstantStrategy(encoder, ConstantStrategy.DEFAULT_STRATEGY);
        }
        else if ( value.equalsIgnoreCase("repeat") )
        {
          if ( profile() )
            _profiledEncoders.add(
              strategy = new PConstantStrategy(encoder, ConstantStrategy.REPEAT_STRATEGY));
          else
            strategy = new ConstantStrategy(encoder, ConstantStrategy.REPEAT_STRATEGY);
        }
      }
      
      if ( strategy != null )
        encoder = register(strategy);
    }

    // apply pending events
    {
      Iterator iTargets = _nameToEventList.keySet().iterator();
      while (iTargets.hasNext())
      {
        String  target    = (String) iTargets.next();
        Encoder targetEnc = (Encoder) _nameToEncoder.get(target);

        if ( targetEnc != null )
        {
          Iterator iEvents = ((List) _nameToEventList.get(target)).iterator();
          while (iEvents.hasNext())
            ((EncoderEvent) iEvents.next()).applyToEncoder(targetEnc);

          _nameToEventList.remove(target);
        }
      }
    }

    return encoder;
  }


  /**
   * If the target exists, apply the event immediately, otherwise queue the
   * event until the target encoder is created by this factory.
   */
  public void queueEvent(EncoderEvent event)
  {
    String target = event.target.toString();

    Encoder encoder = (Encoder) _nameToEncoder.get(target);

    if ( encoder == null )
    {
      List eventList = (List) _nameToEventList.get(target);

      if ( eventList == null )
      {
        _nameToEventList.put(target, eventList = new java.util.LinkedList());
      }

      eventList.add(event);
    }
    else
    {
      event.applyToEncoder(encoder);
    }
  }

  private String[] splitValues(List attributes, String group)
  {
    int count = 0; List vList = new java.util.LinkedList();
    Iterator iAttributes = attributes.iterator();
    while (iAttributes.hasNext())
    {
      Attribute attr = (Attribute) iAttributes.next();

      if ( attr.getGroup().equalsIgnoreCase(group) )
      {
        String[] valueSplit = attr.getValue().split(",");
        count += valueSplit.length;
        vList.add(valueSplit);
      }
    }

    String[] values = new String[count]; int j=0;
    Iterator iValues = vList.iterator();
    while (iValues.hasNext())
    {
      String[] valueSplit = (String[]) iValues.next();
      for (int i=0; i<valueSplit.length; i++)
        values[j++] = valueSplit[i].trim();
    }

    return values;
  }

  private Encoder register(Encoder encoder)
  {
    _nameToEncoder.put(encoder.getName(), encoder);
    return encoder;
  }

  private boolean profile()
  {
    return Boolean.getBoolean("step.encoding.profile");
  }

  public Encoder getEncoder(String name)
  {
    return (Encoder) _nameToEncoder.get(name);
  }
}
