/* STEP - An Framework for the Efficient Encoding of Trace Data
 * Copyright (C) 2002 Rhodes Brown
 *
 * STEP is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * STEP is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with STEP; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA
 */

/* STEP is distributed through the Sable Research Group at:
 * http://www.sable.mcgill.ca/step/
 */

/* Reference Version: $ProjectVersion: 0.9.2.dev.36 $ */

package step.encoding;

import java.util.Map;
import java.util.List;

import step.*;
import step.typedef.*;

/**
 * Assume that values derive from a relatively small, fixed distribution.
 * Instead of encoding the full value, use a small integer ID in its place.
 * <p>
 * NOTE: The ID values are encoded with a {@link CreepIntEncoder}, thus the
 * strategy is optimized for distributions with &lt; 128 distinct values.
 */
public class IdentifierStrategy extends RegularValueStrategy
{
  public static final String STRATEGY_NAME = "identifier";

  private Encoder _idEncoder;
  private Map     _valueToId;
  private Map     _idToValue;
  private int     _nextId = 0;

  /**
   * Create a new "identifier" strategy.
   *
   * @param baseEncoder the encoder to defer to when writing previously unmapped
   *   values.
   */
  public IdentifierStrategy(Encoder baseEncoder)
  {
    super(baseEncoder, STRATEGY_NAME);

    _valueToId = new java.util.HashMap(512, 0.5F);
    _idToValue = new java.util.HashMap(512, 0.5F);

    {
      List idEncAttrs = new java.util.ArrayList(2);
      idEncAttrs.add(new Attribute("encoding", "unsigned"));
      idEncAttrs.add(new Attribute("encoding", "size=creep"));

      _idEncoder =
        getFactory().newEncoder(makeFieldName("id"), StepInt.TYPE, idEncAttrs);
    }

    setNextIsRegular(false); // => no mapping yet
  }

  /**
   * Encode a value as an ID.
   */
  public void encode(StepObject o, EncodeContext context)
    throws EncoderException, java.io.IOException
  {
    StepInt id = (StepInt) valueToId(o);

    if ( id == null )
    {
      map(o, id = new StepInt(_nextId++));

      context.metaWrite(new IrregularValueEvent(new StepString(getName())));
      baseEncode(o, context);
    }

    _idEncoder.encode(id, context);
  }

  /**
   * Decode an ID, map it to a value.
   */
  public StepObject decode(DecodeContext context)
    throws EncoderException, java.io.IOException
  {
    StepObject value;

    if ( _nextIsRegular )
    {
      value = (StepObject) idToValue(_idEncoder.decode(context));

      if ( value == null )
        throw new RuntimeException("no value mapping for id");
    }
    else // => new mapping
    {
      setNextIsRegular(true); // IMPORTANT: set this before decoding

      map(value = baseDecode(context), (StepInt) _idEncoder.decode(context));
    }

    return value;
  }

  /**
   * Get the current ID for this value.
   *
   * @return the current ID (as a {@link StepInt}) or
   *   <code>null</code> if no ID is defined for the value.
   */
  protected Object valueToId(StepObject value)
  {
    return _valueToId.get(value);
  }

  /**
   * Get the current value for this ID.
   *
   * @return the current value associated with the ID or
   *   <code>null</code> if no value is defined for the ID.
   */
  protected Object idToValue(StepObject id)
  {
    return _idToValue.get(id);
  }

  private void map(StepObject value, StepInt id)
  {
    _valueToId.put(value, id);
    _idToValue.put(id, value);
  }

  /**
   * Pre-map a value in the expected distribution.
   *
   * <p>
   * WARNING: This method is provided for bootstrapping purposes. Don't use it
   * unless you understand the consequences.
   *
   * @throws IllegalArgumentException if a mapping already exists for the value.
   */
  protected void add(StepObject value)
  {
    if ( _valueToId.containsKey(value) )
      throw new IllegalArgumentException("mapping already exists for "+value);

    map(value, new StepInt(_nextId++));
  }

  public void dump(int level)
  {
    indent(level);   System.out.println(getName());
    indent(level+1); System.out.println("map="+_valueToId);

    _idEncoder.dump(level+1);
    _baseEncoder.dump(level+1);
  }
}
