/* STEP - An Framework for the Efficient Encoding of Trace Data
 * Copyright (C) 2002 Rhodes Brown
 *
 * STEP is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * STEP is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with STEP; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA
 */

/* STEP is distributed through the Sable Research Group at:
 * http://www.sable.mcgill.ca/step/
 */

/* Reference Version: $ProjectVersion: 0.9.2.dev.36 $ */

package step.encoding;

import step.StepObject;
import step.StepInt;
import step.StepString;

import step.util.*;

/**
 * An encoder for STEP <code>int</code> objects that uses a specific number of bytes.
 * <p>
 * Bytes are written in LITTLE endian (least significant -&gt; most significant)
 * order. If the required number of bits is greater than the allowed size,
 * either an {@link EncoderMessageEvent} signals the deviant size value or an
 * exception is thrown if the size is fixed. If the number of bytes is too
 * small, the value is padded.
 *
 * @see StepInt
 * @see CreepIntEncoder
 */
public class IntEncoder extends Encoder
{
  public static final int FIXED_SIZE   = 1;
  public static final int ELASTIC_SIZE = 2;
  public static final int GROW_SIZE    = 3;

  static final int LONG_SIZE = 8;

  private byte[]  _buf = new byte[LONG_SIZE];
  private int     _size;
  private int     _resize;
  private boolean _unsigned;
  private Integer _nextSize; // for deviant decode sizes

  /**
   * Create a new encoder for STEP <code>int</code> types.
   *
   * @param name      the name for this encoder.
   * @param size      the (minimum) number of bytes to use for encoding
   * @param resize    the policy regarding size changes.
   *   One of {@link #FIXED_SIZE}, {@link #ELASTIC_SIZE} or {@link #GROW_SIZE}.
   * @param unsigned  <code>true</code> if encoded/decoded values are always &gt;= 0
   * @param factory   the {@link EncoderFactory} used to create this encoder
   *
   * @throws IllegalArgumentException if the size or resize values are invalid.
   */
  public IntEncoder(String name, int size, int resize, boolean unsigned,
                    EncoderFactory factory)
  {
    super(name, factory);

    if ( (size <= 0) || (LONG_SIZE < size) )
      throw new IllegalArgumentException("size must be in the range [1.."+LONG_SIZE+"]");

    if ( (resize != FIXED_SIZE) &&
         (resize != ELASTIC_SIZE) &&
         (resize != GROW_SIZE) )
      throw new IllegalArgumentException("invalid resize policy "+resize);

    _size     = size;
    _resize   = resize;
    _unsigned = unsigned;
  }

  /**
   * Encode the given STEP <code>int</code> value.
   *
   * @throws EncoderException if
   * <ul>
   *   <li> the value was not a {@link StepInt}.
   *   <li> the "unsigned" option was set and the value was negative
   *   <li> the "fixed" resize policy was set and the value requires too
   *        many bytes to encode
   * </ul>
   */
  public void encode(StepObject o, EncodeContext context)
    throws EncoderException, java.io.IOException
  {
    long n;
    try
    {
      n = ((StepInt) o).longValue();
    }
    catch (ClassCastException e)
    {
      throw new EncoderException(this, "incorrect value type");
    }

    int used = 0;

    if ( n < 0L )
    {
      if ( _unsigned )
        throw new EncoderException(this, "attempt to encode negative value as unsigned");

      long n_comp = n ^ 0xFFFFFFFFFFFFFFFFL; // invert the bits
      do
      {
        _buf[used++] = (byte) (n & 0xFF); // grab the low bytes
        n >>= 8; n_comp >>= 8;            // shift right
      }
      while ( n_comp != 0L );

      while ( used < _size ) // => fill unused bytes
        _buf[used++] = (byte) 0xFF;
    }
    else // n >= 0
    {
      do
      {
        _buf[used++] = (byte) (n & 0xFF); // grab the low byte
        n >>= 8;                          // shift right
      }
      while ( n != 0L );

      while ( used < _size ) // => fill unused bytes
        _buf[used++] = 0x00;

      if ( (!_unsigned) && (_buf[used-1] < 0) ) // => distinguish value as positive
        _buf[used++] = 0x00;
    }


    assert used >= _size;

    if ( used == _size )
    {
      byte[] bytes = ByteArray.allocator().alloc(_size);
      {
        for (int i=0; i<_size; i++) bytes[i] = _buf[i];
        context.write(bytes);
      }
      ByteArray.allocator().free(bytes);
    }
    else if ( _resize == FIXED_SIZE )
    {
      throw new EncoderException(this, "value cannot fit in fixed size");
    }
    else // => signal deviant size value with meta data
    {
      context.metaWrite(
        new EncoderMessageEvent(new StepString(getName()), new StepString("size:="+used)));
      
      if ( _resize == ELASTIC_SIZE )
      {
        int oldSize = _size;
        _size = used;
        {
          encode(o, context); // try again...
        }
        _size = oldSize;
      }
      else
      {
        assert _resize == GROW_SIZE;

        _size = used;
        encode(o, context); // try again...
      }
    }
  }

  /**
   * Decode a STEP <code>int</code> value.
   */
  public StepObject decode(DecodeContext context)
    throws EncoderException, java.io.IOException
  {
    long n = 0;
    byte[] bytes = ByteArray.allocator().alloc(nextSize());
    {
      assert (0 < bytes.length) && (bytes.length <= LONG_SIZE);

      context.read(bytes);

      int i;
      for (i=0; i<bytes.length; i++)
        n |= ( (((long) bytes[i]) << (8*i)) & (0xFFL << (8*i)) );

      if ( (!_unsigned) && (bytes[bytes.length-1] < 0) ) // negative => sign extend
      {
        while ( i < LONG_SIZE ) n |= (0xFFL << (8*i++));
      }
    }
    ByteArray.allocator().free(bytes);

    return new StepInt(n);
  }

  protected void eventMessage(String msg)
  {
    if ( msg.startsWith("size:=") )
    {
      if ( _resize == ELASTIC_SIZE )
      {
        _nextSize = Integer.valueOf(msg.substring(msg.indexOf("=")+1));
      }
      else
      {
        assert _resize == GROW_SIZE;

        _size = Integer.parseInt(msg.substring(msg.indexOf("=")+1));
      }
    }
    else // => defer
    {
      super.eventMessage(msg);
    }
  }

  private int nextSize()
  {
    if ( _nextSize == null )
    {
      return _size;
    }
    else
    {
      int size = _nextSize.intValue();
      _nextSize = null;
      return size;
    }
  }

  public void dump(int level)
  {
    indent(level);   System.out.println(getName()+" (int:normal)");
    indent(level+1); System.out.println("size="+_size);
    indent(level+1); System.out.println("resize="+_resize);
    indent(level+1); System.out.println("unsigned="+_unsigned);
  }
}
