/* STEP - An Framework for the Efficient Encoding of Trace Data
 * Copyright (C) 2002 Rhodes Brown
 *
 * STEP is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * STEP is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with STEP; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA
 */

/* STEP is distributed through the Sable Research Group at:
 * http://www.sable.mcgill.ca/step/
 */

/* Reference Version: $ProjectVersion: 0.9.2.dev.36 $ */

package step.encoding;

import java.util.List;
import java.util.Iterator;

import step.*;
import step.typedef.*;
import step.util.ByteArray;

/**
 * A record that contains another record with interleaved meta-data.
 *
 * @see MetaEvent
 */
public class MetaRecord implements StepRecord
{
  public static final RecordDef DEFINITION;

  static
  {
    DEFINITION = RecordDef.builder().newRecordDef("MetaRecord", "step.encoding")
      .addField(FieldDef.builder().newFieldDef("data", StepData.TYPE.toArrayType())
        .makeFieldDef())
      .addField(FieldDef.builder().newFieldDef("metaData", StepData.TYPE.toArrayType())
        .makeFieldDef())
      .setFactory(new Factory())
      .makeRecordDef();
  }

  public StepArray data;
  public StepArray metaData;

  /**
   * Create a new meta record from data segments and meta segments.
   * <p>
   * The meta segments should consist of complete encoded records, generally
   * either a {@link MetaEvent} or another MetaRecord.
   * <p>
   * There should be <code>n+1</code> data segments for <code>n</code> meta
   * segments.
   */
  public MetaRecord(StepArray data, StepArray metaData)
  {
    assert data.length() == (metaData.length()+1);

    this.data     = data;
    this.metaData = metaData;
  }

  /**
   * Create a new meta record from data segments and meta segments.
   * <p>
   * The meta segments should consist of complete encoded records, generally
   * either a {@link MetaEvent} or another MetaRecord.
   * <p>
   * There should be <code>n+1</code> data segments for <code>n</code> meta
   * segments.
   *
   * @throws ClassCastException if the list elements are not
   *   <code>byte[]</code> objects.
   */
  public MetaRecord(List data, List metaData)
  {
    assert data.size() == (metaData.size()+1);

    StepObject[] dataObjs = new StepObject[data.size()];
    Iterator iData = data.iterator();
    for (int i=0; iData.hasNext(); i++)
      dataObjs[i] = new StepData((byte[]) iData.next()); // throws Cast...

    StepObject[] metaDataObjs = new StepObject[metaData.size()];
    Iterator iMetaData = metaData.iterator();
    for (int i=0; iMetaData.hasNext(); i++)
      metaDataObjs[i] = new StepData((byte[]) iMetaData.next()); // throws Cast...

    this.data     = new StepArray(dataObjs);
    this.metaData = new StepArray(metaDataObjs);
  }

  /**
   * Compare two MetaRecords for equality.
   * @throws ClassCastException if the given object is not a MetaRecord.
   * @throws NullPointerException if the given object is <code>null</code>.
   */
  public boolean equals(Object o) //: StepObject
  {
    MetaRecord rhs = (MetaRecord) o;
    return data.equals(rhs.data) && metaData.equals(rhs.metaData);
  }

  public int hashCode() //: StepObject
  {
    return data.hashCode() + metaData.hashCode();
  }

  public FieldIterator fieldIterator() //: StepRecord
  {
    StepObject[] fields = { data, metaData };
    return new FieldIterator(fields);
  }

  /**
   * Free the memory used by this record.
   * <p>
   * The underlying <code>byte[]</code> objects are freed to the
   * {@link ByteArray#allocator()}.
   */
  public void free()
  {
    Iterator iData = data.iterator();
    while (iData.hasNext())
      ByteArray.allocator().free((byte[]) ((StepData) iData.next()).getData());

    data = null;

    Iterator iMeta = metaData.iterator();
    while (iMeta.hasNext())
      ByteArray.allocator().free((byte[]) ((StepData) iMeta.next()).getData());

    metaData = null;
  }

  public static class Factory implements RecordFactory
  {
    public StepRecord newRecord(StepObject[] fieldData)
    {
      assert fieldData.length == 2;
      return new MetaRecord((StepArray) fieldData[0], (StepArray) fieldData[1]);
    }
  }
}
