/* STEP - An Framework for the Efficient Encoding of Trace Data
 * Copyright (C) 2002 Rhodes Brown
 *
 * STEP is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * STEP is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with STEP; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA
 */

/* STEP is distributed through the Sable Research Group at:
 * http://www.sable.mcgill.ca/step/
 */

/* Reference Version: $ProjectVersion: 0.9.2.dev.36 $ */

package step.encoding;

import step.StepObject;
import step.StepInt;

/**
 * Instead of transmitting the absolute (integer) value, transmit the offset
 * from a fixed base value.
 * <p>
 * NOTE: The same (base) encoder is used for the offset value. Thus, if the
 * base values have the "unsigned" property, so must the offset values.
 *
 * <h3>Algorithm:</h3>
 * <pre>
 *  -&gt; current - base
 * </pre>
 *
 * <b>Credit:</b> Chilimbi, Jones, Zorn (HATF/Meta-TF)
 *
 * @see StepInt
 * @see WindowIntStrategy
 */
public class OffsetIntStrategy extends Strategy
{
  public static final String STRATEGY_NAME = "offset";

  private StepInt _base;
  private Encoder _deltaEncoder;

  /**
   * Create a new "offset" strategy.
   * The first value is used as the base value.
   *
   * @param baseEncoder the encoder to use for the inital value
   */
  public OffsetIntStrategy(Encoder baseEncoder)
  {
    super(baseEncoder, STRATEGY_NAME);

    _base = null;

    _deltaEncoder = getFactory().newEncoder(makeFieldName("delta"), StepInt.TYPE);
  }

  /**
   * Create a new "offset" strategy.
   *
   * @param baseEncoder the base encoder, only used for identification
   * @param base        the base value to subtract from input values.
   */
  public OffsetIntStrategy(Encoder baseEncoder, StepInt base)
  {
    super(baseEncoder, STRATEGY_NAME);

    _base = base; assert _base != null;

    _deltaEncoder = getFactory().newEncoder(makeFieldName("delta"), StepInt.TYPE);
  }

  /**
   * Encode an offset value.
   *
   * @throws EncoderException if the value is not a {@link StepInt}.
   */
  public void encode(StepObject o, EncodeContext context)
    throws EncoderException, java.io.IOException
  {
    StepInt current;
    try
    {
      current = (StepInt) o;
    }
    catch (ClassCastException e)
    {
      throw new EncoderException(this, "incorrect value type");
    }

    try
    {
      _deltaEncoder.encode(current.sub(_base), context);
    }
    catch (NullPointerException e)
    {
      assert _base == null; // => first value

      baseEncode(_base = current, context);
    }
  }

  /**
   * Decode an offset value.
   */
  public StepObject decode(DecodeContext context)
    throws EncoderException, java.io.IOException
  {
    if ( _base == null )
    {
      return (_base = (StepInt) baseDecode(context));
    }
    else
    {
      return _base.add((StepInt) _deltaEncoder.decode(context));
    }
  }

  public void dump(int level)
  {
    indent(level);   System.out.println(getName());
    indent(level+1); System.out.println("base="+_base);

    _baseEncoder.dump(level+1);
  }
}
