/* STEP - An Framework for the Efficient Encoding of Trace Data
 * Copyright (C) 2002 Rhodes Brown
 *
 * STEP is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * STEP is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with STEP; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA
 */

/* STEP is distributed through the Sable Research Group at:
 * http://www.sable.mcgill.ca/step/
 */

/* Reference Version: $ProjectVersion: 0.9.2.dev.36 $ */

package step.encoding;

import java.util.Map;
import java.util.List;
import java.util.Iterator;

import step.*;
import step.typedef.*;

/**
 * An encoder for STEP record objects.
 *
 * @see StepRecord
 * @see FieldIterator
 * @see RecordFactory
 */
public class RecordEncoder extends Encoder
{
  public static final String TYPE_FIELD_NAME = "$type";

  private Encoder _typeEncoder;
  private TypeDef _baseDef;
  private Map     _typeToEncoders = new java.util.HashMap();

  /**
   * Create a "universal" record encoder, capable of encoding all STEP records.
   * <p>
   * Types are tagged with their type name, encoded the {@link IdentifierStrategy}.
   */
  public RecordEncoder(String name, EncoderFactory factory)
  {
    super(name, factory);

    EncoderEvent.DEFINITION.getType(); // hack: need to pre-load RecordDef

    _baseDef = TypeDef.getDefault(StepRecord.TYPE);

    // create the type field encoder:
    {
      List typeEncAttrs = new java.util.LinkedList();
      typeEncAttrs.add(new Attribute("encoding", "identifier"));

      _typeEncoder = getFactory().newEncoder(
        makeFieldName(TYPE_FIELD_NAME), StepString.TYPE, typeEncAttrs);
    }

    // bootstrap $type encoder:
    {
      IdentifierStrategy typeIdEncoder = (IdentifierStrategy) _typeEncoder;

      Type[] knownTypes =
      {
        MetaRecord.DEFINITION.getType(),
        IrregularValueEvent.DEFINITION.getType(),
        EncoderMessageEvent.DEFINITION.getType()
      };

      for (int i=0; i<knownTypes.length; i++)
      {
        _typeToEncoders.put(knownTypes[i], makeFieldEncoders(knownTypes[i]));
        typeIdEncoder.add(new StepString(knownTypes[i].toString()));
      }

      // assume first is MetaRecord (-> already mapped)
      typeIdEncoder.setNextIsRegular(true);
    }

    // bootstrap IrregularValueEvent:
    {
      // NOTE: IdentifierStrategy uses IrregularValueEvent, but
      //       IrregularValueEvent.target uses IdentifierStrategy...

      Type iveType = IrregularValueEvent.DEFINITION.getType();
        
      IdentifierStrategy iveTargetEncoder = (IdentifierStrategy)
        getFactory().getEncoder("identifier("+makeFieldName(iveType+":target")+")");

      iveTargetEncoder.add(new StepString(iveTargetEncoder.getName())); // itself
      iveTargetEncoder.add(new StepString(_typeEncoder.getName()));  // this.$type

      // assume first target is itself or this.$type (-> already mapped)
      iveTargetEncoder.setNextIsRegular(true);
    }
  }

  /**
   * Create a new encoder for the given record definition.
   * <p>
   * This encoder will support the given record type and all its sub-types.
   *
   * @param name         the name of this encoder.
   * @param baseDef      the root type to allow through this encoder.
   * @param typeEncAttrs attributes for the type tag encoder.
   * @param factory      the {@link EncoderFactory} used to create this encoder.
   *
   * @throws IllegalArgumentException if the given base type may not be used.
   *   NOTE: The internal records (such as {@link MetaRecord} or the
   *   {@link EncoderEvent} records) are not permitted because the require
   *   subtle bootstrapping.
   */
  public RecordEncoder(String name, RecordDef baseDef, List typeEncAttrs,
                       EncoderFactory factory)
  {
    super(name, factory);

    if ( baseDef.getType().isParent(MetaRecord.DEFINITION.getType()) ||
         baseDef.getType().isParent(EncoderEvent.DEFINITION.getType()) )
    {
      throw new IllegalArgumentException("illegal base type");
    }

    _baseDef = baseDef; assert _baseDef != null;
    
    _typeEncoder = getFactory().newEncoder(
      makeFieldName(TYPE_FIELD_NAME), StepString.TYPE, typeEncAttrs);
  }

  /**
   * Encode a record.
   *
   * @throws EncoderException if the type is not allowed by this encoder.
   */
  public void encode(StepObject o, EncodeContext context)
    throws EncoderException, java.io.IOException
  {
    Type type = Type.v(o);

    if ( !(type.equals(_baseDef.getType()) || _baseDef.getType().isParent(type)) )
      throw new EncoderException(this, "can't encode "+type+" types");

    Encoder[] fieldEncoders = (Encoder[]) _typeToEncoders.get(type);

    if ( fieldEncoders == null )
    {
      _typeToEncoders.put(type, fieldEncoders = makeFieldEncoders(type));
    }

    _typeEncoder.encode(new StepString(type.toString()), context);

    int i; Iterator iFields = ((StepRecord) o).fieldIterator();
    for (i=0; iFields.hasNext(); i++)
      fieldEncoders[i].encode((StepObject) iFields.next(), context);

    assert i == fieldEncoders.length; // wrote all fields
  }
  
  /**
   * Decode a record.
   *
   * @throws RecordDefNotFoundException if a definition for the type could
   *   not be loaded.
   */
  public StepObject decode(DecodeContext context)
    throws EncoderException, java.io.IOException
  {
    Type type = Type.v(((StepString) _typeEncoder.decode(context)).toString());

    Encoder[] fieldEncoders = (Encoder[]) _typeToEncoders.get(type);

    if ( fieldEncoders == null )
    {
      _typeToEncoders.put(type, fieldEncoders = makeFieldEncoders(type));
    }

    StepObject[] fieldData = new StepObject[fieldEncoders.length];
    for (int i=0; i<fieldData.length; i++)
      fieldData[i] = fieldEncoders[i].decode(context);

    return ((RecordDef) TypeDef.getDefault(type)).getFactory().newRecord(fieldData);
  }

  private Encoder[] makeFieldEncoders(Type type)
  {
    assert type != null;

    RecordDef recordDef = (RecordDef) TypeDef.getDefault(type); 
      // -> throws RecordDefNotFound

    recordDef.resolve(); // => ensure def is complete

    Encoder[] fieldEncoders = new Encoder[recordDef.getFields().size()];
    Iterator iFieldDefs = recordDef.getFields().iterator();
    for (int i=0; iFieldDefs.hasNext(); i++)
    {
      FieldDef fieldDef = (FieldDef) iFieldDefs.next();

      if ( _baseDef.getField(fieldDef.getName()) != null ) // => merge with base definitions
        fieldDef = FieldDef.merge(fieldDef, _baseDef.getField(fieldDef.getName()));

      String encName = makeFieldName(type+":"+fieldDef.getName());

      TypeDef encDef = fieldDef.hasAltTypeDef() ?
        fieldDef.getAltTypeDef() :
        TypeDef.getDefault(fieldDef.getType());

      List encAttrs = fieldDef.getAttributes();

      fieldEncoders[i] = getFactory().newEncoder(encName, encDef, encAttrs);
    }

    return fieldEncoders;
  }

  public void dump(int level)
  {
    indent(level); System.out.println(getName());

    _typeEncoder.dump(level+1);

    Iterator iTypes = _typeToEncoders.keySet().iterator();
    while (iTypes.hasNext())
    {
      Type type = (Type) iTypes.next();

      indent(level+1); System.out.println(type);

      Encoder[] fieldEncoders = (Encoder[]) _typeToEncoders.get(type);
      for (int i=0; i<fieldEncoders.length; i++)
        fieldEncoders[i].dump(level+2);
    }
  }
}
