/* STEP - An Framework for the Efficient Encoding of Trace Data
 * Copyright (C) 2002 Rhodes Brown
 *
 * STEP is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * STEP is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with STEP; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA
 */

/* STEP is distributed through the Sable Research Group at:
 * http://www.sable.mcgill.ca/step/
 */

/* Reference Version: $ProjectVersion: 0.9.2.dev.36 $ */

package step.encoding;

import step.*;

/**
 * Assume that the (integer) values increase or decrease with a regular
 * increment.
 *
 * <h3>Algorithm:</h3>
 * <pre>
 * if current - last == stride
 *     then -&gt; <i>nothing</i>
 *     else -&gt; meta:{@link IrregularValueEvent}, current
 * </pre>
 *
 * <b>Credit:</b> Chilimbi, Jones, Zorn (HATF/Meta-TF)
 *
 * @see StepInt
 */
public class StrideIntStrategy extends RegularValueStrategy
{
  public static final String STRATEGY_NAME = "stride";

  private StepInt _last;
  private int     _stride;
  private StepInt _stride2; // avoid unnecessary allocation

  /**
   * Create a new "stride" strategy.
   *
   * @param baseEncoder the encoder to defer to for deviant values.
   * @param stride the increment from the previous value.
   */
  public StrideIntStrategy(Encoder baseEncoder, int stride)
  {
    super(baseEncoder, STRATEGY_NAME);

    _stride  = stride;
    _stride2 = new StepInt(_stride);

    setNextIsRegular(false); // => initial value is irregular
  }

  /**
   * Encode a stride value.
   *
   * @throws EncoderException if the value is not a {@link StepInt}.
   */
  public void encode(StepObject o, EncodeContext context)
    throws EncoderException, java.io.IOException
  {
    StepInt current;
    try
    {
      current = (StepInt) o;
    }
    catch (ClassCastException e)
    {
      throw new EncoderException(this, "incorrect value type");
    }

    StepInt delta;
    try
    {
      delta = current.sub(_last);
    }
    catch (NullPointerException e)
    {
      assert _last == null; // => initial value

      baseEncode(_last = current, context);
      return;
    }
    // => got delta value

    if ( delta.intValue() == _stride )
    {
      // encode nothing
    }
    else
    {
      context.metaWrite(new IrregularValueEvent(new StepString(getName())));
      baseEncode(current, context);
    }

    _last = current;
  }

  /**
   * Decode a stride value.
   */
  public StepObject decode(DecodeContext context)
    throws EncoderException, java.io.IOException
  {
    StepInt current;

    if ( _nextIsRegular )
    {
      assert _last != null;

      current = _last.add(_stride2);
    }
    else // => deviant value
    {
      setNextIsRegular(true); // IMPORTANT: set this before decoding

      current = (StepInt) baseDecode(context);
    }

    return ( _last = current );
  }

  public void dump(int level)
  {
    indent(level);   System.out.println(getName());
    indent(level+1); System.out.println("last="+_last);
    indent(level+1); System.out.println("stride="+_stride);

    _baseEncoder.dump(level+1);
  }
}
