/* STEP - An Framework for the Efficient Encoding of Trace Data
 * Copyright (C) 2002 Rhodes Brown
 *
 * STEP is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * STEP is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with STEP; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA
 */

/* STEP is distributed through the Sable Research Group at:
 * http://www.sable.mcgill.ca/step/
 */

/* Reference Version: $ProjectVersion: 0.9.2.dev.36 $ */

package step.encoding;

import java.util.List;

import step.*;
import step.typedef.*;

/**
 * An encoder for STEP <code>string</code> objects.
 *
 * @see StepString
 */
public class StringEncoder extends Encoder
{
  private Encoder _lengthEncoder;
  private String  _charset;

  private final StepInt _length = new StepInt(0);
    // avoid unnecessary allocation

  /**
   * Create a new encoder for STEP <code>string</code> objects.
   *
   * @param name       the name of this encoder
   * @param charset    the character set to use when converting values to bytes
   * @param lengthDef  a definition for the (integer) length of string objects
   * @param factory    the {@link EncoderFactory} used to create this encoder
   *
   * @throws IllegalArgumentException if the length definition does not handle
   *   {@link StepInt} values.
   *
   * @see java.nio.charset.Charset
   */
  public StringEncoder(String name, String charset, FieldDef lengthDef,
                       EncoderFactory factory)
  {
    super(name, factory);

    if ( ! lengthDef.getType().equals(StepInt.TYPE) )
      throw new IllegalArgumentException("length field must encode StepInt values");

    _charset = charset; assert _charset != null;

    String lengthName = makeFieldName(lengthDef.getName());

    TypeDef lengthTypeDef = lengthDef.hasAltTypeDef() ?
      lengthDef.getAltTypeDef() :
      TypeDef.getDefault(lengthDef.getType());

    List lengthAttrs = lengthDef.getAttributes();

    _lengthEncoder = getFactory().newEncoder(lengthName, lengthTypeDef, lengthAttrs);
  }

  /**
   * Encode a STEP <code>string</code> value.
   * @throws EncoderException if the value is not a {@link StepString}.
   */
  public void encode(StepObject o, EncodeContext context)
    throws EncoderException, java.io.IOException
  {
    String str;
    try
    {
      str = ((StepString) o).toString();
    }
    catch (ClassCastException e)
    {
      throw new EncoderException(this, "incorrect value type");
    }

    try
    {
      byte[] bytes = str.getBytes(_charset);

      _length.setValue(bytes.length);
      _lengthEncoder.encode(_length, context);

      context.write(bytes);
    }
    catch (java.io.UnsupportedEncodingException e)
    {
      throw new RuntimeException(e);
    }
  }

  /**
   * Decode a STEP <code>string</code> value.
   */
  public StepObject decode(DecodeContext context)
    throws EncoderException, java.io.IOException
  {
    StepInt length = (StepInt) _lengthEncoder.decode(context);

    byte[] bytes = new byte[length.intValue()]; // => don't use ByteArray.allocator()

    context.read(bytes);

    try
    {
      return new StepString(new String(bytes, _charset));
    }
    catch (java.io.UnsupportedEncodingException e)
    {
      throw new RuntimeException(e);
    }
  }

  public void dump(int level)
  {
    indent(level);   System.out.println(getName()+" (string)");
    indent(level+1); System.out.println("charset="+_charset);

    _lengthEncoder.dump(level+1);
  }
}
