/* STEP - An Framework for the Efficient Encoding of Trace Data
 * Copyright (C) 2002 Rhodes Brown
 *
 * STEP is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * STEP is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with STEP; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA
 */

/* STEP is distributed through the Sable Research Group at:
 * http://www.sable.mcgill.ca/step/
 */

/* Reference Version: $ProjectVersion: 0.9.2.dev.36 $ */

package step.encoding;

import step.*;

/**
 * Assume (integer) values are in the neighborhood of a base value, and transmit
 * the offset from the base. Irregular values shift the base of computation.
 * <p>
 * NOTE: Although values may be more than 4 bytes (i.e. <code>long</code>),
 * their difference must be in the <code>int</code> range. Otherwise, the
 * behavior of this encoder is undefined.
 *
 * <h3>Algorithm:</h3>
 * <pre>
 * delta := current - base
 * if |delta| &lt;= threshold
 *     then -&gt; delta
 *     else -&gt; meta:{@link IrregularValueEvent}, base:=current
 * </pre>
 *
 * @see StepInt
 * @see OffsetIntStrategy
 */
public class WindowIntStrategy extends RegularValueStrategy
{
  public static final String STRATEGY_NAME = "window";

  private StepInt _base;
  private int     _threshold;
  private Encoder _deltaEncoder;

  /**
   * Create a new "window" strategy.
   *
   * @param baseEncoder the encoder to use for deviant values.
   * @param threshold   the maximum (absolute) difference from the current base
   *   to allow.
   *
   * @throws IllegalArgumentException if the threshold is not &gt;= 0.
   */
  public WindowIntStrategy(Encoder baseEncoder, int threshold)
  {
    super(baseEncoder, STRATEGY_NAME);

    if ( threshold < 0 )
      throw new IllegalArgumentException("threshold value must be >= 0");

    _threshold    = threshold;
    _deltaEncoder = getFactory().newEncoder(makeFieldName("delta"), StepInt.TYPE);

    setNextIsRegular(false); // => initial read is starting base
  }

  /**
   * Encode a window value.
   *
   * @throws EncoderException if the value is not a {@link StepInt}.
   */
  public void encode(StepObject o, EncodeContext context)
    throws EncoderException, java.io.IOException
  {
    StepInt current;
    try
    {
      current = (StepInt) o;
    }
    catch (ClassCastException e)
    {
      throw new EncoderException(this, "incorrect value type");
    }

    StepInt delta;
    try
    {
      delta = current.sub(_base);
    }
    catch (NullPointerException e)
    {
      assert _base == null; // => initial value

      baseEncode(_base = current, context);
      return;
    }

    if ( Math.abs(delta.intValue()) <= _threshold )
    {
      _deltaEncoder.encode(delta, context);
    }
    else
    {
      context.metaWrite(new IrregularValueEvent(new StepString(getName())));
      baseEncode(_base = current, context);
    }
  }

  /**
   * Decode a window value.
   */
  public StepObject decode(DecodeContext context)
    throws EncoderException, java.io.IOException
  {
    StepInt current;

    if ( _nextIsRegular )
    {
      assert _base != null;

      StepInt delta = (StepInt) _deltaEncoder.decode(context);

      current = _base.add(delta);
    }
    else // => deviant value
    {
      setNextIsRegular(true); // IMPORTANT: set this before decoding

      _base = current = (StepInt) baseDecode(context);
    }

    return current;
  }

  public void dump(int level)
  {
    indent(level);   System.out.println(getName());
    indent(level+1); System.out.println("base="+_base);
    indent(level+1); System.out.println("threshold="+_threshold);

    _deltaEncoder.dump(level+1);
    _baseEncoder.dump(level+1);
  }
}
