/* STEP - An Framework for the Efficient Encoding of Trace Data
 * Copyright (C) 2002 Rhodes Brown
 *
 * STEP is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * STEP is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with STEP; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA
 */

/* STEP is distributed through the Sable Research Group at:
 * http://www.sable.mcgill.ca/step/
 */

/* Reference Version: $ProjectVersion: 0.9.2.dev.36 $ */

package step.io;

import java.io.InputStream;
import java.io.DataInput;
import java.io.IOException;

import java.util.Map;
import java.util.List;
import java.util.Iterator;

import step.*;
import step.typedef.*;
import step.encoding.*;
import step.util.*;

/**
 * An input stream for reading {@link StepRecord} objects.
 *
 * @see StepRecord
 * @see Encoder
 * @see StepEncodedOutput
 */
public class StepEncodedInput implements StepRecordInput
{
  private InputStream _input;

  private SizeContext   _sizeContext;
  private RecordContext _recordContext;

  private EncoderFactory _encFactory;
  private Encoder _recordEncoder, _sizeEncoder;

  /**
   */
  // TODO: throws...
  public StepEncodedInput(InputStream input) throws IOException
  {
    if ( input == null )
      throw new IllegalArgumentException("input stream is null");

    _input = input;

    // check magic:
    {
      byte[] magic = new byte[StepEncodedOutput.FILE_MAGIC.length];
      if ( _input.read(magic) != magic.length )
        throw new java.io.EOFException("incomplete file header");

      for (int i=0; i<magic.length; i++)
        if ( magic[i] != StepEncodedOutput.FILE_MAGIC[i] )
          throw new java.io.IOException("not a valid STEP data file");
    }

    // check version:
    {
      byte[] version = new byte[StepEncodedOutput.FILE_VERSION.length];
      if ( _input.read(version) != version.length )
        throw new java.io.EOFException("incomplete file header");

      for (int i=0; i<version.length; i++)
        if ( version[i] > StepEncodedOutput.FILE_VERSION[i] )
          throw new java.io.IOException("file version new than decoder");
    }

    // get file options:
    int options;
    String recordEncName, sizeEncName;
    {
      DataInput dataIn = new java.io.DataInputStream(_input);

      options       = dataIn.readInt();
      sizeEncName   = dataIn.readUTF(); 
      recordEncName = dataIn.readUTF(); 
    }

    _encFactory = new EncoderFactory();

    _recordEncoder = _encFactory.newRecordEncoder(recordEncName);
    _recordContext = new RecordContext();

    _sizeEncoder = _encFactory.newEncoder(sizeEncName, StepInt.TYPE);
    _sizeContext = new SizeContext();

    if ( (options & StepEncodedOutput.COMPRESSED_DATA) != 0 )
    {
      if ( verbose() ) System.out.println("STEP: data is compressed");
      _input = new java.util.zip.GZIPInputStream(_input, zipBufferSize());
    }
  }

  // NOTE: Can't distinguish between auto and user MetaRecords
  //   => metaData must be MetaEvent records!
  //
  // NOTE: Discarded records may also discard meta-info
  //   => can't do otherwise (eg. EncoderEvent needs target...)
  //   => user meta may get corrupted, internal should be OK.
  public StepRecord readRecord() throws IOException, ClassNotFoundException
  {
    // NOTE:
    // We loop until a record is available (vs. recurse) because there may be
    // long sequences of unknown records and we don't want a stack overflow.
    do
    {
      StepInt size;
      try
      {
        size = (StepInt) _sizeEncoder.decode(_sizeContext);
      }
      catch (java.io.EOFException e) // => OK, end of input
      {
        return null;
      }
      catch (EncoderException e)
      {
        throw new RuntimeException(e.getSource().getName()+": "+e.getMessage(), e);
      }

      byte[] recordBytes = ByteArray.allocator().alloc(size.intValue());
      {
        if ( _input.read(recordBytes) < recordBytes.length )
          throw new java.io.EOFException("unexpected end of input");

        _recordContext.setData(recordBytes);

        try
        {
          return doReadRecord(_recordContext);
        }
        catch (RecordDefNotFoundException e)
        {
          if ( ! ignoreUnknownRecords() )
            throw e;

          // else, loop again...
        }
      }
      ByteArray.allocator().free(recordBytes);
    }
    while (true);
  }

  private StepRecord doReadRecord(DecodeContext context) throws IOException
  {
    StepRecord record;
    try
    {
      record = (StepRecord) _recordEncoder.decode(context);
    }
    catch (EncoderException e)
    {
      throw new RuntimeException(e.getSource().getName()+": "+e.getMessage(), e);
    }

    if ( record instanceof MetaRecord )
    {
      MetaRecord metaRecord = (MetaRecord) record;
      {
        record = doReadRecord(new MetaContext(metaRecord));
      }
      metaRecord.free();
    }

    return record;
  }

  public void close() throws IOException
  {
    _input.close();
  }

  static boolean ignoreUnknownRecords()
  {
    return Boolean.getBoolean("step.ignoreUnknownRecords");
  }

  public class SizeContext implements DecodeContext
  {
    public Object read(Object request) throws IOException //: DecodeContext
    {
      byte[] bytes = (byte[]) request;

      long count = _input.read(bytes);

      if ( count < bytes.length )
        throw new java.io.EOFException("unexpected end of input");

      return bytes;
    }
  }

  public class RecordContext implements DecodeContext
  {
    private byte[] _dataBytes;
    private int    _dataPos;

    public RecordContext()
    {
      _dataBytes = null;
      _dataPos   = -1;
    }

    public void setData(byte[] dataBytes)
    {
      assert (_dataPos == -1) || (_dataPos == _dataBytes.length);

      _dataBytes = dataBytes;
      _dataPos   = 0;
    }

    public Object read(Object request) throws IOException //: DecodeContext
    {
      byte[] bytes = (byte[]) request;

      for (int i=0; i<bytes.length; i++)
        bytes[i] = _dataBytes[_dataPos++];

      if ( _dataPos >= _dataBytes.length ) // TODO: not necessary
      {
        _dataBytes = null;
        _dataPos   = -1;
      }

      return bytes;
    }

    private void dump()
    {
      System.out.println("RecordContext: "+Hex.toString(_dataBytes)+" @ "+_dataPos);
    }
  }

  public class MetaContext implements DecodeContext
  {
    private RecordContext _metaContext;
    private Iterator      _iData, _iMeta;
    private byte[]        _dataBytes;
    private int           _dataPos;

    public MetaContext(MetaRecord metaRecord) throws IOException
    {
      _metaContext = new RecordContext();

      List data = new java.util.ArrayList(metaRecord.data.length());

      for (int i=0; i<metaRecord.data.length(); i++)
        data.add( ((StepData) metaRecord.data.getElements()[i]).getData() );

      _iData = data.iterator();


      List meta = new java.util.ArrayList(metaRecord.metaData.length());

      for (int i=0; i<metaRecord.metaData.length(); i++)
        meta.add( ((StepData) metaRecord.metaData.getElements()[i]).getData() );

      _iMeta = meta.iterator();

      assert data.size() == (meta.size() + 1);

      _dataBytes = (byte[]) _iData.next();
      _dataPos   = 0;

      while ( _dataPos >= _dataBytes.length )
      {
        assert _iMeta.hasNext();

        metaRead();

        _dataBytes = (byte[]) _iData.next();
        _dataPos   = 0;
      }
    }

    private void metaRead() throws IOException
    {
      _metaContext.setData((byte[]) _iMeta.next());

      MetaEvent event = (MetaEvent) doReadRecord(_metaContext);

      event.apply(_encFactory);
    }

    public Object read(Object request) throws IOException //: DecodeContext
    {
      byte[] bytes = (byte[]) request;

      for (int i=0; i<bytes.length; i++)
      {
        while ( _dataPos >= _dataBytes.length )
        {
          assert _iMeta.hasNext();

          metaRead();

          _dataBytes = (byte[]) _iData.next();
          _dataPos   = 0;
        }

        bytes[i] = _dataBytes[_dataPos++];
      }

      while ( _dataPos >= _dataBytes.length )
      {
        if ( _iMeta.hasNext() )
        {
          metaRead();

          _dataBytes = (byte[]) _iData.next();
          _dataPos   = 0;
        }
        else
        {
          // TODO: not necessary
          {
            _metaContext = null;
            _iData = null;
            _iMeta = null;
            _dataBytes = null;
            _dataPos = -1;
          }
          break;
        }
      }

      return bytes;
    }

    private void dump()
    {
      System.out.println("MetaContext: "+Hex.toString(_dataBytes)+" @ "+_dataPos);
    }
  }

  static boolean verbose()
  {
    return Boolean.getBoolean("step.verbose");
  }

  static int zipBufferSize()
  {
    return Integer.parseInt(System.getProperty("step.zipBufferSize", "65536"));
  }
}
