/* STEP - An Framework for the Efficient Encoding of Trace Data
 * Copyright (C) 2002 Rhodes Brown
 *
 * STEP is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * STEP is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with STEP; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA
 */

/* STEP is distributed through the Sable Research Group at:
 * http://www.sable.mcgill.ca/step/
 */

/* Reference Version: $ProjectVersion: 0.9.2.dev.36 $ */

package step.io;

import java.io.OutputStream; import java.io.DataOutput;
import java.io.IOException;

import java.util.Map;
import java.util.List;
import java.util.Iterator;

import step.*;
import step.encoding.*;
import step.util.*;

/**
 * An output stream for writing {@link StepRecord} objects.
 *
 * <h2>File Format</h2>
 * The file consists of a series of <code>[size][record]</code> elements.
 * Records are encoded with a universal {@link RecordEncoder}.
 * The size (in bytes) of the record data precedes the data itself, which
 * permits a {@linkplain StepEncodedInput reader} to skip the record if it is
 * unwanted or its decoder is unavailable.
 *
 * <p>
 * NOTE: Some records are {@link MetaRecord}s, which contain a segmented record
 * along with interpretive meta-data. The format of a MetaRecord is:
 * <code>[data1][data2][data3]...[meta1][meta2]...</code>, where the "meta"
 * portions are {@link MetaEvent} records. MetaEvents indicate some form of
 * state change in the encoding/decoding process. MetaRecords are unpacked by
 * decoding a data segment, then interpreting the next MetaEvent. There are
 * always n-1 MetaEvents for n data segments.
 * <p>
 * NOTE: This class is not synchronized. When used in a multi-threaded
 * environment, the write process must be externally synchronized.
 * <em>Failure to do so may result in corrupted data.</em>
 *
 * @see StepRecord
 * @see Encoder
 * @see StepEncodedInput
 */
// TODO: header format
public class StepEncodedOutput implements StepRecordOutput
{
  public static final byte[] FILE_MAGIC   = { 0x53, 0x54, 0x45, 0x50 }; // "STEP"

  /* $Format: "  public static final byte[] FILE_VERSION = { $StepMajorVersion$, $StepMinorVersion$, $StepMicroVersion$ };"$ */
  public static final byte[] FILE_VERSION = { 0, 9, 2 };

  public static final String SIZE_ENCODER_NAME   = "$size";
  public static final String RECORD_ENCODER_NAME = "$record";

  // File options:
  static final int COMPRESSED_DATA = 0x00000001;


  private OutputStream _output;

  private RecordContext _recordContext;
  private Encoder       _recordEncoder;
      
  private SizeContext _sizeContext;
  private Encoder     _sizeEncoder;

  private final StepInt _size = new StepInt(0); // avoid unnecessary allocation

  /**
   */
  public StepEncodedOutput(OutputStream output) throws IOException
  {
    EncoderFactory encFactory = new EncoderFactory();

    _recordEncoder = encFactory.newRecordEncoder(RECORD_ENCODER_NAME);
    _recordContext = new RecordContext(0);

    _sizeEncoder = encFactory.newEncoder(SIZE_ENCODER_NAME, StepInt.TYPE);
    _sizeContext = new SizeContext();

    _output = output;
    _output.write(FILE_MAGIC);
    _output.write(FILE_VERSION);

    int options = 0x00000000;
    if ( zipData() ) options |= COMPRESSED_DATA;

    // Write file header:
    {
      DataOutput dataOut = new java.io.DataOutputStream(_output);

      // file options
      if ( verbose() ) System.out.println("STEP: options = "+Hex.toString(options));
      dataOut.writeInt(options);

      // top-level encoder names
      dataOut.writeUTF(_sizeEncoder.getName());
      dataOut.writeUTF(_recordEncoder.getName());
    }

    if ( (options & COMPRESSED_DATA) != 0 )
    {
      if ( verbose() ) System.out.println("STEP: data is compressed");
      _output = new java.util.zip.GZIPOutputStream(_output, zipBufferSize());
    }
  }

  /**
   * NOTE: This operation is not synchronized. When used in a multi-threaded
   * environment, the write process must be externally synchronized on this
   * object. Failure to do so may result in corrupted data.
   */
  // NOTE: We invoke this method recursively to write meta data so as to ensure
  // that all records (plain & meta-nested) are encoded in the same format,
  // using the same tags, etc.
  public void writeRecord(StepRecord record) throws IOException
  {
    _recordContext.encode(record);

    if ( _recordContext.writeLevel() == 0 )
    {
      byte[] recordBytes = _recordContext.close();

      _size.setValue(recordBytes.length);
      try
      {
        _sizeEncoder.encode(_size, _sizeContext);
      }
      catch (EncoderException e)
      {
        throw new RuntimeException(e.getSource().getName()+": "+e.getMessage(), e);
      }
      byte[] sizeBytes = _sizeContext.close();

      _output.write(sizeBytes);   ByteArray.allocator().free(sizeBytes);
      _output.write(recordBytes); ByteArray.allocator().free(recordBytes);
    }
  }

  public void close() throws IOException
  {
    _output.close();
  }

  public void flush() throws IOException
  {
    _output.flush();
  }

  public class RecordContext implements EncodeContext
  {
    private int           _writeLevel;
    private RecordContext _metaContext; // lazy create
    private boolean       _metaWrite;
    private List          _data, _meta;
    private byte[]        _dataBytes;
    private int           _dataPos;

    private RecordContext(int writeLevel)
    {
      assert writeLevel >= 0;

      _writeLevel  = writeLevel;
      _metaContext = null;
      _metaWrite   = false;

      _data = new java.util.ArrayList();
      _meta = new java.util.ArrayList();

      _dataBytes = ByteArray.allocator().alloc(256);
      _dataPos   = 0;
    }

    public void encode(StepRecord record) throws IOException
    {
      if ( _metaWrite ) { metaContext().encode(record); return; }

      assert _data.size() == 0;
      assert _meta.size() == 0;
      assert _dataPos     == 0;

      try
      {
        _recordEncoder.encode(record, _recordContext); // NOTE: encode to top context
      }
      catch (EncoderException e)
      {
        throw new RuntimeException(e.getSource().getName()+": "+e.getMessage(), e);
      }

      _data.add(ByteArray.dup(_dataBytes, 0, _dataPos));
      _dataPos = 0;

      assert _data.size() == (_meta.size() + 1);
    }

    /**
     * @throws ClassCastException if the object is not a <code>byte[]</code>.
     * @throws IOException if an I/O error occurred writing the object.
     */
    public void write(Object o) throws IOException //: EncodeContext
    {
      if ( _metaWrite ) { metaContext().write(o); return; }

      byte[] bytes = (byte[]) o;

      for (int i=0; i<bytes.length; i++)
      {
        try
        {
          _dataBytes[_dataPos++] = bytes[i];
        }
        catch (ArrayIndexOutOfBoundsException e)
        {
          assert _dataPos >= _dataBytes.length;

          byte[] newDataBytes = ByteArray.allocator().alloc(3*(_dataBytes.length+1)/2);
            // => grow by 50%

          for (int j=0; j<_dataBytes.length; j++)
            newDataBytes[j] = _dataBytes[j];

          i--;                          // backup 1
          _dataPos = _dataBytes.length; // reset

          ByteArray.allocator().free(_dataBytes);

          _dataBytes = newDataBytes;
        }
      }
    }

    /**
     * @throws ClassCastException if the event is not a {@link StepRecord}.
     * @throws IOException if an I/O error occurred writing the event.
     */
    public void metaWrite(MetaEvent event) throws IOException //: EncodeContext
    {
      if ( _metaWrite ) { metaContext().metaWrite(event); return; }

      assert _data.size() == _meta.size();

      _data.add(ByteArray.dup(_dataBytes, 0, _dataPos));
      _dataPos = 0;

      _metaWrite = true;
      {
        writeRecord((StepRecord) event);
        _meta.add(metaContext().close());
      }
      _metaWrite = false;
      
      assert _data.size() == _meta.size();
    }

    /**
     * @throws IOException if an I/O error occurred writing a {@link MetaRecord}.
     */
    public byte[] close() throws IOException
    {
      if ( _metaWrite ) return metaContext().close();

      assert _data.size() > 0;

      if ( (_data.size() == 1) )
      {
        assert _meta.size() == 0;

        byte[] bytes = (byte[]) _data.get(0);

        _data.clear();

        return bytes;
      }
      else // need a meta record!
      {
        assert _data.size() == (_meta.size()+1);

        MetaRecord metaRecord = new MetaRecord(_data, _meta);

        _meta.clear();
        _data.clear();

        _writeLevel++; // temporarily inflate level so writeRecord() won't close()
        {
          writeRecord(metaRecord);
        }
        _writeLevel--;

        metaRecord.free(); // free resources used by the meta record

        return close();
      }
    }

    public int writeLevel()
    {
      if ( _metaWrite ) return metaContext().writeLevel();

      return _writeLevel;
    }

    private RecordContext metaContext()
    {
      if ( _metaContext == null )
        _metaContext = new RecordContext(_writeLevel+1);

      return _metaContext;
    }
  }

  public class SizeContext implements EncodeContext
  {
    private byte[] _dataBytes;
    private int    _dataPos;

    public SizeContext()
    {
      _dataBytes = new byte[10];
      _dataPos   = 0;
    }

    public void write(Object o) throws IOException //: EncodeContext
    {
      byte[] bytes = (byte[]) o;

      for (int i=0; i<bytes.length; i++)
        _dataBytes[_dataPos++] = bytes[i];
    }

    public byte[] close()
    {
      byte[] bytes = ByteArray.dup(_dataBytes, 0, _dataPos);

      _dataPos = 0;

      return bytes;
    }

    /**
     * @throws UnsupportedOperationException always
     */
    public void metaWrite(MetaEvent event) throws IOException //: EncodeContext
    {
      throw new UnsupportedOperationException("size encoding should not generate meta-data");
    }
  }

  static boolean verbose()
  {
    return Boolean.getBoolean("step.verbose");
  }

  static boolean zipData()
  {
    return Boolean.valueOf(System.getProperty("step.zipData", "false")).booleanValue();
  }

  static int zipBufferSize()
  {
    return Integer.parseInt(System.getProperty("step.zipBufferSize", "65536"));
  }
}
