/* STEP - An Framework for the Efficient Encoding of Trace Data
 * Copyright (C) 2002 Rhodes Brown
 *
 * STEP is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * STEP is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with STEP; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA
 */

/* STEP is distributed through the Sable Research Group at:
 * http://www.sable.mcgill.ca/step/
 */

/* Reference Version: $ProjectVersion: 0.9.2.dev.36 $ */

package step.typedef;

import java.util.List;
import java.util.Iterator;

import step.util.StringConstant;

/**
 * Abstraction of a STEP attribute.
 * <p>
 * Attribute values may contain arbitrary string values, however, the generally
 * contain a comma separated list of items.
 * Each attribute is assigned to a particular group, which indicates its
 * applicability.
 */
public class Attribute implements StepDef
{
  private String _group;
  private String _value;

  /** Create a new attribute for the given group, with the given value. */
  public Attribute(String group, String value)
  {
    _group = group; assert _group != null;
    _value = value; assert _value != null;
  }

  /** Get the group for this attribute. */
  public String getGroup()
  {
    return _group;
  }

  /** Get the value of this attribute. */
  public String getValue()
  {
    return _value;
  }

  /** Get this attribute as STEP-DL text. */
  public String toStepDL() //: StepDef
  {
    return "<"+_group+":"+StringConstant.encode(_value)+">";
  }

  /** Compare this attribute to another for equality. */
  public boolean equals(Object o)
  {
    try
    {
      Attribute rhs = (Attribute) o;
      return getGroup().equals(rhs.getGroup()) && getValue().equals(rhs.getValue());
    }
    catch (ClassCastException e)
    {
      return false;
    }
    catch (NullPointerException e)
    {
      return false;
    }
  }

  /** Get the hash code for this attribute. */
  public int hashCode()
  {
    return getGroup().hashCode() + getValue().hashCode();
  }

  /**
   * Combine two attribute lists, based on priority.
   *
   * <h3>Algorithm:</h3>
   * <pre>
   * while low[i] == high[i]
   *   add low[i]
   *
   * add remaining low
   *
   * add remaining high
   * </pre>
   * <p>
   * This algorithm follows the general property that items in an attribute list
   * are processed from first (lowest priority) to last (highest priority).
   *
   * @param lowList  the list of lower priority attributes
   * @param highList the list of higher priority attributes
   *
   * @return a new list that is the "merge" of the two input lists.
   */
  public static List merge(List lowList, List highList)
  {
    List newList = new java.util.ArrayList(lowList.size()+highList.size());

    Iterator iLow  = lowList.iterator();
    Iterator iHigh = highList.iterator();

    while (iLow.hasNext() && iHigh.hasNext())
    {
      Attribute low  = (Attribute) iLow.next();
      Attribute high = (Attribute) iHigh.next();

      if ( low.equals(high) )
      {
        newList.add(low);
      }
      else
      {
        newList.add(low);
        while (iLow.hasNext())
          newList.add((Attribute) iLow.next());

        newList.add(high);
        while (iHigh.hasNext())
          newList.add((Attribute) iHigh.next());
      }
    }

    while (iLow.hasNext())
      newList.add((Attribute) iLow.next());

    while (iHigh.hasNext())
      newList.add((Attribute) iHigh.next());

    return newList;
  }
}
