/* STEP - An Framework for the Efficient Encoding of Trace Data
 * Copyright (C) 2002 Rhodes Brown
 *
 * STEP is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * STEP is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with STEP; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA
 */

/* STEP is distributed through the Sable Research Group at:
 * http://www.sable.mcgill.ca/step/
 */

/* Reference Version: $ProjectVersion: 0.9.2.dev.36 $ */

package step.typedef;

import java.util.List;
import java.util.Iterator;
import java.util.Collections;
import java.util.Stack;
import java.util.Map;

import step.util.StringConstant;

/**
 * The definition of a derived STEP record type.
 *
 * @see step.StepRecord
 * @see InheritedFieldDef
 */
public class DerivedRecordDef extends RecordDef
{
  private Type    _parent;
  private boolean _cleanInherit;
  private List    _allAttributes = null;  // non-null signals type info is resolved
  private boolean _resolveLock   = false; // detect circular inheritance

  // NOTE: RecordDef's _attributes field stores the extra attributes for this type.

  /**
   * Create a new derived record type definition.
   */
  public DerivedRecordDef(String name, Package pkg, Type parent, boolean cleanInherit)
  {
    super(name, pkg);

    setParent(parent);
    setCleanInherit(cleanInherit);
  }

  /** Get the parent type for this definition. */
  public Type getParent()
  {
    return _parent;
  }

  protected void setParent(Type parent)
  {
    _parent = parent; assert _parent != null;
  }

  /** Does this type discard attributes inherited from the parent? */
  public boolean isCleanInherit()
  {
    return _cleanInherit;
  }

  protected void setCleanInherit(boolean cleanInherit)
  {
    _cleanInherit = cleanInherit;
  }

  /** Get all the record attributes associated with this type. */
  public List getAttributes()
  {
    if ( ! isTypeResolved() )
      throw new IllegalStateException("type has not been resolved yet.");

    assert _allAttributes != null;

    return _allAttributes;
  }

  /**
   * Get only the record attributes added by this type, not the inherited
   * attributes.
   */
  public List getExtraAttributes()
  {
    if ( ! isTypeResolved() )
      throw new IllegalStateException("type has not been resolved yet.");

    return super.getAttributes(); // -> see NOTE above
  }

  /**
   * Is the type information for this definition complete?
   * @see #isResolved()
   */
  public boolean isTypeResolved()
  {
    return ( ( _allAttributes != null ) && super.isTypeResolved() );
  }

  /**
   * Complete the type information for this definition.
   * The purpose of this method is to incorporate inherited information
   * from the parent type.
   *
   * @throws ResolveException if
   * <ul>
   *   <li> a circular inheritance relationship was detected
   *   <li> the parent type is undefined
   *   <li> the parent type is not a record type
   *   <li> a field name conflicts with one inherited from the parent
   * </ul>
   *
   * @see #resolve()
   */
  protected void resolveType()
  {
    if ( isTypeResolved() ) return;

    if ( _resolveLock )
    {
      throw new ResolveException(this,
        "circular inheritance; \""+getFullName()+"\" derives from itself");
    }
    _resolveLock = true;


    RecordDef parentDef;
    try
    {
      try
      {
        parentDef = (RecordDef) TypeDef.getDefault(getParent());
      }
      catch (RecordDefNotFoundException e)
      {
        throw new ResolveException(this,
          "parent type \""+getParent()+"\" is undefined");
      }
    }
    catch (ClassCastException e) // TODO: could be a warning
    {
      throw new ResolveException(this,
        "parent type \""+getParent()+"\" is not a record type");
    }

    parentDef.resolveType();

    // update the field list
    {
      Map fields = new java.util.LinkedHashMap();

      // inherit the parent's fields
      {
        Iterator iParentFields = parentDef.getFields().iterator();
        while (iParentFields.hasNext())
        {
          FieldDef fieldDef = (FieldDef) iParentFields.next();
          fields.put(fieldDef.getName(), new InheritedFieldDef(fieldDef));
        }
      }

      // add the local fields
      {
        Iterator iLocalFields = getFields().iterator();
        while (iLocalFields.hasNext())
        {
          FieldDef fieldDef = (FieldDef) iLocalFields.next();

          if ( fields.containsKey(fieldDef.getName()) )
          {
            throw new ResolveException(this,
              "redefinition of field \""+fieldDef.getName()+"\"");
          }

          fields.put(fieldDef.getName(), fieldDef);
        }
      }

      setFields(new java.util.ArrayList(fields.values()));
    }


    // inherit the parent's attributes
    {
      List newList;

      if ( isCleanInherit() )
      {
        newList = new java.util.ArrayList(super.getAttributes());
      }
      else
      {
        newList = new java.util.ArrayList(parentDef.getAttributes()); // parent's
        newList.addAll(super.getAttributes());                        // local
      }

      _allAttributes = Collections.unmodifiableList(newList);
    }

    super.resolveType(); // complete the default resolution...
  }

  public Object clone()
  {
    RecordDef recordDef = (RecordDef) super.clone();

    DerivedRecordDef def =
      new DerivedRecordDef(getName(), getPackage(), getParent(), isCleanInherit());

    def.setLabel(getLabel());
    def.setDescription(getDescription());
    def._allAttributes = _allAttributes;
    def.setAttributes(recordDef.getAttributes());
    def.setFields(recordDef.getFields());
    def.setModifiers(recordDef.getModifiers());

    if ( getFactory() != null )
      def.setFactory(getFactory());

    return def;
  }

  /**
   * Get a compact string version of this definition as STEP-DL.
   * The definition includes package definitions, thus it is complete STEP
   * definition for the type.
   */
  public String toStepDL() //: StepDef
  {
    if ( ! isResolved() ) resolve();

    StringBuffer buf = new StringBuffer();

    int pkgScope;
    {
      Stack pkgs = new Stack();
      for (Package pkg = getPackage(); !pkg.equals(Package.ROOT); pkg = pkg.getParent())
        pkgs.add(pkg);

      pkgScope = pkgs.size();

      while (pkgs.size() > 0)
        buf.append("package ").append(((Package) pkgs.pop()).getName()).append('{');
    }

    buf.append("record ").append(getName());

    if ( ! getLabel().equals(getName()) )
      buf.append(StringConstant.encode(getLabel()));

    buf.append("extends ").append(isCleanInherit() ? "!" : "").append(getParent());

    buf.append('{');

    if ( ! getDescription().equals("") )
      buf.append(StringConstant.encode(getDescription()));

    {
      Iterator iAttributes = getExtraAttributes().iterator();
      while (iAttributes.hasNext())
        buf.append(((Attribute) iAttributes.next()).toStepDL());
    }

    {
      Iterator iFieldDefs = getFields().iterator();
      while (iFieldDefs.hasNext())
      {
        FieldDef fieldDef = (FieldDef) iFieldDefs.next();
        if ( (fieldDef instanceof LocalFieldDef) )
          buf.append(((LocalFieldDef) fieldDef).toStepDL());
      }
    }

    {
      Iterator iModifiers = getModifiers().iterator();
      while (iModifiers.hasNext())
      {
        FieldModifier modifier = (FieldModifier) iModifiers.next();

        if ( getField(modifier.getTarget()) instanceof LocalFieldDef )
          continue;

        buf.append(modifier.toStepDL());
      }
    }

    buf.append('}');

    for (int i=0; i<pkgScope; i++)
      buf.append('}');

    return buf.toString();
  }
}
