/* STEP - An Framework for the Efficient Encoding of Trace Data
 * Copyright (C) 2002 Rhodes Brown
 *
 * STEP is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * STEP is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with STEP; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA
 */

/* STEP is distributed through the Sable Research Group at:
 * http://www.sable.mcgill.ca/step/
 */

/* Reference Version: $ProjectVersion: 0.9.2.dev.36 $ */

package step.typedef;

import java.util.List;
import java.util.Iterator;
import java.util.Collections;

/**
 * Abstraction of a STEP field defintion.
 * <p>
 * Field definitions are a component of type definitions.
 *
 * @see TypeDef
 */
// NOTE:
// An inherited field definition decorates another field definition,
// hence the need for an abstract base class.

public abstract class FieldDef implements Cloneable
{
  private List _attributes = Collections.unmodifiableList(new java.util.LinkedList());

  protected FieldDef()
  {
  }

  /** Get the name of the field. */
  public abstract String getName();

  /** Set the name of the field. */
  protected abstract void setName(String name);

  /** Get the type of the field. */
  public abstract Type getType();

  /** Set the type of the field. */
  protected abstract void setType(Type type);

  /** Get the description of the field. */
  public abstract String getDescription();

  /** Set the description of the field. */
  protected abstract void setDescription(String description);

  /** Get the field attributes associated with this definition. */
  public List getAttributes()
  {
    return _attributes;
  }

  /**
   * Set the field attributes for this definition.
   * <p>
   * NOTE: The given list is copied and checked to make sure the elements are in
   * fact {@link Attribute}s.
   */
  protected void setAttributes(List attributes)
  {
    List newList = new java.util.ArrayList(attributes.size());
    Iterator iAttributes = attributes.iterator();
    while (iAttributes.hasNext())
      newList.add((Attribute) iAttributes.next());

    _attributes = Collections.unmodifiableList(newList);
  }

  /**
   * Get an alternate type definition specific to this field.
   * The alternate definition is normally used to indicate that subfields 
   * of this field have extended/overridden attributes.
   */
  public abstract TypeDef getAltTypeDef();

  /** Set a new, alternate, definition for the type of this field. */
  protected abstract void setAltTypeDef(TypeDef altDef);

  /** Does this field have an alternate type definition? */
  public boolean hasAltTypeDef()
  {
    return ( getAltTypeDef() != null );
  }

  /** Get a copy (clone) of this definition. */
  public abstract Object clone();

  public static FieldDef merge(FieldDef lowDef, FieldDef highDef)
  {
    if ( (!lowDef.getName().equals(highDef.getName())) ||
         (!lowDef.getType().equals(highDef.getType())) )
    {
      throw new IllegalArgumentException("incompatible field definitions");
    }

    String name = lowDef.getName();
    Type   type = lowDef.getType();
    String desc = lowDef.getDescription();

    List attrs = Attribute.merge(lowDef.getAttributes(), highDef.getAttributes());

    TypeDef altTypeDef = null;

    if ( lowDef.hasAltTypeDef() && highDef.hasAltTypeDef() )
    {
      altTypeDef = TypeDef.merge(lowDef.getAltTypeDef(), highDef.getAltTypeDef());
    }
    else if ( lowDef.hasAltTypeDef() )
    {
      altTypeDef = lowDef.getAltTypeDef();
    }
    else if ( highDef.hasAltTypeDef() )
    {
      altTypeDef = highDef.getAltTypeDef();
    }
    // else => do nothing, use null


    return new ArtificialFieldDef(name, type, desc, attrs, altTypeDef);
  }

  // -- Field Definition Builder -----------------------------------------------

  private static Builder _Builder;

  /** Get the builder object for creating field definitions. */
  public static Builder builder()
  {
    if ( _Builder == null )
      _Builder = new Builder();  // lazy create

    return _Builder;
  }

  /**
   * Create field definitions with the <em>Builder</em> pattern.
   */
  public static class Builder
  {
    private boolean  _building;
    private String   _name;
    private Type     _type;
    private String   _description;
    private List     _attributes;
    private TypeDef  _altTypeDef;

    private Builder()
    {
      _building   = false;
      _attributes = new java.util.ArrayList();
    }

    /**
     * Start building a new field definition with the given name and type name.
     * @see #newFieldDef(String,Type)
     */
    public Builder newFieldDef(String name, String typeName)
    {
      return newFieldDef(name, Type.v(typeName));
    }

    /**
     * Start building a new field definition with the given name and type.
     * <p>
     * The default record description is the empty string ("").
     *
     * @param name the brief name of the new field.
     * @param type the type of the field.
     *
     * @return the current builder.
     *
     * @throws IllegalStateException if already building a definition.
     * @throws IllegalArgumentException <code>name</code> or <code>type</code>
     *   are <code>null</code>.
     */
    public Builder newFieldDef(String name, Type type)
    {
      if ( _building )
        throw new IllegalStateException("already building a definition");

      if ( name == null )
        throw new IllegalArgumentException("name is null");

      if ( type == null )
        throw new IllegalArgumentException("type is null");

      _building    = true;
      _name        = name;
      _type        = type;
      _description = "";

      return this;
    }

    /**
     * Set the description for the definition.
     * @return the current builder.
     * @throws IllegalStateException if not building a definition.
     * @throws IllegalArgumentException <code>description</code> is <code>null</code>.
     */
    public Builder setDescription(String description)
    {
      if ( ! _building )
        throw new IllegalStateException("not building a definition");

      if ( description == null )
        throw new IllegalArgumentException("description is null");

      _description = description;

      return this;
    }

    /**
     * Get the description for the definition.
     * @return the description,
     *   or <code>null</code> if no description is currently defined.
     */
    public String getDescription()
    {
      return _description;
    }

    /**
     * Add an attribute to the definition.
     * @return the current builder.
     * @throws IllegalStateException if not building a definition.
     * @throws IllegalArgumentException <code>attribute</code> is <code>null</code>.
     */
    public Builder addAttribute(Attribute attribute)
    {
      if ( ! _building )
        throw new IllegalStateException("not building a definition");

      if ( attribute == null )
        throw new IllegalArgumentException("attribute is null");

      _attributes.add(attribute);
      
      return this;
    }

    public Builder addAttributes(List attributes)
    {
      if ( ! _building )
        throw new IllegalStateException("not building a definition");

      Iterator iAttributes = attributes.iterator();
      while (iAttributes.hasNext())
        addAttribute((Attribute) iAttributes.next());

      return this;
    }

    /**
     * Get the list of attributes for the definition.
     * @return the current list of attributes, which may be empty.
     */
    public List getAttributes()
    {
      return _attributes;
    }

    /**
     * Set an alternate type definition for the field.
     * @return the current builder.
     * @throws IllegalStateException if not building a definition.
     * @throws IllegalArgumentException <code>altDef</code> is for a
     *   different type than the current one.
     */
    public Builder setAltTypeDef(TypeDef altDef)
    {
      if ( ! _building )
        throw new IllegalStateException("not building a definition");

      if ( !altDef.getType().equals(_type) )
        throw new IllegalArgumentException("altDef specifies a different type");

      _altTypeDef = altDef;

      return this;
    }

    public TypeDef getAltTypeDef()
    {
      return _altTypeDef;
    }

    /**
     * Obtain the completed field definition.
     * @return the fully constructed field definition.
     * @throws IllegalStateException if not building a definition.
     */
    public FieldDef makeFieldDef()
    {
      if ( ! _building )
        throw new IllegalStateException("not building a definition");

      FieldDef fieldDef = new LocalFieldDef(_name, _type);
      fieldDef.setDescription(_description);
      fieldDef.setAttributes(_attributes);
      if ( _altTypeDef != null ) fieldDef.setAltTypeDef(_altTypeDef);

      clear();

      return fieldDef;
    }

    /** Reset the build context. */
    public void clear()
    {
      _building    = false;
      _name        = null;
      _type        = null;
      _description = null;
      _altTypeDef  = null;

      _attributes.clear();
    }
  }
}
