/* STEP - An Framework for the Efficient Encoding of Trace Data
 * Copyright (C) 2002 Rhodes Brown
 *
 * STEP is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * STEP is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with STEP; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA
 */

/* STEP is distributed through the Sable Research Group at:
 * http://www.sable.mcgill.ca/step/
 */

/* Reference Version: $ProjectVersion: 0.9.2.dev.36 $ */

package step.typedef;

import java.util.List;
import java.util.Iterator;

/**
 * Abstraction of a STEP field modifier.
 * <p>
 * NOTE: Field modifiers are only meaningful before a type definition has been
 *       resolved.
 *
 * @see FieldDef
 * @see TypeDef#resolve()
 */
public class FieldModifier implements StepDef, Cloneable
{
  private String  _target;
  private boolean _clean;
  private List    _attributes = new java.util.LinkedList();

  /**
   * Create a new field modifier.
   * @param target the field to apply this modifier to.
   *   NOTE: The target may be a qualified name.
   * @param clean override (discard) previous attributes?
   */
  public FieldModifier(String target, boolean clean)
  {
    _target = target; assert _target != null;
    _clean  = clean;
  }

  /**
   * Get the name of the target field for this modifier.
   * <p>
   * NOTE: The target may be a qualified name.
   */
  public String getTarget()
  {
    return _target;
  }

  /**
   * Does this modifier cause previous (including inherited)
   * attributes to be discarded?
   */
  public boolean isClean()
  {
    return _clean;
  }

  /** Get the attributes held by this modifier. */
  public List getAttributes()
  {
    return _attributes;
  }

  /**
   * Add an attribute to this modifier.
   * <p>
   * The current modifier is returned by this method to allow chaining of
   * multiple calls or immediate addition to a newly created instance.
   * For example,
   * <pre>FieldModifier m = new FieldModifier(..).addAttribute(..)...</pre>
   *
   * @return this object.
   */
  public FieldModifier addAttribute(Attribute attribute)
  {
    assert attribute != null;

    _attributes.add(attribute);

    return this;
  }

  /** Get a copy of this modifier. */
  public Object clone() //: Cloneable
  {
    FieldModifier modifier = new FieldModifier(getTarget(), isClean());

    Iterator iAttributes = getAttributes().iterator();
    while (iAttributes.hasNext())
      modifier.addAttribute((Attribute) iAttributes.next());

    return modifier;
  }

  /** Get this modifier as STEP-DL text. */
  public String toStepDL()
  {
    StringBuffer buf = new StringBuffer();

    buf.append( isClean() ? '!' : '~' );
    buf.append( getTarget() ).append( ' ' );

    Iterator iAttributes = getAttributes().iterator();
    while (iAttributes.hasNext())
      buf.append(((Attribute) iAttributes.next()).toStepDL());

    buf.append( ';' );

    return buf.toString();
  }
}
