/* STEP - An Framework for the Efficient Encoding of Trace Data
 * Copyright (C) 2002 Rhodes Brown
 *
 * STEP is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * STEP is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with STEP; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA
 */

/* STEP is distributed through the Sable Research Group at:
 * http://www.sable.mcgill.ca/step/
 */

/* Reference Version: $ProjectVersion: 0.9.2.dev.36 $ */

package step.typedef;

import java.util.Map;

/**
 * A STEP package.
 * <p>
 * Packages are unique objects.
 */
public class Package
{
  /** The name separator for nested packages. */
  public static final String SEPARATOR = ".";

  /** The "root" (global) package scope. */
  public static final Package ROOT;

  private static Map _NameToPackage;

  // Ensure proper initialization order.
  static
  {
    _NameToPackage = new java.util.Hashtable();
    _NameToPackage.put("", ROOT = new Package("", null));
  }

  private String  _name;
  private Package _parent;

  private Package(String name, Package parent) // => v() enforces uniqueness
  {
    _name   = name; assert _name != null;
    _parent = parent;
  }

  /**
   * Get the unique package with the given name and parent.
   * If the requested package does not exist, it is created.
   *
   * @param name the name of the package, which may be relative.
   *   Passing "x.y.z", "parent" gives "parent.x.y.z".
   *
   * @param parent the package that contains the requested name.
   *   The parent may be the {@link #ROOT} package.
   */
  public static Package v(String name, Package parent)
  {
    String fullName = parent.qualify(name);

    Package pkg = (Package) _NameToPackage.get(fullName);

    if ( pkg == null )
    {
      int dotPos = name.lastIndexOf(SEPARATOR);

      if ( dotPos >= 0 ) // name == "x.y.z"
      {
        parent = v(name.substring(0, dotPos), parent); // parent <- "parent.x.y"
        name = name.substring(dotPos+1);               // name   <- "z"
      }

      _NameToPackage.put(fullName, pkg = new Package(name, parent));
    }
    
    return pkg;
  }

  /**
   * Get the unique package with the given name.
   * @see #v(String,Package)
   */
  public static Package v(String name)
  {
    return v(name, ROOT);
  }

  /**
   * Get the short name of this package.
   * @see #getFullName()
   */
  public String getName()
  {
    return _name;
  }

  /**
   * Get the fully-qualified name of this package.
   * The form is, for example, "grandparent.parent.name".
   * @see #getName()
   * @see #qualify(String)
   */
  public String getFullName()
  {
    return ( this == ROOT ) ? getName() : getParent().qualify(getName());
  }

  /**
   * Combine a package and a name to get a fully-qualified name.
   * @return "pkg.baseName", or "baseName" if pkg equals {@link #ROOT}.
   * @see #getFullName()
   */
  public String qualify(String baseName)
  {
    assert baseName != null;

    return ( this == ROOT ) ? baseName : (getFullName() + SEPARATOR + baseName);
  }

  /**
   * Get the parent of this package.
   * <p>
   * NOTE: The {@link #ROOT} package has no parent.
   *
   * @return the parent package, or <code>null</code> if there is no parent.
   */
  public Package getParent()
  {
    return _parent;
  }

  /**
   * Compare two packages for equality.
   * <p>
   * Two packages are equal if they are the <em>same</em> object.
   * This equality property is enforced by the creational pattern for packages.
   * @see #v(String,Package)
   */
  public boolean equals(Object o)
  {
    return ( o == this );
  }

  /** Get the hash code for this package. */
  public int hashCode()
  { 
    return ( this == ROOT ) ?
      getName().hashCode() :
      (getName().hashCode() + getParent().hashCode());
  }

  public String toString()
  {
    return getFullName();
  }
}
