/* STEP - An Framework for the Efficient Encoding of Trace Data
 * Copyright (C) 2002 Rhodes Brown
 *
 * STEP is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * STEP is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with STEP; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA
 */

/* STEP is distributed through the Sable Research Group at:
 * http://www.sable.mcgill.ca/step/
 */

/* Reference Version: $ProjectVersion: 0.9.2.dev.36 $ */

package step.typedef;

import java.util.List;
import java.util.Iterator;
import java.util.Stack;
import java.util.Collections;

import step.*;
import step.util.StringConstant;

/**
 * The definition of a STEP record type.
 *
 * @see StepRecord
 */
public class RecordDef extends TypeDef implements StepDef
{
  static
  {
    // NOTE: We can't do any initialization of StepRecord in StepRecord.java
    // since it is an interface. Thus we do the init here.

    StepRecord.TYPE.setParent(StepObject.TYPE);

    new TypeDef("StepRecord", Package.v("step"), null, null)
      .setAsDefault();
  }

  private String        _label;
  private String        _description;
  private List          _attributes;
  private RecordFactory _factory = null;

  public RecordDef(String name, Package pkg)
  {
    super(name, pkg, null, null);

    setLabel(name);
    setDescription("");

    _attributes = Collections.unmodifiableList(new java.util.LinkedList());
  }

  /** Get the label for this record definition. */
  public String getLabel()
  {
    return _label;
  }

  /** Set the label for this record definition. */
  protected void setLabel(String label)
  {
    _label = label; assert _label != null;
  }

  /** Get the description for this record definition. */
  public String getDescription()
  {
    return _description;
  }

  /** Set the description for this record definition. */
  protected void setDescription(String description)
  {
    _description = description; assert _description != null;
  }

  /** Get the record attributes for this definition. */
  public List getAttributes()
  {
    return _attributes;
  }

  protected void setAttributes(List attributes)
  {
    List newList = new java.util.ArrayList(attributes.size());
    Iterator iAttributes = attributes.iterator();
    while (iAttributes.hasNext())
      newList.add((Attribute) iAttributes.next());

    _attributes = Collections.unmodifiableList(newList);
  }

  /**
   * Get the factory for this record definition.
   * @return the factory, or <code>null</code> if no factory is defined.
   */
  public RecordFactory getFactory()
  {
    return _factory;
  }

  protected void setFactory(RecordFactory factory)
  {
    _factory = factory; assert _factory != null;
  }

  public Object clone()
  {
    TypeDef typeDef = (TypeDef) super.clone();

    RecordDef recordDef = new RecordDef(getName(), getPackage());
    recordDef.setLabel(getLabel());
    recordDef.setDescription(getDescription());
    recordDef.setAttributes(getAttributes());
    recordDef.setFields(typeDef.getFields());
    recordDef.setModifiers(typeDef.getModifiers());

    if ( getFactory() != null )
      recordDef.setFactory(getFactory());

    return recordDef;
  }

  /**
   * Get a compact string version of this definition as STEP-DL.
   * The definition includes package definitions, thus it is a complete STEP
   * definition for the type.
   */
  public String toStepDL() //: StepDef
  {
    if ( ! isResolved() ) resolve();

    StringBuffer buf = new StringBuffer();

    int pkgScope;
    {
      Stack pkgs = new Stack();
      for (Package pkg = getPackage(); !pkg.equals(Package.ROOT); pkg = pkg.getParent())
        pkgs.add(pkg);

      pkgScope = pkgs.size();

      while (pkgs.size() > 0)
        buf.append("package ").append(((Package) pkgs.pop()).getName()).append('{');
    }

    buf.append("record ").append(getName());

    if ( ! getLabel().equals(getName()) )
      buf.append(StringConstant.encode(getLabel()));

    buf.append('{');

    if ( ! getDescription().equals("") )
      buf.append(StringConstant.encode(getDescription()));

    {
      Iterator iAttributes = getAttributes().iterator();
      while (iAttributes.hasNext())
        buf.append(((Attribute) iAttributes.next()).toStepDL());
    }

    {
      Iterator iFieldDefs = getFields().iterator();
      while (iFieldDefs.hasNext())
        buf.append(((LocalFieldDef) iFieldDefs.next()).toStepDL());
    }

    buf.append('}');

    for (int i=0; i<pkgScope; i++)
      buf.append('}');

    return buf.toString();
  }

  /**
   * Attempt to load the definition for the given record name.
   * <p>
   * NOTE: It is assumed that the record type provides its definition in a
   * <code>DEFINITION</code> static field.
   *
   * @throws RecordDefNotFoundException if the named type could not be loaded.
   */
  public static RecordDef load(Type recordType)
  {
    RecordDef def = null;

    try
    {
      def = (RecordDef) Class.forName(
       recordType.toString()).getField("DEFINITION").get(null);

      if ( def instanceof DerivedRecordDef ) // => load parents
      {
        load(((DerivedRecordDef) def).getParent());
      }

      def.setAsDefault();

      return def;
    }
    catch (ClassNotFoundException e)   { } // .forName(String)
    catch (NoSuchFieldException e)     { } // .getField(String)
    catch (IllegalArgumentException e) { } // .get(Object)
    catch (IllegalAccessException e)   { } // .get(Object)

    throw new RecordDefNotFoundException("could not load definition for "+recordType);
  }


  // -- Record Definition Builder ---------------------------------------------

  private static Builder _Builder;

  /**
   * Get the builder object for creating new record definitions.
   */
  public static Builder builder()
  {
    if ( _Builder == null )
      _Builder = new Builder(); // lazy create

    return _Builder;
  }

  /**
   * Create record definitions with the <em>Builder</em> pattern.
   */
  public static class Builder
  {
    private boolean       _building;
    private String        _name;
    private Package       _pkg;
    private Type          _parent;
    private boolean       _cleanInherit;
    private String        _label;
    private String        _description;
    private List          _attributes;
    private List          _fieldDefs;
    private List          _modifiers;
    private RecordFactory _factory;

    private Builder()
    {
      _building = false;

      _attributes = new java.util.ArrayList();
      _fieldDefs  = new java.util.ArrayList();
      _modifiers  = new java.util.ArrayList();
    }

    /**
     * Start building a new record definition with the given name in the
     * {@link Package#ROOT ROOT} package.
     * @see #newRecordDef(String,Package)
     */
    public Builder newRecordDef(String name)
    {
      return newRecordDef(name, Package.ROOT);
    }

    /**
     * Start building a new record definition with the given name in the given
     * package name.
     * @see #newRecordDef(String,Package)
     */
    public Builder newRecordDef(String name, String pkgName)
    {
      return newRecordDef(name, Package.v(pkgName));
    }

    /**
     * Start building a new record definition with the given name in the given
     * package.
     * <p>
     * The default record label is the record's name.
     * The default record description is the empty string ("").
     *
     * @param name the brief name of the record type.
     * @param pkg  the package containing the record.
     *
     * @return the current builder.
     *
     * @throws IllegalStateException if already building a definition.
     * @throws IllegalArgumentException if <code>name</code> or <code>pkg</code>
     *   are <code>null</code>.
     */
    public Builder newRecordDef(String name, Package pkg)
    {
      if ( _building )
        throw new IllegalStateException("already building a definition");

      if ( name == null )
        throw new IllegalArgumentException("name is null");

      if ( pkg == null )
        throw new IllegalArgumentException("pkg is null");

      _building    = true;
      _name        = name;
      _pkg         = pkg;
      _label       = _name;
      _description = "";

      return this;
    }

    /**
     * Make the definition inherit from another.
     * @see #setParent(Type,boolean)
     */
    public Builder setParent(String parentName)
    {
      return setParent(Type.v(parentName), false);
    }

    /**
     * Make the definition inherit from another.
     * @see #setParent(Type,boolean)
     */
    public Builder setParent(String parentName, boolean cleanInherit)
    {
      return setParent(Type.v(parentName), cleanInherit);
    }

    /**
     * Make the definition inherit from another.
     * @see #setParent(Type,boolean)
     */
    public Builder setParent(Type parent)
    {
      return setParent(parent, false);
    }

    /**
     * Make the definition inherit from another.
     *
     * @param parent       the type to inherit from
     * @param cleanInherit if <code>true</code>,
     *   do not inherit the parent's record attributes
     *
     * @return the current builder.
     *
     * @throws IllegalStateException if not building a definition.
     * @throws IllegalArgumentException if <code>parent</code> is <code>null</code>.
     *
     * @see DerivedRecordDef
     */
    public Builder setParent(Type parent, boolean cleanInherit)
    {
      if ( ! _building )
        throw new IllegalStateException("not building a definition");

      if ( parent == null )
        throw new IllegalArgumentException("parent is null");

      _parent       = parent;
      _cleanInherit = cleanInherit;

      return this;
    }

    /**
     * Get the parent for the current definition.
     * @return the parent type,
     *   or <code>null</code> if no parent is currently defined.
     */
    public Type getParent()
    {
      return _parent;
    }

    /**
     * Set the label for the definition.
     * @return the current builder.
     * @throws IllegalStateException if not building a definition.
     * @throws IllegalArgumentException if <code>label</code> is <code>null</code>.
     */
    public Builder setLabel(String label)
    {
      if ( ! _building )
        throw new IllegalStateException("not building a definition");

      if ( label == null )
        throw new IllegalArgumentException("label is null");

      _label = label;

      return this;
    }

    /**
     * Get the label for the definition.
     * @return the label, or <code>null</code> if no label is currently defined.
     */
    public String getLabel()
    {
      return _label;
    }

    /**
     * Set the description for the definition.
     * @return the current builder.
     * @throws IllegalStateException if not building a definition.
     * @throws IllegalArgumentException if <code>description</code> is <code>null</code>.
     */
    public Builder setDescription(String description)
    {
      if ( ! _building )
        throw new IllegalStateException("not building a definition");

      if ( description == null )
        throw new IllegalArgumentException("description is null");

      _description = description;

      return this;
    }

    /**
     * Get the description for the definition.
     * @return the description,
     *   or <code>null</code> if no description is currently defined.
     */
    public String getDescription()
    {
      return _description;
    }

    /**
     * Add a record attribute to the definition.
     * @return the current builder.
     * @throws IllegalStateException if not building a definition.
     * @throws IllegalArgumentException if <code>attribute</code> is <code>null</code>.
     */
    public Builder addAttribute(Attribute attribute)
    {
      if ( ! _building )
        throw new IllegalStateException("not building a definition");

      if ( attribute == null )
        throw new IllegalArgumentException("attribute is null");

      _attributes.add(attribute);

      return this;
    }

    /**
     * Get the list of record attributes for the definition.
     * @return the current list of attributes, which may be empty.
     */
    public List getAttributes()
    {
      return _attributes;
    }

    /**
     * Add a field to the definition.
     * @return the current builder.
     * @throws IllegalStateException if not building a definition.
     * @throws IllegalArgumentException if <code>fieldDef</code> is <code>null</code>.
     */
    public Builder addField(FieldDef fieldDef)
    {
      if ( ! _building )
        throw new IllegalStateException("not building a definition");

      if ( fieldDef == null )
        throw new IllegalArgumentException("fieldDef is null");

      _fieldDefs.add(fieldDef);

      return this;
    }

    /**
     * Get the list of fields for the definition.
     * @return the current list of field definitions, which may be empty.
     */
    public List getFields()
    {
      return _fieldDefs;
    }

    /**
     * Add a field modifier to the definition.
     * @return the current builder.
     * @throws IllegalStateException if not building a definition.
     * @throws IllegalArgumentException if <code>modifier</code> is <code>null</code>.
     */
    public Builder modifyField(FieldModifier modifier)
    {
      if ( ! _building )
        throw new IllegalStateException("not building a definition");

      if ( modifier == null )
        throw new IllegalArgumentException("modifier is null");

      _modifiers.add(modifier);

      return this;
    }

    /**
     * Get the list of field modifiers for the definition.
     * @return the current list of field modifiers, which may be empty.
     */
    public List getModifiers()
    {
      return _modifiers;
    }

    /**
     * Set the factory object for the definition.
     * @return the current builder.
     * @throws IllegalStateException if not building a definition.
     * @throws IllegalArgumentException if <code>factory</code> is <code>null</code>.
     */
    public Builder setFactory(RecordFactory factory)
    {
      if ( ! _building )
        throw new IllegalStateException("not building a definition");

      if ( factory == null )
        throw new IllegalArgumentException("factory is null");

      _factory = factory;

      return this;
    }

    /**
     * Get the factory for the definition.
     * @return the factory, or <code>null</code> if no factory is currently defined.
     */
    public RecordFactory getFactory()
    {
      return _factory;
    }

    /**
     * Obtain the completed record definition.
     * <p>
     * NOTE: This definition will be set as the default for the type.
     *       Also, if the type is derived its parent is set as the parent in
     *       the definition.
     *
     * @return the fully constructed record definition.
     *
     * @throws IllegalStateException if not building a definition.
     *
     * @see TypeDef#getDefault(Type)
     * @see Type#getParent()
     */
    public RecordDef makeRecordDef()
    {
      if ( ! _building )
        throw new IllegalStateException("not building a definition");

      RecordDef recordDef = ( _parent == null ) ?
        new RecordDef(_name, _pkg) : new DerivedRecordDef(_name, _pkg, _parent, _cleanInherit);

      recordDef.setAttributes(_attributes);
      recordDef.setFields(_fieldDefs);
      recordDef.setModifiers(_modifiers);

      if ( _factory != null )
        recordDef.setFactory(_factory);

      recordDef.setAsDefault();

      if ( _parent != null )
        recordDef.getType().setParent(_parent);
      else
        recordDef.getType().setParent(StepRecord.TYPE);

      clear();

      return recordDef;
    }

    /**
     * Clear the current build context.
     */
    public void clear()
    {
      _building     = false;
      _name         = null;
      _pkg          = null;
      _parent       = null;
      _cleanInherit = false;
      _label        = null;
      _description  = null;
      _factory      = null;

      _attributes.clear();
      _fieldDefs.clear();
      _modifiers.clear();
    }
  }
}
