/* STEP - An Framework for the Efficient Encoding of Trace Data
 * Copyright (C) 2002 Rhodes Brown
 *
 * STEP is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * STEP is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with STEP; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA
 */

/* STEP is distributed through the Sable Research Group at:
 * http://www.sable.mcgill.ca/step/
 */

/* Reference Version: $ProjectVersion: 0.9.2.dev.36 $ */

package step.typedef;

import java.util.Map;

import step.*;

/**
 * Abstraction of a STEP type.
 */
public class Type implements StepDef
{
  /** The name extension used for array types. */
  public static final String ARRAY_EXTENSION = "[]";

  private static Map _NameToType   = new java.util.HashMap();
  private static Map _TypeToStepDL = new java.util.HashMap();

  private String _name;
  private Type   _parent;

  private Type(String name) // NOTE: only v() should create => enforce uniqueness
  {
    _name   = name; assert _name != null;
    _parent = null;

    _NameToType.put(_name, this);

    if ( isArrayType() )
    {
      setParent(StepObject.TYPE);
    }
  }

  /**
   * Get the unique type with the given name.
   * If the type does not exist, it will be created.
   * @see #v(String,Package)
   */
  public static Type v(String name)
  {
    return v(name, Package.ROOT);
  }

  /**
   * Get the type for the given object.
   * If the type does not exist, it will be created.
   * @see #v(String,Package)
   */
  public static Type v(StepObject o)
  {
    if ( o instanceof StepArray )
    {
      return v(((StepArray) o).getElements()[0]).toArrayType();
      // hack: assumes array size > 1
    }
    else
    {
      return v(o.getClass().getName());
    }
  }

  /**
   * Get the unique type with the given name in the given package.
   * If the type does not exist, it will be created.
   */
  public static Type v(String name, Package pkg)
  {
    Type type = (Type) _NameToType.get(pkg.qualify(name));

    if ( type == null )
      type = new Type(pkg.qualify(name)); // -> auto-register

    return type;
  }

  /**
   * Get the parent of this type, if it is inherited.
   * @return the parent type, or <code>null</code> if there is no parent.
   */
  public Type getParent()
  {
    return _parent;
  }

  /**
   * Set the parent for an inherited type.
   * @throws IllegalStateException if a parent has already been set for this type.
   * @throws IllegalArgumentException if <code>parent</code> is <code>null</code>.
   */
  public void setParent(Type parent)
  {
    if ( parent == null )
      throw new IllegalArgumentException("parent is null");

    if ( _parent != null )
      throw new IllegalStateException("attempt to redefine the parent of "+this);

    _parent = parent;
  }

  public boolean isParent(Type child)
  {
    if ( child == null ) // recursion base
      return false;

    return ( equals(child.getParent()) || isParent(child.getParent()) );
  }

  /** Is this type an array type? */ // ends in "[]"
  public boolean isArrayType()
  {
    return ( (_name.length() >= ARRAY_EXTENSION.length()) &&
             ( _name.lastIndexOf(ARRAY_EXTENSION) ==
               (_name.length()-ARRAY_EXTENSION.length()) ) );
  }

  /** Get the array type with this type as its base. */
  public Type toArrayType()
  {
    return v(_name + ARRAY_EXTENSION);
  }

  /**
   * If this is an array type, get the base type.
   * @return if this is an array type, the base type of the array.
   * @throws IllegalStateException if the current type is not an array type.
   */
  public Type toBaseType()
  {
    if ( isArrayType() )
    {
      return v(_name.substring(0, _name.length()-ARRAY_EXTENSION.length()));
    }
    else
    {
      throw new IllegalStateException(this+" is not an array type");
    }
  }

  /**
   * Compare two types for equality.
   * <p>
   * Two types are equal if they are the <em>same</em> object.
   * This equality property is enforced by the creational pattern for types.
   */
  public boolean equals(Object o)
  {
    return ( this == o );
  }

  /** Get a hash code for this type. */
  public int hashCode()
  {
    return _name.hashCode();
  }

  /** Get this type as a string. */
  public String toString()
  {
    return _name;
  }

  /** Get the STEP-DL name for this type. */
  public String toStepDL()
  {
    if ( isArrayType() )
    {
      return toBaseType().toStepDL()+ARRAY_EXTENSION;
    }
    else
    {
      String typeName = (String) _TypeToStepDL.get(this);

      if ( typeName == null ) typeName = toString(); // default

      return typeName;
    }
  }

  /** Set the STEP-DL name for this type. */
  public void setStepDL(String typeName)
  {
    _TypeToStepDL.put(this, typeName);
  }
}
