/* STEP - An Framework for the Efficient Encoding of Trace Data
 * Copyright (C) 2002 Rhodes Brown
 *
 * STEP is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * STEP is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with STEP; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA
 */

/* STEP is distributed through the Sable Research Group at:
 * http://www.sable.mcgill.ca/step/
 */

/* Reference Version: $ProjectVersion: 0.9.2.dev.36 $ */

package step.typedef;

import java.util.Map;
import java.util.List;
import java.util.Iterator;
import java.util.Collections;

import step.*;

/**
 * A definition for a STEP type.
 * <p>
 * Types have a {@linkplain #getDefault(Type) default} definition, however some
 * fields may choose to use an alternate, context-sensitive version of the
 * definition.
 *
 * @see FieldDef#getAltTypeDef()
 */
public class TypeDef implements Cloneable
{
  private static Map _TypeToDefaultDef = new java.util.HashMap();

  private String  _name;
  private Package _pkg;

  private List _fieldDefs;
  private Map  _nameToField;
  private List _modifiers;

  private boolean _resolved     = false;
  private boolean _typeResolved = false;

  private boolean _resolveLock = false;

  /**
   * @throws ClassCastException if
   * <ul>
   *   <li> the field list has a non-{@link FieldDef} element
   *   <li> the modifier list has a non-{@link FieldModifier} element
   * </ul>
   */
  public TypeDef(String name, Package pkg, List fieldDefs, List modifiers)
  {
    _name = name; assert _name != null;
    _pkg  = pkg;  assert _pkg  != null;

    if ( fieldDefs == null )
      setFields(new java.util.LinkedList());
    else
      setFields(fieldDefs);

    if ( modifiers == null )
      setModifiers(new java.util.LinkedList());
    else
      setModifiers(modifiers);
  }

  /**
   * Does the given type have a default definition set?
   * @see #setAsDefault()
   */
  public static boolean isDefined(StepObject o)
  {
    return isDefined(Type.v(o));
  }

  /**
   * Get the default (prototype) definition for the type of the given object.
   * @see #getDefault(Type)
   */
  public static TypeDef getDefault(StepObject o)
  {
    return getDefault(Type.v(o));
  }

  /**
   * Does the given type have a default definition set?
   * @see #setAsDefault()
   */
  public static boolean isDefined(String typeName)
  {
    return isDefined(Type.v(typeName));
  }

  /**
   * Get the default (prototype) definition for the given type name.
   * @see #getDefault(Type)
   */
  public static TypeDef getDefault(String typeName)
  {
    return getDefault(Type.v(typeName));
  }

  /**
   * Does the given type have a default definition set?
   * @see #setAsDefault()
   */
  public static boolean isDefined(Type type)
  {
    return _TypeToDefaultDef.containsKey(type);
  }

  /**
   * Get the default (prototype) definition for the given type.
   * @return the current defualt definition for the type,
   *   or <code>null</code> if no default exists for the type.
   *
   * @throws RecordDefNotFoundException if the type does not have a default
   *   definition and one cannot be loaded from the current environment.
   *
   * @see #setAsDefault()
   */
  public static TypeDef getDefault(Type type)
  {
    assert type != null;

    // do some silly bootstrapping to ensure base types are loaded:
    {
      StepInt.TYPE.toString();
      StepData.TYPE.toString();
      StepString.TYPE.toString();
    }

    TypeDef def = (TypeDef) _TypeToDefaultDef.get(type);

    if ( def == null )
    {
      if ( type.isArrayType() )
      {
        TypeDef baseDef = getDefault(type.toBaseType());

        if ( baseDef == null ) // => can't define
          return null;

        FieldDef elementField = (FieldDef) StepArray.ELEMENT_FIELD.clone();
        elementField.setType(baseDef.getType());

        List fields = new java.util.ArrayList(2);
        fields.add(StepArray.LENGTH_FIELD.clone());
        fields.add(elementField);

        def = new TypeDef(type.toString(), Package.ROOT, fields, null);
        def.setAsDefault();
      }
      else
      {
        def = RecordDef.load(type);
      }
    }

    return def;
  }

  /**
   * Set this definition as the default (prototype) definition for all objects
   * of this type.
   */
  public void setAsDefault()
  {
    _TypeToDefaultDef.put(getType(), this);
  }

  /**
   * Get the short name for the type represented by this definition.
   * @see #getFullName()
   */
  public String getName()
  {
    return _name;
  }

  /**
   * Get the fully-qualified name for the type represented by this definition.
   * @see #getName()
   * @see Package#qualify(String)
   */
  public String getFullName()
  {
    return _pkg.qualify(_name);
  }

  /**
   * Get the package that contains the type represented by this definition.
   */
  public Package getPackage()
  {
    return _pkg;
  }

  /**
   * Get the type that this definition is associated with.
   * @see Type#v(String,Package)
   */
  public Type getType()
  {
    return Type.v(_name, _pkg);
  }

  /**
   * Get the list of field definitions associated with this definition.
   * <p>
   * The list returned is {@linkplain Collections#unmodifiableList(List) unmodifiable}.
   */
  public List getFields()
  {
    return _fieldDefs;
  }

  /**
   * Replace the current list of field definitions.
   * @throws ClassCastException if an element of the list is not a {@link FieldDef}.
   */
  protected void setFields(List fieldDefs)
  {
    Map  nameToField  = new java.util.HashMap(2*fieldDefs.size());
    List newFieldDefs = new java.util.ArrayList(fieldDefs.size());

    Iterator iFieldDefs = fieldDefs.iterator();
    while (iFieldDefs.hasNext())
    {
      FieldDef fieldDef = (FieldDef) iFieldDefs.next();

      nameToField.put(fieldDef.getName(), fieldDef);
      newFieldDefs.add(fieldDef);
    }

    _nameToField = nameToField;
    _fieldDefs   = Collections.unmodifiableList(newFieldDefs);
  }

  /**
   * Get the field definition for the field with the given name.
   * @return the field definition,
   *   or <code>null</code> if no field is defined with the given name.
   */
  public FieldDef getField(String fieldName)
  {
    assert fieldName != null;

    return (FieldDef) _nameToField.get(fieldName);
  }

  /**
   * Get the list of field modifiers associated with this definition.
   * <p>
   * The list returned is {@linkplain Collections#unmodifiableList(List) unmodifiable}.
   */
  public List getModifiers()
  {
    return _modifiers;
  }

  /**
   * Replace the current list of field modifiers.
   * @throws ClassCastException if an element of the list is not a {@link FieldModifier}.
   */
  protected void setModifiers(List modifiers)
  {
    List newModifiers = new java.util.ArrayList(modifiers.size());
    Iterator iModifiers = modifiers.iterator();
    while (iModifiers.hasNext())
      newModifiers.add((FieldModifier) iModifiers.next());

    _modifiers = Collections.unmodifiableList(newModifiers);
  }

  private void addModifier(FieldModifier modifier)
  {
    assert modifier != null;

    List newModifiers = new java.util.ArrayList(getModifiers().size()+1);
    newModifiers.addAll(getModifiers());
    newModifiers.add(modifier);

    _modifiers = Collections.unmodifiableList(newModifiers);
  }

  /**
   * Is the type information for this definition complete?
   * @see #resolveType()
   * @see #isResolved()
   */
  public boolean isTypeResolved()
  {
    return _typeResolved;
  }

  /**
   * Complete the type information for this definition.
   * The default action is to check that the fields have defined types.
   * <p>
   * NOTE: Clients of this object can invoke this method indirectly via
   * {@link #resolve()}.
   *
   * @throws ResolveException if a field has an undefined type.
   */
  protected void resolveType()
  {
    if ( isTypeResolved() ) return;

    Iterator iFields = _fieldDefs.iterator();
    while (iFields.hasNext())
    {
      FieldDef field = (FieldDef) iFields.next();
      try
      {
        getDefault(field.getType());
      }
      catch (RecordDefNotFoundException e)
      {
        throw new ResolveException(this, "the type of field \""+
          field.getName()+"\" ("+field.getType()+") is undefined");
      }
    }

    _typeResolved = true;
  }

  /**
   * Is this definition complete?
   * <p>
   * In general, a type definition is complete (resolved) when all inherited
   * information has been incorporated, the existence of field types have been
   * verified, and the field modifiers have been applied.
   *
   * @see #resolve()
   */
  public boolean isResolved()
  {
    return isTypeResolved() && _resolved;
  }

  /**
   * Execute all operations needed to complete this definition.
   * <p>
   * By default, a type definition is completed when the type information is
   * resolved and the field modifiers have been applied.
   *
   * @see #resolveType()
   * @see #applyModifiers()
   */
  public void resolve()
  {
    resolveType();
    applyModifiers();
  }

  /**
   * Apply the list of modifiers to the various fields of this definition.
   * <p>
   * NOTE: Clients of this object can invoke this method indirectly via
   * {@link #resolve()}.
   *
   * @throws ResolveException if
   * <ul>
   *   <li> a modifier targeted at an undefined field
   *   <li> the {@linkplain #resolveType() type resolution} failed
   * </ul>
   */
  protected void applyModifiers()
  {
    if ( isResolved() ) return;

    if ( ! isTypeResolved() ) resolveType();

    if ( _resolveLock )
    {
      throw new ResolveException(this, "circular dependency detected");
    }
    _resolveLock = true;

    Iterator iModifiers = getModifiers().iterator();
    while (iModifiers.hasNext())
    {
      FieldModifier modifier = (FieldModifier) iModifiers.next();

      String target = modifier.getTarget();

      if ( target.indexOf(".") < 0 ) // => not a qualified modifier
      {
        FieldDef fieldDef = getField(target);

        if ( fieldDef == null )
        {
          throw new ResolveException(this, "can't modify field \""+target+"\""+
            ", it is not defined for \""+getType()+"\" types");
        }

        // modify the field's attributes
        {
          List attributes = new java.util.LinkedList(fieldDef.getAttributes());

          if ( modifier.isClean() ) attributes.clear(); // discard previous attributes

          attributes.addAll(modifier.getAttributes());

          fieldDef.setAttributes(attributes);
        }

        if ( fieldDef instanceof InheritedFieldDef )
        {
          ((InheritedFieldDef) fieldDef).setModified(true);
        }
      }
      else // => a qualified modifier
      {
        String subField = target.substring(target.indexOf(".")+1);
        target = target.substring(0, target.indexOf("."));

        FieldDef fieldDef = getField(target);

        if ( fieldDef == null )
        {
          throw new ResolveException(this, "can't modify field \""+target+"\""+
            ", it is not defined for \""+getType()+"\" types");
        }

        TypeDef altDef = fieldDef.getAltTypeDef();

        if ( altDef == null )
        {
          altDef = (TypeDef) getDefault(fieldDef.getType()).clone();
          fieldDef.setAltTypeDef(altDef);
        }

        getDefault(altDef.getType()).resolve(); // prevent circular mod references

        FieldModifier subModifier = new FieldModifier(subField, modifier.isClean());
        Iterator iAttributes = modifier.getAttributes().iterator();
        while (iAttributes.hasNext())
          subModifier.addAttribute((Attribute) iAttributes.next());

        altDef.addModifier(subModifier);
      }
    }

    // recurse to resolve sub-field modifiers
    {
      Iterator iFieldDefs = getFields().iterator();
      while (iFieldDefs.hasNext())
      {
        FieldDef fieldDef = (FieldDef) iFieldDefs.next();

        if ( fieldDef.hasAltTypeDef() )
        {
          fieldDef.getAltTypeDef().applyModifiers();
        }
      }
    }

    _resolved = true;
  }

  /**
   * Combine two definitions into one.
   * This method is used by {@link step.encoding.RecordEncoder}s to handle the
   * encoding of sub-types.
   *
   * @see FieldDef#merge(FieldDef,FieldDef)
   */
  public static TypeDef merge(TypeDef lowDef, TypeDef highDef)
  {
    if ( ! lowDef.getType().equals(highDef.getType()) )
      throw new IllegalArgumentException("types do not match");

    lowDef.resolve();
    highDef.resolve();

    List newFields = new java.util.ArrayList(lowDef.getFields().size());
    Iterator iLowFields  = lowDef.getFields().iterator();
    Iterator iHighFields = highDef.getFields().iterator();
    while (iLowFields.hasNext())
      newFields.add(FieldDef.merge((FieldDef) iLowFields.next(), (FieldDef) iHighFields.next()));

    return new TypeDef(lowDef.getName(), lowDef.getPackage(), newFields, null);
  }

  /**
   * Get a copy (clone) of this definition.
   * The clone is a deep copy of the field definitions and modifiers.
   */
  public Object clone() //: Cloneable
  {
    List dupFields = new java.util.ArrayList(getFields().size());
    Iterator iFieldDefs = getFields().iterator();
    while (iFieldDefs.hasNext())
      dupFields.add(((FieldDef) iFieldDefs.next()).clone());

    List dupModifiers = new java.util.ArrayList(getModifiers().size());
    Iterator iModifiers = getModifiers().iterator();
    while (iModifiers.hasNext())
      dupModifiers.add(((FieldModifier) iModifiers.next()).clone());

    return new TypeDef(getName(), getPackage(), dupFields, dupModifiers);
  }

  public void dump(int level)
  {
    for (int i=0; i<level; i++) System.out.print("  ");
    System.out.println("type="+getType());
    for (int i=0; i<level; i++) System.out.print("  ");
    System.out.println("resolved="+isResolved());

    Iterator iModifiers = getModifiers().iterator();
    while (iModifiers.hasNext())
    {
      for (int i=0; i<level+1; i++) System.out.print("  ");
      System.out.println(((FieldModifier) iModifiers.next()).toStepDL());
    }

    Iterator iFields = getFields().iterator();
    while (iFields.hasNext())
    {
      FieldDef fieldDef = (FieldDef) iFields.next();

      for (int i=0; i<level+1; i++) System.out.print("  ");
      System.out.println("field="+fieldDef.getName());

      if ( fieldDef.hasAltTypeDef() )
        fieldDef.getAltTypeDef().dump(level+2);
      else
      {
        for (int i=0; i<level+2; i++) System.out.print("  ");
        System.out.println("type="+fieldDef.getType());
      }
    }
  }
}
