/* STEP - An Framework for the Efficient Encoding of Trace Data
 * Copyright (C) 2002 Rhodes Brown
 *
 * STEP is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * STEP is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with STEP; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA
 */

/* STEP is distributed through the Sable Research Group at:
 * http://www.sable.mcgill.ca/step/
 */

/* Reference Version: $ProjectVersion: 0.9.2.dev.36 $ */

package step.util;

import java.util.LinkedList;
import java.util.Iterator;
import java.util.Map;

/**
 */
public class ByteArray
{
  public static byte[] dup(byte[] bytes)
  {
    return dup(bytes, 0, bytes.length);
  }

  public static byte[] dup(byte[] bytes, int offset, int len)
  {
    assert (offset+len) <= bytes.length;

    byte[] newBytes = allocator().alloc(len);

    int i, j;
    for (i=0,j=offset; i<len; i++,j++) newBytes[i] = bytes[j];

    return newBytes;
  }

  public static byte[] cat(byte[] bytes1, byte[] bytes2)
  {
    byte[] newBytes = allocator().alloc(bytes1.length+bytes2.length);

    int i,j=0;
    for (i=0; i<bytes1.length; i++) newBytes[j++] = bytes1[i];
    for (i=0; i<bytes2.length; i++) newBytes[j++] = bytes2[i];

    return newBytes;
  }

  public static byte[] cat(byte[] bytes1, byte[] bytes2, byte[] bytes3)
  {
    byte[] newBytes = allocator().alloc(bytes1.length+bytes2.length+bytes3.length);

    int i,j=0;
    for (i=0; i<bytes1.length; i++) newBytes[j++] = bytes1[i];
    for (i=0; i<bytes2.length; i++) newBytes[j++] = bytes2[i];
    for (i=0; i<bytes3.length; i++) newBytes[j++] = bytes3[i];

    return newBytes;
  }

  private static Allocator _Allocator;

  public static Allocator allocator()
  {
    if ( _Allocator == null )
    {
      String allocatorName =
        System.getProperty("step.ByteArray.allocator", "step.util.ByteArray$Allocator");

      try
      {
        _Allocator = (Allocator) Class.forName(allocatorName).newInstance();
      }
      catch (ClassNotFoundException e)
      {
        throw new RuntimeException(e);
      }
      catch (InstantiationException e)
      {
        throw new RuntimeException(e);
      }
      catch (IllegalAccessException e)
      {
        throw new RuntimeException(e);
      }
    }

    return _Allocator;
  }

  public static class Allocator
  {
    private LinkedList[] _L1; // lazy fill
    private Map          _L2; // lazy fill

    public Allocator()
    {
      _L1 = new LinkedList[256];
      _L2 = new java.util.HashMap(256, 0.5F);
    }

    public byte[] alloc(int size)
    {
      assert size >= 0;

      if ( size < _L1.length )
      {
        do
        {
          LinkedList sizeList = _L1[size];

          try
          {
            if ( sizeList.size() > 0 )
            {
              return (byte[]) sizeList.removeFirst(); // LIFO
            }
            else
            {
              return new byte[size];
            }
          }
          catch (NullPointerException e)
          {
            assert _L1[size] == null;

            _L1[size] = new LinkedList();
          }
        }
        while (true);
      }
      else // => L2
      {
        Integer    sizeInt  = new Integer(size);
        LinkedList sizeList = (LinkedList) _L2.get(sizeInt);

        do
        {
          try
          {
            if ( sizeList.size() > 0 )
            {
              return (byte[]) sizeList.removeFirst(); // LIFO
            }
            else
            {
              return new byte[size];
            }
          }
          catch (NullPointerException e)
          {
            assert sizeList == null;

            _L2.put(sizeInt, sizeList = new LinkedList());
          }
        }
        while (true);
      }
    }

    public void free(byte[] bytes)
    {
      int size = bytes.length;

      if ( size < _L1.length )
      {
        do
        {
          try
          {
            _L1[size].addFirst(bytes); // LIFO
            return;
          }
          catch (NullPointerException e)
          {
            assert _L1[size] == null;

            _L1[size] = new LinkedList();
          }
        }
        while (true);
      }
      else // => L2
      {
        Integer    sizeInt  = new Integer(size);
        LinkedList sizeList = (LinkedList) _L2.get(sizeInt);

        do
        {
          try
          {
            sizeList.addFirst(bytes); // LIFO
            return;
          }
          catch (NullPointerException e)
          {
            assert sizeList == null;

            _L2.put(sizeInt, sizeList = new LinkedList());
          }
        }
        while (true);
      }
    }
  }

  public static class StatsAllocator extends Allocator
  {
    private Map _sizeToStats = new java.util.LinkedHashMap();

    public StatsAllocator()
    {
      Runtime.getRuntime().addShutdownHook(new Thread()
      {
        public void run()
        {
          Iterator iStats = _sizeToStats.values().iterator();
          while (iStats.hasNext())
            ((Stats) iStats.next()).printSummary();
        }
      });
    }

    public byte[] alloc(int size)
    {
      Integer sizeInt = new Integer(size);
      Stats   stats   = (Stats) _sizeToStats.get(sizeInt);

      if ( stats == null )
        _sizeToStats.put(sizeInt, stats = new Stats(size));

      stats.alloc();

      return super.alloc(size);
    }

    public void free(byte[] bytes)
    {
      Integer sizeInt = new Integer(bytes.length);
      Stats   stats   = (Stats) _sizeToStats.get(sizeInt);

      if ( stats == null )
        _sizeToStats.put(sizeInt, stats = new Stats(bytes.length));

      stats.free();

      super.free(bytes);
    }
  }

  static class Stats
  {
    final int _size;
    int _allocCount = 0;
    int _freeCount  = 0;
    int _hitCount   = 0;
    int _load       = 0;
    int _totalLoad  = 0;

    public Stats(int size)
    {
      _size = size;
    }

    public void alloc()
    {
      _allocCount++;
      if ( _load > 0 )
      {
        _load--;
        _hitCount++;
      }
      _totalLoad += _load;
    }

    public void free()
    {
      _freeCount++;
      _load++;
      _totalLoad += _load;
    }

    public void printSummary()
    {
      System.out.println("byte["+_size+"]");
      System.out.println("\tallocs:   "+_size+" ("+((_size+4)*_allocCount)+" bytes)");
      System.out.println("\tfrees:    "+_freeCount+" ("+((_size+4)*_freeCount)+" bytes)");
      System.out.println("\thit rate: "+(((double) _hitCount)/((double) _allocCount)));
      System.out.println("\tavg load: "+(((double) _totalLoad)/((double) _allocCount+_freeCount)));
    }
  }
}
