/* STEP - An Framework for the Efficient Encoding of Trace Data
 * Copyright (C) 2002 Rhodes Brown
 *
 * STEP is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * STEP is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with STEP; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA
 */

/* STEP is distributed through the Sable Research Group at:
 * http://www.sable.mcgill.ca/step/
 */

/* Reference Version: $ProjectVersion: 0.9.2.dev.36 $ */

package stepc;

import java.io.File;
import java.io.FileReader;
import java.io.BufferedReader;
import java.io.PushbackReader;
import java.io.Reader;
import java.io.PrintWriter;
import java.io.PrintStream;
import java.io.IOException;

import java.util.List;
import java.util.Iterator;
import java.util.Map;
import java.util.Stack;

import step.*;
import step.typedef.*;
import step.util.*;

import stepc.lexer.*;
import stepc.parser.*;
import stepc.node.*;

/**
 */
public class Compiler
{
  /* $Format: "  public static final String VERSION = \"$StepcVersion$\";"$ */
  public static final String VERSION = "0.9.2";

  public static final int NORMAL_EXIT = 0;
  public static final int ERROR_EXIT  = 1;
  public static final int USAGE_EXIT  = 2;

  private PrintWriter _msgOut = new PrintWriter(new NullWriter());
  private PrintWriter _errOut = new PrintWriter(new NullWriter());

  private List _verifiedDefs = new java.util.LinkedList();
  private Map  _defNodes     = new java.util.HashMap();

  private int _errors   = 0;
  private int _warnings = 0;

  private Token _token;
    // As we visit the AST nodes, _token is set to the most recent representative
    // token so that when an error occurs it can be tagged to the appropriate token.

  /**
   * Create a new STEP-DL compiler.
   * <p>
   * NOTE: By default, <em>no</em> messages are printed.
   *
   * @see #setMsgOutput(PrintWriter)
   * @see #setErrOutput(PrintWriter)
   */
  public Compiler()
  {
  }

  /**
   * The default compiler application.
   * 
   * @param args the command line arguments.
   *   <br>Run <code>stepc.Compiler --help</code> for details.
   */
  public static void main(String[] args)
  {
    try
    {
      if ( args.length == 0 )
      {
        printHelp(System.err);
        System.exit(USAGE_EXIT);
      }

      final int NO_OUTPUT     = 0;
      final int STEPDL_OUTPUT = 1;
      final int JAVA_OUTPUT   = 2;
      final int XML_OUTPUT    = 3;

      int mode = NO_OUTPUT;
      File outputDir = null, outputFile = null;
      List inputFiles = new java.util.LinkedList();

      // parse args:
      for (int i=0; i<args.length; i++)
      {
        if ( args[i].equals("-h") || args[i].equals("--help") )
        {
          printHelp(System.out);
          System.exit(NORMAL_EXIT);
        }
        else if ( args[i].equals("--version") )
        {
          printVersion();
          System.exit(NORMAL_EXIT);
        }
        else if ( args[i].equals("-v") || args[i].equals("--verbose") )
        {
          System.setProperty("stepc.verbose", "true");
        }
        else if ( args[i].equals("--debug") )
        {
          System.setProperty("stepc.debug", "true");
        }
        else if ( args[i].equals("-d") )
        {
          if ( (++i) < args.length )
          {
            outputDir = new File(args[i]);

            if ( ! outputDir.isDirectory() )
            {
              System.err.println(outputDir+" is not a directory");
              printHelp(System.err);
              System.exit(USAGE_EXIT);
            }
          }
          else // missing PATH value
          {
            printHelp(System.err);
            System.exit(USAGE_EXIT);
          }
        }
        else if ( args[i].equals("-o") )
        {
          if ( (++i) < args.length )
          {
            outputFile = new File(args[i]);
          }
          else // missing FILE value
          {
            printHelp(System.err);
            System.exit(USAGE_EXIT);
          }
        }
        else if ( args[i].equals("--stepdl") )
        {
          if ( mode != NO_OUTPUT ) // other mode set
          {
            printHelp(System.err);
            System.exit(USAGE_EXIT);
          }
          mode = STEPDL_OUTPUT;
        }
        else if ( args[i].equals("-J") || args[i].equals("--java") )
        {
          if ( mode != NO_OUTPUT ) // other mode set
          {
            printHelp(System.err);
            System.exit(USAGE_EXIT);
          }
          mode = JAVA_OUTPUT;
        }
/*TODO: else if ( args[i].equals("-X") || args[i].equals("--xml") )
        {
          if ( mode != NO_OUTPUT ) // other mode set
          {
            printHelp(System.err);
            System.exit(USAGE_EXIT);
          }
          mode = XML_OUTPUT;
        }
*/
        else
        {
          File inputFile = new File(args[i]);

          if ( inputFile.canRead() )
          {
            inputFiles.add(inputFile);
          }
          else
          {
            System.err.println("unable to read "+inputFile);
            printHelp(System.err);
            System.exit(USAGE_EXIT);
          }
        }
      }

      if ( mode == NO_OUTPUT )
      {
        System.err.println("no output mode specified");
        printHelp(System.err);
        System.exit(USAGE_EXIT);
      }

      // defaults:
      PrintWriter output;
      {
        if ( outputDir == null ) outputDir = new File(".");

        output = ( outputFile == null ) ? 
          new PrintWriter(System.out) :
          new PrintWriter(new java.io.FileWriter(outputFile));
      }

      // emitter setup:
      Emitter emitter = null;
      switch ( mode )
      {
        case STEPDL_OUTPUT:
          emitter = new StepdlEmitter(output);
          break;
        case JAVA_OUTPUT:
          emitter = new JavaEmitter(outputDir);
          break;
        default:
          throw new RuntimeException("unhandled mode "+mode);
      }

      Compiler compiler = new Compiler();

      // message setup:
      compiler.setMsgOutput(new PrintWriter(System.out, true/*:autoFlush*/));
      compiler.setErrOutput(new PrintWriter(System.err, true/*:autoFlush*/));

      // read input:
      if ( inputFiles.size() == 0 ) // => read from stdin
      {
        compiler.parse(new java.io.InputStreamReader(System.in));
      }
      else
      {
        Iterator iInputs = inputFiles.iterator();
        while (iInputs.hasNext())
        {
          File inputFile = (File) iInputs.next();

          if ( (inputFiles.size() > 1) && isVerbose() ) System.out.println(inputFile);

          compiler.parse(new BufferedReader(new FileReader(inputFile), ioBufferSize()));
        }
      }

      if ( compiler.errCount() == 0 ) // => do something with the input...
      {
        compiler.emit(emitter);
      }

      if ( compiler.warnCount() > 0 )
      {
        System.err.println(compiler.warnCount()+" warning(s)");
      }
      if ( compiler.errCount() > 0 )
      {
        System.err.println(compiler.errCount()+" error(s)");
        System.exit(ERROR_EXIT);
      }
    }
    catch (Exception e)
    {
      if ( isDebug() )
        e.printStackTrace();
      else
        System.err.println(e);

      System.exit(ERROR_EXIT);
    }
  }

  /**
   * Read a text stream of STEP-DL definitions;
   * parse, resolve and verify the definitions.
   *
   * @return <code>true</code> if the parse was successful,
   *         <code>false</code> if there were errors.
   *
   * @throws IOException if an I/O failure occurred while reading the input.
   *
   * @see #errCount()
   * @see #warnCount()
   */
  public boolean parse(Reader input) throws IOException
  {
    _token = null; // clear context

    // 1. parse the input
    Start AST;
    try
    {
      AST = new Parser(new Lexer(new PushbackReader(input, ioBufferSize()))).parse();
    }
    catch (LexerException e)
    {
      error(e.getMessage());
      return false;
    }
    catch (ParserException e)
    {
      error(e.getMessage());
      return false;
    }

    // 2. build the type definitions from the AST
    List defs = new java.util.LinkedList();
    try
    {
      AST.apply(new Analysis(defs)); // -> add to defs
    }
    catch (AnalysisException e)
    {
      error(e.getMessage());
      return false;
    }

    // 3. resolve the type references and apply the modifiers
    {
      Iterator iDefs = defs.iterator();
      while (iDefs.hasNext())
      {
        RecordDef def = (RecordDef) iDefs.next();

        if ( isVerbose() ) _msgOut.println("Resolving \""+def.getFullName()+"\"...");

        try
        {
          def.resolve();
        }
        catch (ResolveException e)
        {
          _token = (Token) _defNodes.get(def);
          error(e.getMessage());
        }
      }
    }

_verifiedDefs.addAll(defs);
/*
    if ( _errors > 0 ) return false; // => don't try to verify attributes

    // 4. verify the attributes
    {
      Iterator iDefs = defs.iterator();
      while (iDefs.hasNext())
      {
        TypeDef def = (TypeDef) iDefs.next();

        if ( isVerbose() ) _msgOut.println("Verifying \""+def.getFullName()+"\"...");

        try
        {
          def.verifyAttributes();
          _verifiedDefs.add(def);
        }
        catch (AttributeVerifyException e)
        {
          _token = (Token) _defNodes.get(e.getSource());
          error(e.getMessage());
        }
      }
    }
*/

    return ( _errors == 0 );
  }
      
  /**
   * Apply an emitter to the current set of parsed definitions.
   *
   * @throws IOException if the emitter encountered an I/O failure.
   */
  public void emit(Emitter emitter) throws IOException
  {
    if ( _verifiedDefs.size() == 0 )
    {
      if ( isVerbose() ) _errOut.println("nothing to do");
      return;
    }

    Iterator iDefs = _verifiedDefs.iterator();
    while (iDefs.hasNext())
    {
      RecordDef def = (RecordDef) iDefs.next();

      if ( isVerbose() ) _msgOut.println("Creating \""+def.getFullName()+"\"...");

      emitter.emit(def);
    }
  }

  /**
   * Get the total parse error count so far.
   */
  public int errCount()
  {
    return _errors;
  } 

  /**
   * Get the total parse warning count so far.
   */
  public int warnCount()
  {
    return _warnings;
  } 

  /**
   * Direct normal messages to a given output stream.
   */
  public void setMsgOutput(PrintWriter output)
  {
    _msgOut = output;
  }

  /**
   * Direct error/warning messages to a given output stream.
   */
  public void setErrOutput(PrintWriter output)
  {
    _errOut = output;
  }


  // -- Local Interface --------------------------------------------------------

  static int ioBufferSize()
  {
    return Integer.getInteger("stepc.ioBufferSize", 4096).intValue();
  }

  static boolean isVerbose()
  {
    return Boolean.getBoolean("stepc.verbose");
  }

  static boolean isDebug()
  {
    return Boolean.getBoolean("stepc.debug");
  }

  private static void printHelp(PrintStream out)
  {
    System.out.println("STEPc version "+VERSION);
    out.println("Usage: stepc mode [option]* [input]*");
    out.println();
    out.println("Input:");
    out.println("  A series of STEP-DL files.");
    out.println("  If no input files are specified, the input is read from the");
    out.println("  standard input stream.");
    out.println();
    out.println("General Options:");
    out.println("  --version      Display the version information.");
    out.println("  -h, --help     Display this message.");
    out.println();
    out.println("Mode Options:");
    out.println("  -J, --java     Produce Java source.");
    //TODO: out.println("  -X, --xml      Produce an XML version of the input.");
    out.println("  --stepdl       Parse, verify & reprint the input as STEP-DL.");
    out.println();
    out.println("Output Options:");
    out.println("  -o FILE        For single file output, write to FILE.");
    out.println("  -d PATH        For multiple file output, produced files in PATH");
    out.println();
    out.println("Miscellaneous Options:");
    out.println("  -v, --verbose  Produce verbose output messages.");
    out.println("  --debug        Enable debugging messages.");
  }

  private static void printVersion()
  {
    System.out.println("STEPc version "+VERSION);
    System.out.println("Copyright (C) 2002 Rhodes Brown (rbrown@sable.mcgill.ca).");
    System.out.println("All rights reserved.");
    System.out.println();
    System.out.println("STEP (including STEPc) comes with ABSOLUTELY NO WARRANTY.");
    System.out.println("STEP is free software, and you are welcome to redistribute");
    System.out.println("it under certain conditions.");
    System.out.println("See the accompanying files 'COPYING.txt' or 'license.html'");
    System.out.println("for details.");
    System.out.println();
    System.out.println("Visit the STEP website:");
    System.out.println("  http://www.sable.mcgill.ca/step/");
  }

  private void error(String message)
  {
    if ( _token != null )
      _errOut.println("["+_token.getLine()+","+_token.getPos()+"]: "+message);
    else
      _errOut.println(message);

    _errors++;
  }

  private void warning(String message)
  {
    if ( _token != null )
      _errOut.println("["+_token.getLine()+","+_token.getPos()+"]: "+message);
    else
      _errOut.println(message);

    _warnings++;
  }

  // -- AST Visitor ------------------------------------------------------------

  private class Analysis extends stepc.analysis.AnalysisAdapter
  {
    private List _defs; 

    private step.typedef.Package _pkg = step.typedef.Package.ROOT;
      // current package scope

    private Stack _tmp = new Stack(); // multipurpose work stack

    private Type _type; // the current type, for list definitions

    public Analysis(List defs)
    {
      _defs = defs;
    }

    public void defaultCase(Node node)
    {
      throw new RuntimeException("unhandled syntax rule");
    }

    public void caseStart(Start node)
    {
      node.getPFile().apply(this);

      _token = null; // no more tokens
    }

    public void caseAFile(AFile node) // definition (, definition)*
    {
      Object defs[] = node.getDefinitions().toArray();
      for (int i=0; i<defs.length; i++)
      {
        try
        {
          ((PDefinition) defs[i]).apply(this);
        }
        catch (AnalysisException e)
        {
          error(e.getMessage());
          reset(); // try to recover...
        }
      }
    }

    public void caseAPackageDefinition(APackageDefinition node)
    {
      _token = node.getPackage();

      String pkgName = node.getName().getText();

      // open a new package scope
      _pkg = step.typedef.Package.v(pkgName, _pkg);

      Object defs[] = node.getContents().toArray();
      for (int i=0; i<defs.length; i++)
      {
        try
        {
          ((PDefinition) defs[i]).apply(this);
        }
        catch (AnalysisException e)
        {
          error(e.getMessage());
          reset(); // try to recover...
        }
      }

      // pop the package scope
      _pkg = _pkg.getParent();
    }

    public void caseARecordDefinition(ARecordDefinition node)
    {
      _token = node.getName();

      String recordName = node.getName().getText();

      if ( TypeDef.isDefined(_pkg.qualify(recordName)) )
        throw new AnalysisException("redefinition of type \""+recordName+"\"");

      RecordDef.builder().newRecordDef(recordName, _pkg);

      if (node.getSuperDeclaration() != null)
      {
        node.getSuperDeclaration().apply(this); // -> setParent()
      }

      if (node.getLabel() != null)
      {
        String label = StringConstant.decode(node.getLabel().getText());
        RecordDef.builder().setLabel(label);
      }

      if (node.getDescription() != null)
      {
        String description = StringConstant.decode(node.getDescription().getText());
        RecordDef.builder().setDescription(description);
      }

      {
        Object attrs[] = node.getAttributes().toArray();
        for (int i=0; i<attrs.length; i++)
        {
          ((PAttribute) attrs[i]).apply(this);
          Attribute attribute = (Attribute) pop();

          RecordDef.builder().addAttribute(attribute);
        }
      }

      {
        Object fields[] = node.getFields().toArray();
        for (int i=0; i<fields.length; i++)
          ((PFieldList) fields[i]).apply(this); // -> addField()
      }

      {
        Object modifiers[] = node.getModifiers().toArray();
        for (int i=0; i<modifiers.length; i++)
          ((PModifier) modifiers[i]).apply(this); // -> modifyField()
      }

      // get the completed record definition
      RecordDef recordDef = RecordDef.builder().makeRecordDef();

      // save the representative token (record name) for this definition
      _defNodes.put(recordDef, node.getName());

      // add the definition to the list of results
      _defs.add(recordDef);
    }

    public void caseASuperDeclaration(ASuperDeclaration node)
    {
      _token = node.getExtends();

      node.getParentName().apply(this);
      String parentName = (String) pop();

      boolean cleanInherit = (node.getBang() != null);

      if ( isRelative(parentName) )
        parentName = _pkg.qualify(parentName);

      RecordDef.builder().setParent(Type.v(parentName), cleanInherit);
    }

    public void caseAFieldList(AFieldList node) // field (, field)*
    {
      node.getTypeWithAttributes().apply(this);
      // -> type in _type
      // -> attributes in _tmp

      node.getFieldDefinition().apply(this);
      Object fields[] = node.getMoreFields().toArray();
      for (int i=0; i<fields.length; i++)
        ((PAnotherFieldDefinition) fields[i]).apply(this);
      
      // cleanup context
      _type = null;
      _tmp.clear();
    }
    public void caseAAnotherFieldDefinition(AAnotherFieldDefinition node)
    {
      node.getFieldDefinition().apply(this);
    }

    public void caseAFieldDefinition(AFieldDefinition node)
    {
      _token = node.getName();

      String fieldName = node.getName().getText();

      Iterator iFieldDefs = RecordDef.builder().getFields().iterator();
      while (iFieldDefs.hasNext())
      {
        if ( fieldName.equals(((FieldDef) iFieldDefs.next()).getName()) )
          throw new AnalysisException("redefinition of field");
      }

      FieldDef.builder().newFieldDef(fieldName, _type);

      if (node.getDescription() != null)
      {
        String description = StringConstant.decode(node.getDescription().getText());
        FieldDef.builder().setDescription(description);
      }

      // add the attributes for all fields in the current list
      {
        Iterator iTypeAttrs = _tmp.iterator();
        while (iTypeAttrs.hasNext())
          FieldDef.builder().addAttribute((Attribute) iTypeAttrs.next());
      }

      // add the attributes specific to this field
      {
        Object attrs[] = node.getAttributes().toArray();
        for (int i=0; i<attrs.length; i++)
        {
          ((PAttribute) attrs[i]).apply(this);
          Attribute attribute = (Attribute) pop();

          FieldDef.builder().addAttribute(attribute);
        }
      }

      // get the completed field definition
      FieldDef fieldDef = FieldDef.builder().makeFieldDef();

      // add the field definition to the current record definition
      RecordDef.builder().addField(fieldDef);

      // save the representative token (field name) for this definition
      _defNodes.put(fieldDef, node.getName());
    }

    public void caseAExtendModifier(AExtendModifier node)
    {
      _token = node.getTilde();

      node.getFieldName().apply(this);
      String fieldName = (String) pop();

      FieldModifier modifier = new FieldModifier(fieldName, false/*:clean*/);

      Object attrs[] = node.getAttributes().toArray();
      for (int i=0; i<attrs.length; i++)
      {
        ((PAttribute) attrs[i]).apply(this);
        Attribute attribute = (Attribute) pop();

        modifier.addAttribute(attribute);
      }

      if ( modifier.getAttributes().size() > 0 )
      {
        // add the modifier to the current record definition
        RecordDef.builder().modifyField(modifier);

        // save the representative token (~) for this definition
        _defNodes.put(modifier, node.getTilde());
      }
      else // => no modifications
      {
        warning("modifier contains no extensions");
      }
    }

    public void caseAOverrideModifier(AOverrideModifier node)
    {
      _token = node.getBang();

      node.getFieldName().apply(this);
      String fieldName = (String) pop();

      FieldModifier modifier = new FieldModifier(fieldName, true/*:clean*/);

      Object attrs[] = node.getAttributes().toArray();
      for (int i=0; i<attrs.length; i++)
      {
        ((PAttribute) attrs[i]).apply(this);
        Attribute attribute = (Attribute) pop();

        modifier.addAttribute(attribute);
      }

      // add the modifier to the current record definition
      RecordDef.builder().modifyField(modifier);

      // save the representative token (!) for this definition
      _defNodes.put(modifier, node.getBang());
    }

    public void caseAAttribute(AAttribute node)
    {
      _token = node.getGroup();

      String group = node.getGroup().getText();
      String value = StringConstant.decode(node.getValue().getText()).trim();

      Attribute attribute = new Attribute(group, value);

      // save the attribute to the _tmp stack
      push(attribute);

      // save the representative token (group name) for this definition
      _defNodes.put(attribute, node.getGroup());
    }

    public void caseATypeWithAttributes(ATypeWithAttributes node)
    {
      node.getType().apply(this);
      // -> type in _type

      {
        Object attrs[] = node.getAttributes().toArray();
        for (int i=0; i<attrs.length; i++)
          ((PAttribute) attrs[i]).apply(this);
      }
      // -> attributes in _tmp
    }

    public void caseAIntType(AIntType node)
    {
      _token = node.getInt();

      _type = StepInt.TYPE;
    }

    public void caseADataType(ADataType node)
    {
      _token = node.getData();

      _type = StepData.TYPE;
    }

    public void caseAStringType(AStringType node)
    {
      _token = node.getString();

      _type = StepString.TYPE;
    }

    // TODO: implement (=> allow) float types
    public void caseAFloatType(AFloatType node)
    {
      _token = node.getFloat();

      throw new AnalysisException(node.getFloat()+" type not supported yet");
    }

    public void caseAUserType(AUserType node)
    {
      node.getTypeName().apply(this);
      String typeName = (String) pop();

      if ( isRelative(typeName) )
        typeName = _pkg.qualify(typeName);

      _type = Type.v(typeName);
    }

    public void caseAArrayType(AArrayType node)
    {
      _token = node.getLBracket();

      node.getType().apply(this);
      // -> type in _type

      _type = _type.toArrayType();
    }

    public void caseASimpleItemName(ASimpleItemName node)
    {
      _token = node.getName();

      push(node.getName().getText());
    }

    public void caseAQualifiedItemName(AQualifiedItemName node)
    {
      _token = node.getInnerName();

      node.getOuterName().apply(this);
      String outerName = (String) pop();
      String innerName = node.getInnerName().getText();

      push(qualify(outerName, innerName));
    }


    // -- Helpers ----------------------------------------------------------------

    private void push(Object o)
    {
      assert o != null;
      _tmp.push(o);
    }

    private Object pop()
    {
      return _tmp.pop();
    }

    private void reset()
    {
      RecordDef.builder().clear();
      FieldDef.builder().clear();

      _type  = null;
      _token = null;

      _tmp.clear();
    }

    private String qualify(String outerName, String innerName)
    {
      return ( outerName + "." + innerName );
    }

    private boolean isRelative(String name)
    {
      return ( name.indexOf(".") < 0 );
    }

    private boolean isQualified(String name)
    {
      return ( ! isRelative(name) );
    }
  }
}
