/* STEP - An Framework for the Efficient Encoding of Trace Data
 * Copyright (C) 2002 Rhodes Brown
 *
 * STEP is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * STEP is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with STEP; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA
 */

/* STEP is distributed through the Sable Research Group at:
 * http://www.sable.mcgill.ca/step/
 */

/* Reference Version: $ProjectVersion: 0.9.2.dev.36 $ */

package stepc;

import java.io.File;
import java.io.PrintWriter;
import java.io.FileWriter;
import java.io.IOException;

import java.util.Stack;
import java.util.Iterator;

import step.*;
import step.typedef.*;
import step.util.*;

/**
 * Convert STEP definitions to Java source.
 */
public class JavaEmitter implements Emitter
{
  private File _outputDir;

  public JavaEmitter(File outputDir)
  {
    assert outputDir.isDirectory();

    _outputDir = outputDir;
  }

  public void emit(RecordDef def) throws IOException
  {
    File pkgDir = _outputDir;
    {
      Stack pkgs = new Stack();
      for (step.typedef.Package pkg = def.getPackage();
          ! pkg.equals(step.typedef.Package.ROOT);
          pkg = pkg.getParent())
      {
        pkgs.add(pkg);
      }

      pkgDir = _outputDir;
      while (pkgs.size() > 0)
        pkgDir = new File(pkgDir, ((step.typedef.Package) pkgs.pop()).getName());
    }
    pkgDir.mkdirs();

    File javaFile = new File(pkgDir, def.getName()+".java");

    javaFile.createNewFile();

    IndentWriter out =
      new IndentWriter(new PrintWriter(new FileWriter(javaFile)));

    out.println(
      "/* This file was generated by STEPc (see http://www.sable.mcgill.ca/step). */");
    out.println();

    if ( ! def.getPackage().equals(step.typedef.Package.ROOT) )
    {
      out.println("package "+def.getPackage().getFullName()+";");
      out.println();
    }

    out.println("import step.*;");
    out.println("import step.typedef.*;");
    out.println();

    out.print("public class "+def.getName());
    if ( def instanceof DerivedRecordDef )
    {
      DerivedRecordDef derivedDef = (DerivedRecordDef) def;
      out.println(" extends "+derivedDef.getParent());
      if ( ! (TypeDef.getDefault(derivedDef.getParent()) instanceof RecordDef) )
        out.println(" implements StepRecord");
    }
    else
    {
      out.println(" implements StepRecord");
    }
    out.println("{"); out.indent();

    // static
    {
      out.println("public static final RecordDef DEFINITION;");
      out.println();
      out.println("static");
      out.println("{"); out.indent();

      out.println("DEFINITION = RecordDef.builder().newRecordDef("+
        StringConstant.encode(def.getName())+", "+
        StringConstant.encode(def.getPackage().getFullName())+")");
      out.indent();

      if ( def instanceof DerivedRecordDef )
      {
        DerivedRecordDef derivedDef = (DerivedRecordDef) def;
        out.println(".setParent("+
          StringConstant.encode(derivedDef.getParent().toString())+", "+
          derivedDef.isCleanInherit()+")");
      }

      out.println(".setLabel("+
        StringConstant.encode(def.getLabel())+")");

      out.println(".setDescription("+
        StringConstant.encode(def.getDescription())+")");

      // record attributes
      {
        Iterator iAttributes = ( def instanceof DerivedRecordDef ) ?
          ((DerivedRecordDef) def).getExtraAttributes().iterator() :
          def.getAttributes().iterator();

        while (iAttributes.hasNext())
        {
          Attribute attr = (Attribute) iAttributes.next();
          out.println(".addAttribute(new Attribute("+
            StringConstant.encode(attr.getGroup())+", "+
            StringConstant.encode(attr.getValue())+"))");
        }
      }

      // local fields
      {
        Iterator iFieldDefs = def.getFields().iterator();
        while (iFieldDefs.hasNext())
        {
          FieldDef fieldDef = (FieldDef) iFieldDefs.next();

          if ( fieldDef instanceof LocalFieldDef )
          {
            out.println(".addField(FieldDef.builder().newFieldDef("+
              StringConstant.encode(fieldDef.getName())+", "+
              StringConstant.encode(fieldDef.getType().toString())+")");
            out.indent();
            out.println(".setDescription("+
              StringConstant.encode(fieldDef.getDescription())+")");
            Iterator iAttributes = fieldDef.getAttributes().iterator();
            while (iAttributes.hasNext())
            {
              Attribute attr = (Attribute) iAttributes.next();
              out.println(".addAttribute(new Attribute("+
                StringConstant.encode(attr.getGroup())+", "+
                StringConstant.encode(attr.getValue())+"))");
            }
            out.println(".makeFieldDef())");
            out.unindent();
          }
          else // inherited => do nothing
            ; 
        }
      }

      // modifiers
      {
        Iterator iModifiers = def.getModifiers().iterator();
        while (iModifiers.hasNext())
        {
          FieldModifier modifier = (FieldModifier) iModifiers.next();

          if ( def.getField(modifier.getTarget()) instanceof LocalFieldDef )
            continue;

          out.print(".modifyField(new FieldModifier("+
            StringConstant.encode(modifier.getTarget())+", "+
            modifier.isClean()+")");

          out.indent();
          Iterator iAttributes = modifier.getAttributes().iterator();
          while (iAttributes.hasNext())
          {
            out.println();
            Attribute attr = (Attribute) iAttributes.next();
            out.print(".addAttribute(new Attribute("+
              StringConstant.encode(attr.getGroup())+", "+
              StringConstant.encode(attr.getValue())+"))");
          }
          out.println(")");
          out.unindent();
        }
      }

      out.println(".setFactory(new Factory())");
      out.println(".makeRecordDef();");
      out.unindent();

      out.unindent();
      out.println("}");
    }
    out.println();

    // fields
    {
      Iterator iFieldDefs = def.getFields().iterator();
      while (iFieldDefs.hasNext())
      {
        FieldDef fieldDef = (FieldDef) iFieldDefs.next();
        if ( ! (fieldDef instanceof InheritedFieldDef) )
          out.println("public "+toJavaType(fieldDef.getType())+" "+fieldDef.getName()+";");
      }
    }
    out.println();

    // constructor
    {
      out.print("public "+def.getName()+"(");
      {
        Iterator iFieldDefs = def.getFields().iterator();
        for (int i=0; iFieldDefs.hasNext(); i++)
        {
          if ( i > 0 ) out.print(", ");
          FieldDef fieldDef = (FieldDef) iFieldDefs.next();
          out.print(toJavaType(fieldDef.getType())+" "+fieldDef.getName());
        }
        out.println(")");
      }
      out.println("{"); out.indent();
      if ( def instanceof DerivedRecordDef )
      {
        out.print("super(");
        Iterator iFieldDefs = def.getFields().iterator();
        for (int i=0; iFieldDefs.hasNext(); i++)
        {
          FieldDef fieldDef = (FieldDef) iFieldDefs.next();
          if ( fieldDef instanceof InheritedFieldDef )
          {
            if ( i > 0 ) out.print(", ");
            out.print(fieldDef.getName());
          }
        }
        out.println(");");
      }
      {
        Iterator iFieldDefs = def.getFields().iterator();
        while (iFieldDefs.hasNext())
        {
          FieldDef fieldDef = (FieldDef) iFieldDefs.next();
          if ( ! (fieldDef instanceof InheritedFieldDef) )
            out.println("this."+fieldDef.getName()+" = "+fieldDef.getName()+";");
        }
      }
      out.unindent(); out.println("}");
    }
    out.println();

    // fieldIterator()
    {
      out.println("public FieldIterator fieldIterator()");
      out.println("{"); out.indent();
      out.print("StepObject[] fields = { ");
      Iterator iFieldDefs = def.getFields().iterator();
      for (int i=0; iFieldDefs.hasNext(); i++)
      {
        if ( i > 0 ) out.print(", ");
        out.print(((FieldDef) iFieldDefs.next()).getName());
      }
      out.println(" };");
      out.println("return new FieldIterator(fields);");
      out.unindent(); out.println("}");
    }
    out.println();

    // equals()
    {
      out.println("public boolean equals(Object o)");
      out.println("{"); out.indent();
      if ( def.getFields().size() > 0 )
      {
        out.println(def.getName()+" rhs = ("+def.getName()+") o;");
        out.print("return ");
        Iterator iFieldDefs = def.getFields().iterator();
        for (int i=0; iFieldDefs.hasNext(); i++)
        {
          if ( i > 0 ) out.print("&&");
          FieldDef fieldDef = (FieldDef) iFieldDefs.next();
          out.print(fieldDef.getName()+".equals(rhs."+fieldDef.getName()+")");
        }
        out.println(";");
      }
      else // no fields!
      {
        out.println("return ( this == o );");
      }
      out.unindent(); out.println("}");
    }
    out.println();

    // hasCode()
    {
      out.println("public int hashCode()");
      out.println("{"); out.indent();
      if ( def.getFields().size() > 0 )
      {
        out.print("return ");
        Iterator iFieldDefs = def.getFields().iterator();
        for (int i=0; iFieldDefs.hasNext(); i++)
        {
          if ( i > 0 ) out.print("+");
          FieldDef fieldDef = (FieldDef) iFieldDefs.next();
          out.print(fieldDef.getName()+".hashCode()");
        }
        out.println(";");
      }
      else // no fields!
      {
        out.println("return super.hashCode();");
      }
      out.unindent(); out.println("}");
    }
    out.println();

    // Factory class
    {
      out.println("private static class Factory implements RecordFactory");
      out.println("{"); out.indent();
      out.println("public StepRecord newRecord(StepObject[] fieldData)");
      out.println("{"); out.indent();
      out.println("return new "+def.getName()+"(");
      out.indent(); 
      Iterator iFieldDefs = def.getFields().iterator();
      for (int i=0; iFieldDefs.hasNext(); i++)
      {
        FieldDef fieldDef = (FieldDef) iFieldDefs.next();
        out.print("("+toJavaType(fieldDef.getType())+") fieldData["+i+"]");
        if ( i < (def.getFields().size()-1) )
          out.println(",");
        else
          out.println();
      }
      out.unindent(); out.println(");");
      out.unindent(); out.println("}");
      out.unindent(); out.println("}");
    }

    out.unindent();
    out.println("}");

    out.close();
  }

  private String toJavaType(Type type)
  {
    if ( type.equals(StepInt.TYPE) )
    {
      return "StepInt";
    }
    else if ( type.equals(StepData.TYPE) )
    {
      return "StepData";
    }
    else if ( type.equals(StepString.TYPE) )
    {
      return "StepString";
    }
    else if ( type.isArrayType() )
    {
      return "StepArray";
    }
    else if ( TypeDef.getDefault(type) instanceof RecordDef )
    {
      return type.toString();
    }
    else
    {
      throw new RuntimeException("unknown type "+type);
    }
  }
}
